Package: poissonsuperlearner
Title: Poisson Super Learner
Version: 0.2.0
Authors@R: 
    c(person(given="Gabriele", family="Pittarello", email ="gabriele.pittarello@sund.ku.dk", role = c("aut", "cre")),
    person(given="Helene", family="Rytgaard", email ="hely@sund.ku.dk", role = c("aut")),
    person(given="Thomas", family="Gerds", email = "tag@biostat.ku.dk", role = c("aut")))
Description: Provides tools for fitting piecewise-constant hazard models for survival and competing risks data, including ensemble hazard estimation via the Super Learner framework. The package supports estimation of survival functions and absolute risk predictions from fitted cause-specific hazard models. For the Super Learner framework see van der Laan, Polley and Hubbard (2007) <doi:10.2202/1544-6115.1309>.
License: GPL (>= 2)
LinkingTo: Rcpp
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: data.table, sampling, riskRegression
Imports: Rcpp, methods, lava, Matrix, glmnet, mgcv
Suggests: knitr, rmarkdown, survival, prodlim, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-05-18 16:11:00 UTC; pwt887
Author: Gabriele Pittarello [aut, cre],
  Helene Rytgaard [aut],
  Thomas Gerds [aut]
Maintainer: Gabriele Pittarello <gabriele.pittarello@sund.ku.dk>
Repository: CRAN
Date/Publication: 2026-05-18 16:30:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-18 18:59:22 UTC; unix
Archs: poissonsuperlearner.so.dSYM

Package: poissonsuperlearner
Title: Poisson Super Learner
Version: 0.2.0
Authors@R: 
    c(person(given="Gabriele", family="Pittarello", email ="gabriele.pittarello@sund.ku.dk", role = c("aut", "cre")),
    person(given="Helene", family="Rytgaard", email ="hely@sund.ku.dk", role = c("aut")),
    person(given="Thomas", family="Gerds", email = "tag@biostat.ku.dk", role = c("aut")))
Description: Provides tools for fitting piecewise-constant hazard models for survival and competing risks data, including ensemble hazard estimation via the Super Learner framework. The package supports estimation of survival functions and absolute risk predictions from fitted cause-specific hazard models. For the Super Learner framework see van der Laan, Polley and Hubbard (2007) <doi:10.2202/1544-6115.1309>.
License: GPL (>= 2)
LinkingTo: Rcpp
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: data.table, sampling, riskRegression
Imports: Rcpp, methods, lava, Matrix, glmnet, mgcv
Suggests: knitr, rmarkdown, survival, prodlim, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-05-18 16:11:00 UTC; pwt887
Author: Gabriele Pittarello [aut, cre],
  Helene Rytgaard [aut],
  Thomas Gerds [aut]
Maintainer: Gabriele Pittarello <gabriele.pittarello@sund.ku.dk>
Repository: CRAN
Date/Publication: 2026-05-18 16:30:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-18 19:14:00 UTC; unix
Archs: poissonsuperlearner.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(coef,base_learner)
S3method(coef,poisson_superlearner)
S3method(predict,base_learner)
S3method(predict,poisson_superlearner)
S3method(predictRisk,base_learner)
S3method(predictRisk,poisson_superlearner)
S3method(print,base_learner)
S3method(print,poisson_superlearner)
S3method(summary,base_learner)
S3method(summary,poisson_superlearner)
export(Learner_gam)
export(Learner_glmnet)
export(Learner_hal)
export(Superlearner)
export(fit_learner)
export(pch_absolute_risk)
export(pch_absolute_risk_euler)
export(pch_survival)
export(predictRisk.poisson_superlearner)
export(simulateStenoT1)
exportClasses(Learner_gam)
exportClasses(Learner_glmnet)
exportClasses(Learner_hal)
import(data.table)
import(sampling)
importFrom(Matrix,sparse.model.matrix)
importFrom(Rcpp,evalCpp)
importFrom(glmnet,cv.glmnet)
importFrom(glmnet,glmnet)
importFrom(methods,new)
importFrom(mgcv,gam)
importFrom(riskRegression,predictRisk)
importFrom(stats,coef)
importFrom(stats,deviance)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,relevel)
importFrom(stats,time)
importFrom(utils,head)
importFrom(utils,tail)
useDynLib(poissonsuperlearner, .registration = TRUE)

[![R-CMD-check](https://github.com/gpitt71/poissonsuperlearner/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/gpitt71/poissonsuperlearner/actions/workflows/R-CMD-check.yaml)

# Poisson SuperLearner

The package provides an implementation of piecewise-constant hazard models for time-to-event analysis of survival and competing risks data. The piecewise-constant hazard models can be combined in an ensemble, the Poisson Superlearner, via cross-validated risk minimization for flexible hazard estimation. It enables estimation of survival functions and absolute risk predictions.

The package provides:

* Flexible specification of base learners (e.g. penalized and unpenalized Poisson regression).
* Automatic time-splitting into piecewise-constant hazard intervals.
* Cross-validated ensemble learning.
* Absolute risk prediction at arbitrary time horizons.

---

## Installation

```r
# install.packages("devtools")
# devtools::install_github("gpitt71/poissonsuperlearner")

library(poissonsuperlearner)
```

---

## Example: Piecewise-Constant Hazard Model

Fit a single PCH model and obtain absolute risk predictions.

```r
library(poissonsuperlearner)

set.seed(42)

# Simulate synthetic survival data
d <- simulateStenoT1(
  n = 50,
  scenario = "alpha"
)

# Define an unpenalized Poisson hazard learner
l_glm <- Learner_glmnet(
  covariates = c("sex", "diabetes_duration"),
  cross_validation = FALSE,
  lambda = 0
)

# Fit piecewise-constant hazard model
fit_glm <- fit_learner(
  data = d,
  id = "id",
  status = "event",
  event_time = "time",
  learner = l_glm,
  number_of_nodes = 5
)

# Absolute risk prediction at time horizon t = 5
predictRisk(fit_glm, newdata = d[1, ], times = 5)
```

**What happens internally?**

* The follow-up time is split into `number_of_nodes` intervals.
* A Poisson regression is fitted to the expanded dataset.
* The estimated hazards are integrated to produce survival and absolute risk.

---

## Example: Superlearner for Piecewise Hazards

Combine multiple hazard learners into an ensemble.

```r
library(poissonsuperlearner)

set.seed(42)

d <- simulateStenoT1(
  n = 50,
  scenario = "alpha"
)

# Base learner 1: unpenalized Poisson regression
l_glm <- Learner_glmnet(
  covariates = c("sex", "diabetes_duration"),
  cross_validation = FALSE,
  lambda = 0
)

# Base learner 2: Lasso-penalized Poisson regression
l_lasso <- Learner_glmnet(
  covariates = c("value_Smoking", "value_LDL"),
  cross_validation = TRUE,
  alpha = 1
)

learners_list <- list(
  glm   = l_glm,
  lasso = l_lasso
)

# Fit the superlearner
sl_fit <- Superlearner(
  data = d,
  id = "id",
  status = "event",
  event_time = "time",
  learners = learners_list,
  number_of_nodes = 5
)

# Absolute risk prediction from the ensemble
predictRisk(sl_fit, newdata = d[1, ], times = 5)
```

**Superlearner workflow**

1. Each base learner is fitted on cross-validation folds.
2. Out-of-sample Poisson deviance is computed.
3. A meta-learner combines the base hazard predictions.
4. Final hazards are aggregated and converted to absolute risk.

For competing risks, `learners` may also be a list of cause-specific learner
libraries, one library per cause. Predictions can be requested from the stacked
ensemble (`model = "sl"`), the discrete super learner (`model = "discrete_sl"`),
or named/indexed base learners.

---

## Main Functions

* `Learner_glmnet()` - define a Poisson hazard learner.
* `Learner_gam()` - define a Poisson GAM hazard learner.
* `Learner_hal()` - define a HAL-style Poisson hazard learner.
* `fit_learner()` - fit a single piecewise-constant hazard model.
* `Superlearner()` - fit an ensemble of hazard learners.
* `predict()` - obtain hazards, survival, and absolute risk predictions.
* `predictRisk()` - obtain risk matrices compatible with `riskRegression`.

---

## Typical Use Case

`poissonsuperlearner` is designed for:

* Survival analysis via Poisson representations.
* Flexible hazard modeling with regularization.
* Ensemble-based absolute risk prediction.
* Methodological research in time-to-event modeling.

The package focuses on modular learners, transparent cross-validation, and direct control of the piecewise hazard structure.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(poissonsuperlearner)
library(riskRegression)

## -----------------------------------------------------------------------------
d <- simulateStenoT1(
  n = 45,
  scenario = "alpha",
  competing_risks = TRUE,
  seed = 1
)

d <- d[, .(
  id,
  time,
  event,
  sex,
  age,
  diabetes_duration,
  value_LDL,
  value_Smoking
)]

head(d)

## -----------------------------------------------------------------------------
shared_library <- list(
  simple = Learner_glmnet(
    covariates = c("sex", "diabetes_duration"),
    cross_validation = FALSE,
    lambda = 0
  ),
  shrink = Learner_glmnet(
    covariates = c("sex", "age", "value_LDL"),
    cross_validation = FALSE,
    lambda = 0.05,
    alpha = 1
  )
)

fit_shared <- Superlearner(
  data = d,
  id = "id",
  status = "event",
  event_time = "time",
  learners = shared_library,
  number_of_nodes = 3,
  nfold = 2
)

## -----------------------------------------------------------------------------
summary(fit_shared)

## -----------------------------------------------------------------------------
newdata <- d[1:2]
times <- c(1, 2)

risk_shared_sl <- predictRisk(
  fit_shared, newdata = newdata, times = times, cause = 1, model = "sl"
)

risk_shared_discrete <- predictRisk(
  fit_shared, newdata = newdata, times = times, cause = 1, model = "discrete_sl"
)

risk_shared_simple <- predictRisk(
  fit_shared, newdata = newdata, times = times, cause = 1, model = "simple"
)

risk_shared_shrink <- predictRisk(
  fit_shared, newdata = newdata, times = times, cause = 1, model = "shrink"
)

list(
  sl = risk_shared_sl,
  discrete_sl = risk_shared_discrete,
  simple = risk_shared_simple,
  shrink = risk_shared_shrink
)

## -----------------------------------------------------------------------------
libraries_by_cause <- list(
  cvd = list(
    cvd_simple = Learner_glmnet(
      covariates = c("sex", "diabetes_duration"),
      cross_validation = FALSE,
      lambda = 0
    ),
    cvd_shrink = Learner_glmnet(
      covariates = c("age", "value_LDL"),
      cross_validation = FALSE,
      lambda = 0.05,
      alpha = 1
    )
  ),
  death = list(
    death_simple = Learner_glmnet(
      covariates = c("sex", "age"),
      cross_validation = FALSE,
      lambda = 0
    ),
    death_shrink = Learner_glmnet(
      covariates = c("diabetes_duration", "value_Smoking"),
      cross_validation = FALSE,
      lambda = 0.05,
      alpha = 1
    )
  )
)

fit_by_cause <- Superlearner(
  data = d,
  id = "id",
  status = "event",
  event_time = "time",
  learners = libraries_by_cause,
  number_of_nodes = 3,
  nfold = 2
)

## -----------------------------------------------------------------------------
summary(fit_by_cause)

## -----------------------------------------------------------------------------
risk_by_cause_sl <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1, model = "sl"
)

risk_by_cause_discrete <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1, model = "discrete_sl"
)

list(
  sl = risk_by_cause_sl,
  discrete_sl = risk_by_cause_discrete
)

## -----------------------------------------------------------------------------
cause_specific_model <- c("cvd_simple", "death_shrink")
cause_specific_model_alt <- c("cvd_shrink", "death_simple")

risk_by_cause_selected <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1,
  model = cause_specific_model
)

risk_by_cause_selected_alt <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1,
  model = cause_specific_model_alt
)

list(
  selected_learners = risk_by_cause_selected,
  selected_learners_alt = risk_by_cause_selected_alt
)

## -----------------------------------------------------------------------------
risk_by_cause_indexed <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1, model = c(1, 2)
)

risk_by_cause_indexed

#' Extract coefficients from a fitted base learner
#'
#' Convenience method to extract (cause-specific) model coefficients from a fitted
#' `base_learner` returned by [fit_learner()].
#'
#' For competing risks, `fit_learner()` fits one model per cause, stored in
#' `object$learner_fit[[k]]` for `k = 1, 2, ..., K`. This method simply dispatches
#' to the underlying model’s `coef()` method for each fitted object.
#'
#' @param object `base_learner`. A fitted object returned by [fit_learner()].
#' @param cause `numeric(1)` or `NULL`. Which cause to extract coefficients for.
#'   If `NULL`, coefficients are returned for all causes.
#'   Causes are indexed `1, 2, ..., object$data_info$n_crisks` (with `0` reserved for censoring).
#' @param ... Passed to the underlying `coef()` method of the fitted learner object
#'   (learner-dependent; e.g., `s` for `glmnet`).
#'
#' @details
#' **Learner-dependent output.** The returned coefficient object depends on the
#' base learner used (e.g. a numeric vector, a sparse matrix, a list, etc.).
#' This method does not post-process or rename coefficients; it returns the output
#' of `coef(object$learner_fit[[k]], ...)` unchanged.
#'
#' @return
#' If `cause` is a single integer, returns the coefficient object produced by
#' `coef()` for that cause-specific fitted model.
#'
#' If `cause = NULL`, returns a `list` of length `object$data_info$n_crisks`,
#' where element `[[k]]` contains coefficients for cause `k`.
#'
#' If no fitted model is present (`object$learner_fit` is `NULL`), signals a message
#' and returns `invisible(object)`.
#'
#' @examples
#' d <- simulateStenoT1(50, competing_risks = TRUE)
#' lrn <- Learner_glmnet(covariates = c("age", "value_LDL"),
#'                       lambda = 0, cross_validation = FALSE)
#' bl <- fit_learner(d, learner = lrn, id = "id",
#'                   status = "status_cvd", event_time = "time_cvd",
#'                   number_of_nodes = 4)
#'
#' # coefficients for cause 1
#' coef(bl, cause = 1)
#'
#' # coefficients for all causes (list)
#' coef(bl)
#'
#' @export
coef.base_learner <- function(object, cause = NULL, ...) {

  if (is.null(object$learner_fit)) {
    message("No fitted model available (learner_fit is NULL).")
    return(invisible(object))
  }

  if (is.null(cause)) {
    return(lapply(object$learner_fit, coef, ...))
  } else {
    return(coef(object$learner_fit[[cause]], ...))
  }
}

#' Extract stacking (meta-learner) coefficients from a fitted Poisson Super Learner
#'
#' Extracts the **meta-learner coefficients** (stacking weights) from a fitted
#' `poisson_superlearner` object returned by [Superlearner()].
#'
#' For each cause `k`, the ensemble stores a fitted meta-learner in
#' `object$superlearner[[k]]$meta_learner_fit`. This method dispatches to the
#' underlying `coef()` method for that fitted meta-learner.
#'
#' @param object `poisson_superlearner`. A fitted ensemble returned by [Superlearner()].
#' @param cause `numeric(1)` or `NULL`. Which cause to extract meta-learner
#'   coefficients for. If `NULL`, coefficients are returned for all causes.
#'   Causes are indexed `1, 2, ..., object$data_info$n_crisks`.
#' @param model Scalar model selector. Default is `"sl"` for the stacked super learner.
#'   Other allowed values are:
#'   \describe{
#'     \item{`0` or `"sl"`}{Use the super learner prediction.}
#'     \item{learner label}{Use one stored base learner by its label in
#'       `object$data_info$learners_labels`.}
#'     \item{`"learner_j"`}{Use the `j`-th stored learner.}
#'     \item{integer `j >= 1`}{Use the `j`-th stored learner.}
#'   }
#' @param ... Passed to the underlying `coef()` method of the fitted meta-learner
#'   (learner-dependent; e.g., `s` for `glmnet`).
#'
#' @details
#' **What coefficients represent.** These coefficients correspond to the meta-learner
#' regression of the outcome on the cross-validated base-learner predictions
#' (`Z1`, `Z2`, ...). Under the default meta-learner, they are the stacking
#' weights (on the scale defined by the meta-learner).
#'
#' **Learner-dependent output.** The returned coefficient object depends on the
#' meta-learner implementation (by default a `glmnet` fit, often returning a sparse
#' matrix). This method does not rename `Z*` terms or post-process coefficients; it
#' returns the output of `coef(object$superlearner[[k]]$meta_learner_fit, ...)`
#' unchanged.
#'
#' **Single-learner special case.** If the ensemble was fit with only one base learner,
#' no meta-learner is fit and `meta_learner_fit` is `NULL`. In that case, `coef()`
#' for the `poisson_superlearner` does not have meta-learner coefficients to return.
#'
#' @return
#' If `cause` is a single integer, returns the coefficient object produced by
#' `coef()` for the cause-specific fitted meta-learner.
#'
#' If `cause = NULL`, returns a `list` of length `object$data_info$n_crisks`,
#' where element `[[k]]` contains meta-learner coefficients for cause `k`.
#'
#' If no fitted ensemble is present (`object$superlearner` is `NULL`), signals a message
#' and returns `invisible(object)`.
#'
#' @examples
#' d <- simulateStenoT1(50, competing_risks = TRUE)
#' learners <- list(
#'   glm = Learner_glmnet(covariates = c("age", "value_LDL"), lambda = 0, cross_validation = FALSE),
#'   gam = Learner_gam(covariates = c("age", "value_LDL"))
#' )
#' fit <- Superlearner(d, id="id", status="status_cvd", event_time="time_cvd",
#'                     learners=learners, number_of_nodes=4, nfold=2)
#'
#' # meta-learner coefficients (cause 1)
#' coef(fit, cause = 1)
#'
#' # meta-learner coefficients for all causes (list)
#' coef(fit)
#'
#' @export
coef.poisson_superlearner <- function(object, cause = NULL, model = "sl", ...) {

  if (is.null(object$superlearner)) {
    message("No fitted model available (superlearner is NULL).")
    return(invisible(object))
  }

  if (is.null(cause)) {
    out <- lapply(seq_len(object$data_info$n_crisks), function(k) {
      fit_info <- psl_get_stored_fit(object, cause = k, model = model)
      stats::coef(fit_info$fit, ...)
    })
    names(out) <- paste0("cause_", seq_len(object$data_info$n_crisks))
    return(out)
  }

  fit_info <- psl_get_stored_fit(object, cause = cause, model = model)
  stats::coef(fit_info$fit, ...)
}

#' Fit a Poisson Super Learner ensemble
#'
#' Fits an ensemble of cause-specific piecewise-constant hazard models using a
#' long-format Poisson representation and combines them through a meta-learner
#' (stacking).
#'
#' Internally, the function:
#' \enumerate{
#'   \item builds a time grid (`nodes`) and converts the subject-level data to a
#'   long Poisson format;
#'   \item fits each base learner once on the full long data for each cause;
#'   \item removes learners that already fail on the full data;
#'   \item uses `nfold` cross-validation to obtain out-of-sample base-learner
#'   predictions (`Z1`, `Z2`, ...) for stacking;
#'   \item removes learners whose cross-validated prediction column is entirely
#'   missing for at least one cause;
#'   \item fits a cause-specific meta-learner on the retained stacked predictions.
#' }
#'
#' @param data `data.frame`. Subject-level input data, one row per subject.
#' @param id `character(1)`. Name of the subject identifier column. If missing,
#'   an `id` column is created automatically.
#' @param status `character(1)`. Name of the event-status column. It must be coded
#'   with `0` for censoring and `1, 2, ..., K` for event types. If there is no
#'   `0` in `status`, the data are treated as uncensored.
#' @param event_time `character(1)`. Name of the event or censoring time column.
#' @param learners `list`. List of initialized learner reference-class objects,
#'   for example [Learner_glmnet()], [Learner_hal()], or [Learner_gam()]. If
#'   unnamed, learners are named `"learner_1"`, `"learner_2"`, and so on. Each
#'   learner must implement `$private_fit(dt_long)` and
#'   `$private_predictor(model, newdata)`.
#' @param number_of_nodes `numeric(1)` or `NULL`. If not `NULL`, constructs a
#'   quantile-based node grid with `number_of_nodes + 1` cut points. Ignored when
#'   `nodes` is supplied.
#' @param nodes `numeric` or `NULL`. Explicit time-node grid. If supplied,
#'   `number_of_nodes` is ignored. `0` is added if missing, and nodes larger than
#'   `max(event_time)` are dropped.
#' @param variable_transformation Optional transformation specification passed to
#'   `apply_transformations()` on the internally created long-format data.
#' @param nfold `numeric(1)`. Number of folds for cross-validation stacking.
#' @param ... Additional arguments currently ignored.
#'
#' @return An object of class `poisson_superlearner`, stored as a named `list`
#'   with the following components:
#'
#'   `learners`:
#'   the retained base learner objects.
#'
#'   `metalearner`:
#'   the meta-learner object used for stacking. If no stacking is performed because
#'   only one learner remains, `metalearner` is `NULL`.
#'
#'   `superlearner`:
#'   a `list` of length `data_info$n_crisks`, one entry per cause. For cause `k`,
#'   `superlearner[[k]]` is a `list` with two elements:
#'   \itemize{
#'     \item `learners_fit`: the fitted base learner object or objects for cause `k`.
#'     If more than one learner is retained, this is a `list` with one fitted
#'     object per retained learner. If only one learner remains, this is the
#'     single fitted learner object itself.
#'     \item `meta_learner_fit`: the fitted cause-specific meta-learner for cause
#'     `k`. If no stacking is performed, this is `NULL`.
#'   }
#'
#'   `cross_validation_deviance`:
#'   a `data.table` with columns `learner` and `deviance`, giving the mean
#'   cross-validated Poisson deviance for each retained base learner. This
#'   component is present when cross-validated model comparison is available.
#'
#'   `data_info`:
#'   a `list` of bookkeeping information used for prediction and interpretation,
#'   containing:
#'   \itemize{
#'     \item `id`: identifier column name used.
#'     \item `status`: status column name used.
#'     \item `event_time`: event-time column name used.
#'     \item `nodes`: numeric vector of node cut points used for the piecewise grid.
#'     \item `nfold`: number of folds used for stacking.
#'     \item `maximum_followup`: maximum observed follow-up time.
#'     \item `n_crisks`: number of event types detected.
#'     \item `learners_labels`: character vector of retained learner labels.
#'     \item `variable_transformation`: the transformation specification passed in
#'     `variable_transformation`, or `NULL`.
#'   }
#'
#' @details
#' If all learners fail on the full data, the function stops with an error.
#' If only one learner remains after the full-data screening step or after the
#' cross-validation screening step, no meta-learner is fit. In that case,
#' `metalearner` is `NULL`, each `superlearner[[k]]$meta_learner_fit` is `NULL`,
#' and prediction is based directly on the stored fitted base learner.Numeric
#' learner positions always refer to the learners actually retained in the
#' fitted object.
#'
#' @examples
#' data <- simulateStenoT1(50, competing_risks = TRUE)
#'
#' learners <- list(
#'   glm = Learner_glmnet(
#'     covariates = c("sex", "value_LDL"),
#'     lambda = 0,
#'     cross_validation = FALSE
#'   ),
#'   ridge = Learner_glmnet(
#'     covariates = c("sex", "value_LDL"),
#'     alpha = 0,
#'     lambda = 0.01,
#'     cross_validation = FALSE
#'   )
#' )
#'
#' fit <- Superlearner(
#'   data = data,
#'   id = "id",
#'   status = "status_cvd",
#'   event_time = "time_cvd",
#'   learners = learners,
#'   number_of_nodes = 3,
#'   nfold = 2
#' )
#'
#' @export
Superlearner <- function(data,
                         id = "id",
                         status = "status",
                         event_time = NULL,
                         learners,
                         number_of_nodes = NULL,
                         nodes = NULL,
                         variable_transformation = NULL,
                         nfold = 3,
                         ...) {

  if (!(id %in% names(data))) {
    data[["id"]] <- 1:NROW(data)
    id <- "id"
  }

  if (is.null(names(learners))) {
    names(learners) <- paste0("learner_", seq_along(learners))
  }

  learners_labels <- names(learners)

  maximum_followup <- max(data[[event_time]])
  n <- length(unique(data[[id]]))

  if (!(0 %in% data[[status]])) {
    warning(
      paste0(
        "There is no value of ",
        status,
        " equal to zero. We will consider the data uncensored."
      )
    )
    n_crisks <- length(unique(data[[status]]))
    uncensored_01 <- TRUE
  } else {
    n_crisks <- length(unique(data[[status]])) - 1L
    uncensored_01 <- FALSE
  }

  if (!is.null(number_of_nodes)) {
    grid_nodes <- quantile(
      data[[event_time]],
      probs = seq(0, 1, length.out = as.integer(number_of_nodes) + 1L),
      type = 1,
      names = FALSE
    )
  } else if (is.null(nodes)) {
    grid_nodes <- sort(unique(data[[event_time]]))
  } else {
    grid_nodes <- nodes
  }

  if (!(0 %in% grid_nodes)) {
    grid_nodes <- c(0, grid_nodes)
  }

  grid_nodes <- grid_nodes[grid_nodes <= max(data[[event_time]])]

  dt <- data_pre_processing(
    data = data,
    id = id,
    status = status,
    nodes = grid_nodes,
    event_time = event_time,
    uncensored_01 = uncensored_01
  )

  if (!is.null(variable_transformation)) {
    apply_transformations(dt, variable_transformation)
  }

  id_fold <- sample(
    1:nfold,
    n,
    replace = TRUE,
    prob = rep(1 / nfold, nfold)
  )

  dt_id <- data.table(folder = id_fold, id = unique(data[[id]]))
  dt <- merge(dt, dt_id, by = "id", all.x = TRUE)

  dt_by_cause <- split(dt, by = "k")

  z_covariates <- paste0("Z", seq_along(learners))

  ## ------------------------------------------------------------
  ## Step 1: fit learners once on the full data, one list per cause
  ## ------------------------------------------------------------
  full_train_list <- lapply(dt_by_cause, function(dt_k) {
    lapply(learners, function(f) f$private_fit(dt_k))
  })

  ## Remove learners that already fail on the full data
  failed_by_full_fit <- lapply(full_train_list, function(fits_k) {
    z_covariates[vapply(fits_k, is_failed_fit, logical(1))]
  })

  failed_learners <- Reduce(union, failed_by_full_fit)

  if (length(failed_learners) > 0L) {
    keep_z <- setdiff(z_covariates, failed_learners)
    keep_ix <- match(keep_z, z_covariates)

    learners <- learners[keep_ix]
    learners_labels <- learners_labels[keep_ix]
    z_covariates <- keep_z

    full_train_list <- lapply(full_train_list, function(fits_k) fits_k[keep_ix])
  }

  if (length(z_covariates) == 0L) {
    stop("All learners failed on the full data before cross-validation.")
  }

  ## If pruning leaves one learner, return direct fit
  if (length(learners) == 1L) {
    message("Only one usable base learner remains. Fitting the learner directly; no ensemble constructed.")

    one_learner_out <- vector("list", n_crisks)

    for (cause_ix in seq_len(n_crisks)) {
      one_learner_out[[cause_ix]] <- list(
        learners_fit = full_train_list[[cause_ix]][[1L]],
        meta_learner_fit = NULL
      )
    }

    out <- list(
      learners = learners,
      metalearner = NULL,
      superlearner = one_learner_out,
      data_info = list(
        id = id,
        status = status,
        event_time = event_time,
        nodes = sort(unique(as.numeric(levels(dt$node)))),
        nfold = nfold,
        maximum_followup = maximum_followup,
        n_crisks = n_crisks,
        learners_labels = learners_labels,
        variable_transformation = variable_transformation
      )
    )

    class(out) <- "poisson_superlearner"
    return(out)
  }

  ## ------------------------------------------------------------
  ## Step 2: V-fold CV only on retained learners
  ## ------------------------------------------------------------
  dt_by_cause_cv <- lapply(dt_by_cause, function(dt_k) {
    out <- copy(dt_k)
    out[, row_ix := .I]
    out
  })

  oof_buffers <- lapply(
    dt_by_cause_cv,
    psl_init_oof_buffer,
    z_covariates = z_covariates
  )

  for (ix in seq_len(nfold)) {
    pseudo_observations <- mapply(
      function(dt_k, competing_risk) {
        create_pseudo_observations(
          training_data = dt_k[folder != ix, ],
          validation_data = dt_k[folder == ix, ],
          competing_risk = competing_risk,
          learners = learners,
          z_covariates = z_covariates,
          ix = ix
        )
      },
      dt_k = dt_by_cause_cv,
      competing_risk = as.list(seq_len(n_crisks)),
      SIMPLIFY = FALSE,
      USE.NAMES = FALSE
    )

    for (k in seq_len(n_crisks)) {
      oof_buffers[[k]] <- psl_write_oof_buffer(
        oof_buffer = oof_buffers[[k]],
        oof_chunk = pseudo_observations[[k]]
      )
    }
  }

## ------------------------------------------------------------
## Step 3: CV deviance on retained learners
## ------------------------------------------------------------
L <- length(z_covariates)
dev_sum <- matrix(0.0, nrow = nfold, ncol = L)

for (k in seq_len(n_crisks)) {
  loghaz_cols <- psl_oof_loghaz_dt(
    oof_buffer = oof_buffers[[k]],
    z_covariates = z_covariates
  )

  dev_k <- poisson_deviance_by_folder_cols(
    log_hazard_cols = loghaz_cols,
    tij = as.numeric(dt_by_cause_cv[[k]][["tij"]]),
    delta = as.integer(dt_by_cause_cv[[k]][["deltaij"]]),
    folder = as.integer(dt_by_cause_cv[[k]][["folder"]]),
    nfold = nfold
  )

  dev_sum <- dev_sum + dev_k
}

  dev_mean <- colMeans(2.0 * dev_sum)

  dt_learners <- data.table::data.table(
    learner = learners_labels,
    deviance = dev_mean
  )

  ## ------------------------------------------------------------
  ## Step 4: after CV, remove only learners that are ALL-NA
  ## within at least one cause. Partial fold failures are allowed.
  ## ------------------------------------------------------------
  failed_by_risk <- lapply(oof_buffers, function(buf) {
    z_covariates[colSums(!is.na(buf$Z)) == 0L]
  })

  failed_cv_learners <- Reduce(union, failed_by_risk)

  if (length(failed_cv_learners) > 0L) {
    keep_z <- setdiff(z_covariates, failed_cv_learners)
    keep_ix <- match(keep_z, z_covariates)

    oof_buffers <- lapply(oof_buffers, function(buf) {
      buf$Z <- buf$Z[, keep_ix, drop = FALSE]
      buf
    })

    learners <- learners[keep_ix]
    learners_labels <- learners_labels[keep_ix]
    z_covariates <- keep_z
    full_train_list <- lapply(full_train_list, function(fits_k) fits_k[keep_ix])

    dt_learners <- dt_learners[learner %in% learners_labels]
  }

  if (length(z_covariates) == 0L) {
    stop("All learners failed: every cross-validated prediction column was entirely NA in at least one competing risk.")
  }

  ## If post-CV pruning leaves one learner, return direct fit
  if (length(learners) == 1L) {
    message("Only one usable base learner remains after cross-validation screening. No ensemble constructed.")

    one_learner_out <- vector("list", n_crisks)

    for (cause_ix in seq_len(n_crisks)) {
      one_learner_out[[cause_ix]] <- list(
        learners_fit = full_train_list[[cause_ix]][[1L]],
        meta_learner_fit = NULL
      )
    }

    out <- list(
      learners = learners,
      metalearner = NULL,
      superlearner = one_learner_out,
      cross_validation_deviance = dt_learners,
      data_info = list(
        id = id,
        status = status,
        event_time = event_time,
        nodes = sort(unique(as.numeric(levels(dt$node)))),
        nfold = nfold,
        maximum_followup = maximum_followup,
        n_crisks = n_crisks,
        learners_labels = learners_labels,
        variable_transformation = variable_transformation
      )
    )

    class(out) <- "poisson_superlearner"
    return(out)
  }

  ## ------------------------------------------------------------
  ## Step 5: meta-learning, without duplicating the meta-learner
  ## ------------------------------------------------------------
  meta_learner <- Learner_glmnet(
    covariates = z_covariates,
    cross_validation = FALSE,
    intercept = FALSE,
    add_nodes = FALSE,
    penalise_nodes = TRUE,
    lambda = 0
  )

  meta_learner_fits <- mapply(
    function(dt_k, oof_k) {
      fit_meta_learner(
        dt = dt_k,
        dt_z = oof_k,
        meta_learner = meta_learner,
        z_covariates = z_covariates
      )
    },
    dt_by_cause,
    oof_buffers,
    SIMPLIFY = FALSE
  )

  superlearner_out <- lapply(seq_len(n_crisks), function(k) {
    list(
      learners_fit = full_train_list[[k]],
      meta_learner_fit = meta_learner_fits[[k]]
    )
  })

  out <- list(
    learners = learners,
    metalearner = meta_learner,
    superlearner = superlearner_out,
    cross_validation_deviance = dt_learners,
    data_info = list(
      id = id,
      status = status,
      event_time = event_time,
      nodes = sort(unique(as.numeric(levels(dt$node)))),
      nfold = nfold,
      maximum_followup = maximum_followup,
      n_crisks = n_crisks,
      learners_labels = learners_labels,
      variable_transformation = variable_transformation
    )
  )

  class(out) <- "poisson_superlearner"
  out
}

### test-glmnet-learners.R ---
#----------------------------------------------------------------------
## Author: Thomas Alexander Gerds
## Created: feb 12 2026 (06:30)
## Version:
## Last-Updated: feb 12 2026 (15:24)
##           By: Thomas Alexander Gerds
##     Update #: 18
#----------------------------------------------------------------------
##
### Commentary:
##
### Change Log:
#----------------------------------------------------------------------
##
### Code:
# tests/testthat/test-glmnet-learners.R

testthat::test_that("data_pre_processing preserves time at risk and covariates", {

  testthat::skip_if_not_installed("riskRegression")
  testthat::skip_if_not_installed("data.table")

  set.seed(42)

  Xvars <- paste0("X", 1:10)

  d <- riskRegression::sampleData(
    n = 100,
    formula = ~ f(X1, 2) + f(X2, 0) + f(X3, 0) + f(X6, 0) + f(X7, 0) +
      f(X8, 0) + f(X9, 0) + f(X10, 0)
  )
  d <- data.table::as.data.table(d)

  d[, event := as.integer(event == 1L)]
  d[, id := .I]

  id_col <- "id"
  time_col <- "time"
  status_col <- "event"

  testthat::expect_equal(nrow(d), data.table::uniqueN(d[[id_col]]))

  qs <- as.numeric(stats::quantile(d[[time_col]], probs = seq(0.1, 1, by = 0.1), na.rm = TRUE))
  nodes <- sort(unique(c(0, qs)))
  testthat::expect_true(length(nodes) >= 2)

  dt_fit <- poissonsuperlearner:::data_pre_processing(
    data = d,
    id = id_col,
    status = status_col,
    event_time = time_col,
    nodes = nodes,
    predictions = FALSE,
    uncensored_01 = FALSE
  )
  dt_fit <- data.table::as.data.table(dt_fit)

  # 1) time preservation
  time_check <- dt_fit[, .(sum_tij = sum(tij)), by = id][
    d[, .(id = get(id_col), obs_time = get(time_col))],
    on = "id"
  ]

  testthat::expect_true(all(is.finite(time_check$sum_tij)))
  testthat::expect_true(all(is.finite(time_check$obs_time)))

  testthat::expect_true(
    all(abs(time_check$sum_tij - time_check$obs_time) < 1e-8)
  )

  # 2) covariate preservation
  for (v in Xvars) {
    if (!v %in% names(d)) next

    if (!v %in% names(dt_fit)) {
      testthat::fail(paste0("Covariate '", v, "' missing from preprocessed data"))
    }

    varying <- dt_fit[, data.table::uniqueN(get(v)), by = id][V1 != 1]
    testthat::expect_equal(nrow(varying), 0)

    merged <- unique(dt_fit[, .(id, val_fit = get(v))])[
      d[, .(id = get(id_col), val_orig = get(v))],
      on = "id"
    ]

    if (is.factor(merged$val_fit) || is.character(merged$val_fit)) {
      testthat::expect_true(all(as.character(merged$val_fit) == as.character(merged$val_orig)))
    } else {
      testthat::expect_true(all(abs(merged$val_fit - merged$val_orig) < 1e-10))
    }
  }

  # 3) structure
  testthat::expect_true(all(c("tij", "deltaij", "node", "k") %in% names(dt_fit)))
  testthat::expect_true(is.factor(dt_fit$node))
  testthat::expect_true(is.factor(dt_fit$k))
})


testthat::test_that("Poisson piecewise-constant fit reproduces Cox coefficient (X1) - both Superlearner and fit_learner", {

  testthat::skip_if_not_installed("riskRegression")
  testthat::skip_if_not_installed("survival")
  testthat::skip_if_not_installed("glmnet")
  testthat::skip_if_not_installed("data.table")

  set.seed(42)

  d <- riskRegression::sampleData(n = 50, formula = ~ f(X1, 2))
  d <- data.table::as.data.table(d)

  if (!("id" %in% names(d))) d[, id := .I]
  if (!all(c("time", "event", "X1") %in% names(d))) {
    stop("sampleData did not produce expected columns: time, event, X1")
  }

  d[, status := as.integer(event == 1L)]

  fit_cox <- survival::coxph(survival::Surv(time, status) ~ X1, data = d, ties = "breslow")
  beta_cox_vec <- stats::coef(fit_cox)

  # robust name lookup
  if ("X1" %in% names(beta_cox_vec)) {
    beta_cox <- unname(beta_cox_vec[["X1"]])
  } else {
    ix <- grep("^X1", names(beta_cox_vec))
    testthat::expect_true(length(ix) >= 1)
    beta_cox <- unname(beta_cox_vec[ix[1]])
  }

  learner <- poissonsuperlearner::Learner_glmnet(
    covariates = "X1",
    cross_validation = FALSE,
    intercept = FALSE,
    add_nodes = TRUE,
    penalise_nodes = FALSE,
    lambda = 0,
    alpha = 1
  )

  olcheck <- poissonsuperlearner::Superlearner(
    data = d,
    learners = list(learner),
    id = "id",
    status = "status",
    event_time = "time",
    nodes = NULL,
    number_of_nodes = NULL
  )

  fit_ps <- poissonsuperlearner::fit_learner(
    data = d,
    learner = learner,
    id = "id",
    status = "status",
    event_time = "time",
    nodes = NULL,
    number_of_nodes = NULL
  )

  testthat::expect_true(inherits(fit_ps, "base_learner"))
  testthat::expect_true(length(fit_ps$learner_fit) >= 1)

  glmnet_fit <- fit_ps$learner_fit[[1]]
  beta_pois_mat <- as.matrix(stats::coef(glmnet_fit))

  # robust rowname lookup
  if ("X1" %in% rownames(beta_pois_mat)) {
    beta_pois <- unname(beta_pois_mat["X1", 1])
  } else {
    ix <- grep("^X1", rownames(beta_pois_mat))
    testthat::expect_true(length(ix) >= 1)
    beta_pois <- unname(beta_pois_mat[ix[1], 1])
  }

  beta_ol_mat <- as.matrix(stats::coef(olcheck$superlearner[[1]]$learners_fit))
  if ("X1" %in% rownames(beta_ol_mat)) {
    beta_ol <- unname(beta_ol_mat["X1", 1])
  } else {
    ix <- grep("^X1", rownames(beta_ol_mat))
    testthat::expect_true(length(ix) >= 1)
    beta_ol <- unname(beta_ol_mat[ix[1], 1])
  }

  testthat::expect_equal(beta_pois, beta_cox, tolerance = 1e-3)
  testthat::expect_equal(beta_ol,   beta_cox, tolerance = 1e-3)
})

#' Summarize a fitted Poisson Super Learner object
#'
#' Prints:
#' 1) a compact description of the fitted ensemble,
#' 2) cross-validated deviances for base learners (when available),
#' 3) cause-specific meta-learner coefficients (stacking weights).
#'
#' @param object `poisson_superlearner` returned by [Superlearner()].
#' @param cause `numeric(1)` or `NULL`. Which cause’s meta-learner fit to print.
#'   If `NULL`, prints one line per cause (classes only) instead of printing the full
#'   fitted objects.
#' @param model Scalar model selector. Default is `"sl"` for the stacked super learner.
#'   Other allowed values are:
#'   \describe{
#'     \item{`0` or `"sl"`}{Use the super learner prediction.}
#'     \item{learner label}{Use one stored base learner by its label in
#'       `object$data_info$learners_labels`.}
#'     \item{`"learner_j"`}{Use the `j`-th stored learner.}
#'     \item{integer `j >= 1`}{Use the `j`-th stored learner.}
#'   }
#' @param ... Passed to the underlying `coef()` method for the fitted meta-learner
#'   (learner-dependent; e.g. `s` for `glmnet`).
#'
#' @return Invisibly returns a `list` with elements:
#' \describe{
#'   \item{cross_validation_deviance}{`data.table` (or `NULL`).}
#'   \item{meta_coefficients}{List of length `n_crisks` with cause-specific coefficient objects (or `NULL`).}
#' }
#'
#' @examples
#' d <- simulateStenoT1(30, competing_risks = TRUE)
#'
#' learners <- list(
#'   lasso = Learner_glmnet(
#'     covariates = "sex",
#'     alpha = 1,
#'     lambda = 0.01,
#'     cross_validation = FALSE
#'   ),
#'   ridge = Learner_glmnet(
#'     covariates = c("sex", "value_LDL"),
#'     alpha = 0,
#'     lambda = 0.01,
#'     cross_validation = FALSE
#'   )
#' )
#'
#' fit <- Superlearner(
#'   data = d,
#'   id = "id",
#'   status = "status_cvd",
#'   event_time = "time_cvd",
#'   learners = learners,
#'   number_of_nodes = 3,
#'   nfold = 2
#' )
#'
#' s <- summary(fit, cause = 1)
#' names(s)
#'
#'
#' @export
summary.poisson_superlearner <- function(object,
                                         cause = NULL,
                                         model = "sl",
                                         ...) {

  if (is.null(object$superlearner) || length(object$superlearner) == 0L) {
    cat("No fitted model available (superlearner is NULL).\n")
    return(invisible(object))
  }

  model_sel <- resolve_prediction_model(object, model)
  labels <- psl_get_labels(object)

  ## If the user asks for a stored base learner, or if there is no meta-learner
  ## (single-learner special case), dispatch directly to the underlying fit.
  if (model_sel$type == "learner" || is.null(object$metalearner)) {

    if (is.null(cause)) {
      out <- lapply(seq_len(object$data_info$n_crisks), function(k) {
        fit_info <- psl_get_stored_fit(object, cause = k, model = model)
        summary(fit_info$fit, ...)
      })
      names(out) <- paste0("cause_", seq_len(object$data_info$n_crisks))
      return(out)
    }

    fit_info <- psl_get_stored_fit(object, cause = cause, model = model)
    return(summary(fit_info$fit, ...))
  }

  ## Default: summarize the stacked superlearner
  zmap <- stats::setNames(labels, paste0("Z", seq_along(labels)))

  cat("Call:\n")
  if (!is.null(object$metalearner)) {
    ml <- object$metalearner
    ml_class <- class(ml)[1]
    cat(
      sprintf(
        "  Superlearner(..., learners = c(%s), metalearner = %s)\n",
        paste(labels, collapse = ", "),
        ml_class
      )
    )
  } else {
    cat("  Superlearner(...)\n")
  }

  cat("\nFitted object:\n")
  cat("  Class: poisson_superlearner\n")
  cat("  Number of competing risks:", object$data_info$n_crisks, "\n")
  cat("  Number of learners:", length(object$learners), "\n")
  cat("  Learners:", paste(labels, collapse = ", "), "\n")
  cat("  Number of folds:", object$data_info$nfold, "\n")
  cat("  Maximum follow-up:", object$data_info$maximum_followup, "\n")
  cat("  Number of nodes:", length(object$data_info$nodes), "\n")

  cat("\nCross-validation deviance (Average across V-Folds):\n")
  if (!is.null(object$cross_validation_deviance)) {
    print(object$cross_validation_deviance)
  } else if (!is.null(object$meta_learner_cross_validation)) {
    print(object$meta_learner_cross_validation)
  } else {
    cat("  <not available>\n")
  }

  cat("\nMeta-learner coefficients:\n")

  causes_to_show <- if (is.null(cause)) {
    seq_along(object$superlearner)
  } else {
    cause
  }

  meta_out <- vector("list", length(causes_to_show))
  names(meta_out) <- paste0("cause_", causes_to_show)

  for (ii in seq_along(causes_to_show)) {
    k <- causes_to_show[ii]
    fit_k <- object$superlearner[[k]]$meta_learner_fit

    coefs <- tryCatch(
      psl_extract_meta_coefs(fit_k, ...),
      error = function(e) NULL
    )

    if (is.null(coefs)) {
      cat("  k = ", k, ": <not available>\n", sep = "")
      meta_out[[ii]] <- NULL
      next
    }

    nms <- names(coefs)
    names(coefs) <- psl_rename_z_in_text(nms, zmap)

    cat("  k = ", k, ":\n", sep = "")
    print(coefs)
    cat("\n")

    meta_out[[ii]] <- coefs
  }

  invisible(list(
    cross_validation_deviance = object$cross_validation_deviance,
    meta_coefficients = meta_out
  ))
}


#' Summarize a fitted base learner object
#'
#' Dispatches to the underlying fitted model’s `summary()` method for the selected
#' cause, or returns a list of summaries for all causes.
#'
#' @param object `base_learner` returned by [fit_learner()].
#' @param cause `numeric(1)` or `NULL`. Which cause to summarize. If `NULL`,
#'   returns one summary per cause.
#' @param ... Passed to the underlying `summary()` method (learner-dependent).
#'
#' @return If `cause` is a single integer, returns the underlying model summary for
#' that cause. If `cause = NULL`, returns a list of summaries (one per cause).
#'
#' @examples
#' d <- simulateStenoT1(30, competing_risks = TRUE)
#' lrn <- Learner_glmnet(
#'   covariates = c("sex", "value_LDL"),
#'   lambda = 0.01,
#'   cross_validation = FALSE
#' )
#' bl <- fit_learner(
#'   d,
#'   learner = lrn,
#'   id = "id",
#'   status = "status_cvd",
#'   event_time = "time_cvd",
#'   number_of_nodes = 3
#' )
#' out <- summary(bl, cause = 1)
#'
#'
#' @export
summary.base_learner <- function(object, cause=1, ...) {

  if (is.null(object$learner_fit)) {
    cat("No fitted model available (learner_fit is NULL).\n")
    return(invisible(object))
  }

  if (is.null(cause)) {
    return(lapply(object$learner_fit, summary, ...))
  } else{

    return(summary(object$learner_fit[[cause]], ...))

  }

}

utils::globalVariables(c(
  ".", "..columns_ss", "..z_covariates",
  "LDL_hinge3_sq", "age_hinge55_sq",
  "absolute_risk", "covariate", "deviance", "internal_psl_ix",
  "deltaij", "deltatime", "dummy",
  "event", "folder", "id", "k", "node",
  "pwc_cols", "status_cvd", "survival_function",
  "tij", "time", "time.event.0", "time.event.1", "time.event.2",
  "time_cvd", "time_cvd",
  "uncensored_event", "uncensored_status_cvd", "uncensored_time", "uncensored_time_cvd",
  "value_AlbuminuriaMacro", "value_AlbuminuriaMicro",
  "vec_dt", "vec_dt2",
  "log2_eGFR_old", "log2_eGFR_young","row_ix","competing_risk"
))

#' Fit a single base learner
#'
#' Pre-processes subject-level time-to-event data into a long Poisson format on a
#' piecewise-constant time grid, then fits **one** initialized learner object.
#' For competing risks, a separate model is fit for each event type (cause)
#' using the standard cause-specific Poisson likelihood on the long data.
#'
#' @param data `data.frame`. Subject-level input data (one row per subject).
#' @param learner Reference-class learner object (e.g. from [Learner_glmnet()],
#'   [Learner_hal()] or [Learner_gam()]). Must implement a `$private_fit(dt_long)`
#'   method that fits the learner on long Poisson data for one cause.
#' @param id `character(1)`. Name of the subject identifier column. If not found
#'   in `data`, an `id` column is created automatically.
#' @param stratified_k_fold `logical(1)`. Reserved argument for future fold strategy.
#'   Currently ignored.
#' @param status `character(1)`. Name of the event-status column.
#'   Must be coded with `0` = censoring and `1,2,...,K` for event types (causes).
#'   If there is no `0` in `status`, the data are treated as uncensored.
#' @param event_time `character(1)`. Name of the event/censoring time column.
#'   Must be present in `data`.
#' @param number_of_nodes `numeric(1)` or `NULL`. If not `NULL`, constructs a
#'   quantile-based node grid with `number_of_nodes + 1` cut points (including
#'   endpoints), then adds `0` if missing.
#' @param nodes `numeric` or `NULL`. Explicit time-node grid (cut points). If
#'   supplied, `number_of_nodes` is ignored. `0` is added if missing. Nodes
#'   beyond `max(event_time)` are dropped.
#' @param variable_transformation `list`/`character`/`formula` or `NULL`.
#'   Optional transformations applied to the internally created long Poisson data
#'   before fitting (via `apply_transformations()`).
#' @param ... Additional arguments currently ignored.
#'
#' @return An object of class `base_learner`, i.e. a named `list` with:
#'
#' \describe{
#' \item{model}{The **learner object** that was fit (the input `learner`), stored
#'   for later prediction. This contains the learner specification (e.g.,
#'   covariates, tuning parameters).}
#'
#' \item{learner_fit}{A `list` of fitted model objects, **one per cause**.
#'   Its length equals `data_info$n_crisks`. The list is created by splitting the
#'   internally pre-processed long data by cause indicator `k` and calling
#'   `model$private_fit()` on each split.
#'
#'   \itemize{
#'     \item Names typically correspond to the cause labels `"1"`, `"2"`, ..., `"K"`.
#'     \item Each element is **learner-dependent**: e.g. for `Learner_glmnet` it
#'       may be a `"glmnet"` (often wrapped, e.g. `"fishnet"`) fit; for other
#'       learners it will be whatever `$private_fit()` returns.
#'     \item Each fitted object is trained on long Poisson data representing the
#'       piecewise-constant hazard for that cause across the node intervals.
#'   }}
#'
#' \item{data_info}{A `list` of bookkeeping information needed for prediction and
#'   interpretation:
#'   \describe{
#'     \item{id}{Identifier column name used.}
#'     \item{status}{Status column name used.}
#'     \item{event_time}{Event/censoring time column name used.}
#'     \item{nodes}{Numeric vector of node cut points used for the piecewise grid
#'       (includes `0` and is sorted). These are the interval boundaries used in
#'       the long Poisson representation.}
#'     \item{maximum_followup}{`max(data[[event_time]])`.}
#'     \item{n_crisks}{Number of event types (causes) detected.
#'       If censoring is present (`0` in `status`), then `n_crisks = #unique(status) - 1`;
#'       otherwise `n_crisks = #unique(status)`.}
#'     \item{variable_transformation}{The transformation specification passed in
#'       `variable_transformation` (or `NULL`).}
#'   }}
#' }
#'
#' @examples
#' d <- simulateStenoT1(50, competing_risks = TRUE)
#' lrn <- Learner_glmnet(covariates = c("age", "value_LDL"),
#'                       lambda = 0,
#'                       cross_validation = FALSE)
#' bl <- fit_learner(d,
#'                   learner = lrn,
#'                   id = "id",
#'                   status = "status_cvd",
#'                   event_time = "time_cvd",
#'                   number_of_nodes = 4)
#'
#' @export
fit_learner <- function(data,
                    learner,
                    id = "id",
                    stratified_k_fold = FALSE,
                    status = "status",
                    event_time = NULL,
                    number_of_nodes = NULL,
                    nodes = NULL,
                    variable_transformation = NULL,
                    ...) {
  # Multiple checks about the input ----


  if (!(id %in% names(data))) {
    data[["id"]] <- 1:NROW(data)
    id <- "id"
  }


  maximum_followup = max(data[[event_time]])


  n <- length(unique(data[[id]]))

  if (!(0 %in% data[[status]])) {
    warning(
      paste0(
        "There is no value of ",
        status,
        " equal to zero. We will consider the data uncensored."
      )
    )
    n_crisks <- length(unique(data[[status]]))
    uncensored_01 <- TRUE

  } else{
    n_crisks <- length(unique(data[[status]])) - 1
    uncensored_01 <- FALSE
  }


    #  Handle nodes
    ##Either the nodes are given or we take all of the realised times
    if (!is.null(number_of_nodes)) {

      grid_nodes = quantile(
        data[[event_time]],
        probs = seq(0, 1, length.out = as.integer(number_of_nodes) + 1),
        type = 1,
        names = FALSE
      )



    } else{
      if (is.null(nodes)) {
        grid_nodes <- sort(unique(data[[event_time]]))

      } else{
        grid_nodes <- nodes

      }
    }

    # Add zero if missing
    if (!(0 %in% grid_nodes)) {
      grid_nodes <- c(0, grid_nodes)

    }



    grid_nodes <- grid_nodes[grid_nodes <= max(data[[event_time]])]

    # Actual data pp
    dt <- data_pre_processing(
      data = data,
      id = id,
      status = status,
      nodes = grid_nodes,
      event_time = event_time,
      uncensored_01 = uncensored_01
    )






  lhs_string = NULL

  if (!is.null(variable_transformation)) {
    apply_transformations(dt, variable_transformation)
  }

  training_data <- split(dt, by = "k")

  model_fit <- mapply(function(x) {
    out <- learner$private_fit(x)
    return(out)
  }, training_data, SIMPLIFY = FALSE)


  out <- list(
    model=learner,
    learner_fit=model_fit,
    data_info = list(
      id = id,
      status = status,
      event_time = event_time,
      nodes = sort(unique(as.numeric(levels(dt$node)))),
      maximum_followup = maximum_followup,
      n_crisks=n_crisks,
      variable_transformation = variable_transformation    )
  )

  class(out) <- "base_learner"

  return(out)
}

#' Predict hazards, survival and absolute risk from a fitted Poisson Super Learner
#'
#' Computes **cause-specific piecewise-constant hazards** (`pwch_k`), the corresponding
#' **survival function**, and **absolute risk** for a given cause, at user-supplied
#' prediction horizons `times`, for each row in `newdata`.
#'
#' Internally, `newdata` is expanded to a Cartesian product with the requested
#' `times`, converted to long Poisson format on `object$data_info$nodes`, and hazards
#' are predicted either from the stacked super learner (`model = "sl"`) or from one
#' selected fitted base learner. Survival and absolute risk are then computed from
#' the predicted hazards.
#'
#' @param object `poisson_superlearner`. A fitted ensemble from [Superlearner()].
#' @param newdata `data.frame`/`data.table`. New covariate data (one row per subject).
#'   If `newdata` contains the original `event_time`, `status`, or `id` columns used
#'   for fitting, they are ignored for prediction.
#' @param times `numeric`. Prediction horizon(s). May include `0`.
#'   Times larger than `object$data_info$maximum_followup` are not supported:
#'   if **all** requested times exceed the maximum follow-up, a warning is issued and
#'   `NULL` is returned; if only **some** exceed, output rows for those times are
#'   returned with `NA` predictions.
#' @param cause `numeric(1)`. Cause index (1, 2, ...) used for the `absolute_risk`
#'   calculation.
#' @param model Scalar model selector. Default is `"sl"` for the stacked super learner.
#'   Other allowed values are:
#'   \describe{
#'     \item{`0` or `"sl"`}{Use the super learner prediction.}
#'     \item{learner label}{Use one stored base learner by its label in
#'       `object$data_info$learners_labels`.}
#'     \item{`"learner_j"`}{Use the `j`-th stored learner.}
#'     \item{integer `j >= 1`}{Use the `j`-th stored learner.}
#'   }
#'   Numeric positions refer to the learners actually stored in the fitted object.
#' @param ... Additional arguments (currently ignored).
#'
#' @details
#' **Special case `times = 0`:** when `0` is included in `times`, the returned rows
#' have `survival_function = 1`, `absolute_risk = 0`, and all `pwch_k = 0` at time 0.
#'
#' **Identifiers in the output:** if `newdata` contains the `id` column, it is carried
#' into the output. If `newdata` does not contain an id column, an internal id is
#' created for computation, but it is not guaranteed to appear in the returned table
#' unless it was present in `newdata`.
#'
#' @return A `data.table` with one row per `(row in newdata, time in times)` and columns:
#' \describe{
#'   \item{(original columns)}{All columns from `newdata` (excluding ignored event columns).}
#'   \item{time column}{A column with name `object$data_info$event_time` holding the requested horizon.}
#'   \item{pwch_1, pwch_2, ...}{Predicted cause-specific piecewise hazards at the horizon.}
#'   \item{survival_function}{Predicted survival probability at the horizon.}
#'   \item{absolute_risk}{Predicted cumulative incidence (absolute risk) for `cause` at the horizon.}
#' }
#'
#' @examples
#' d <- simulateStenoT1(30, competing_risks = TRUE)
#'
#' learners <- list(
#'   lasso = Learner_glmnet(
#'     covariates = "sex",
#'     alpha = 1,
#'     lambda = 0.01,
#'     cross_validation = FALSE
#'   ),
#'   ridge = Learner_glmnet(
#'     covariates = c("sex", "value_LDL"),
#'     alpha = 0,
#'     lambda = 0.01,
#'     cross_validation = FALSE
#'   )
#' )
#'
#' fit <- Superlearner(
#'   data = d,
#'   id = "id",
#'   status = "status_cvd",
#'   event_time = "time_cvd",
#'   learners = learners,
#'   number_of_nodes = 3,
#'   nfold = 2
#' )
#' p <- predict(fit, newdata = d[1:3], times = c(0, 2), cause = 1)
#' p[, .(id, time_cvd, absolute_risk)]
#'
#'
#' @export
predict.poisson_superlearner <- function(object,
                                         newdata,
                                         times,
                                         cause = 1,
                                         model = "sl",
                                         ...) {

  model_sel <- resolve_prediction_model(object, model)

  setDT(newdata)
  tmp <- copy(newdata)
  tmp[, internal_psl_ix := 1:.N]

  tmp <- tmp[, setdiff(names(tmp), c(object$data_info$event_time,
                                     object$data_info$status,
                                     object$data_info$id)), with = FALSE]

  cond_zero <- 0 %in% times
  all_zero <- all(times == 0)
  cond_times_larger_than_max <- times > object$data_info$maximum_followup

  pwch_cols <- paste0("pwch_", 1:object$data_info$n_crisks)

  if (all(cond_times_larger_than_max)) {
    warning(
      paste0(
        "All the entries in the input times are larger than the maximum follow-up: ",
        as.character(object$data_info$maximum_followup)
      )
    )
    d <- NULL
    return(d)
  } else {
    eval(parse(
      text = paste0(
        "
        vec_dt <- data.table(
          ",
        object$data_info$event_time,
        " = times[times <= object$data_info$maximum_followup]
        )
        "
      )
    ))

    tmp[, dummy := 1]
    vec_dt[, dummy := 1]
    data_pp <- merge(tmp, vec_dt, by = "dummy", allow.cartesian = TRUE)[, dummy := NULL]
  }

  if (is.null(data_pp[[object$data_info$id]])) {
    data_pp[[object$data_info$id]] <- 1:nrow(data_pp)
  }

  if (is.null(data_pp[[object$data_info$status]])) {
    data_pp[[object$data_info$status]] <- 0
  }

  if (cond_zero) {
    tmptcol <- object$data_info$event_time
    zero_time <- data_pp[get(tmptcol) == 0]
    for (cl in pwch_cols) set(zero_time, j = cl, value = 0)
    set(zero_time, j = "survival_function", value = 1)
    set(zero_time, j = "absolute_risk", value = 0)
    data_pp <- data_pp[get(tmptcol) != 0]
  }

  if (all_zero) {
    return(zero_time)
  }

  data_pp <- data_pre_processing(
    data_pp,
    id = object$data_info$id,
    status = object$data_info$status,
    predictions = TRUE,
    event_time = object$data_info$event_time,
    nodes = object$data_info$nodes
  )

  if (!is.null(object$data_info$variable_transformation)) {
    apply_transformations(data_pp, object$data_info$variable_transformation)
  }

  z_covariates <- paste0("Z", seq_along(object$learners))

  data_pp[, deltatime := tij][, tij := 1]

  use_superlearner <- isTRUE(model_sel$type == "sl") &&
    length(object$learners) > 1L &&
    !is.null(object$metalearner)

  if (use_superlearner) {

    learners_predictions <- mapply(
      function(crisk_cause,
               superlearner,
               newdata,
               learners)
        learners_hat(crisk_cause, superlearner, newdata, learners),
      crisk_cause = as.list(1:object$data_info$n_crisks),
      superlearner = object$superlearner,
      MoreArgs = list(newdata = data_pp, learners = object$learners),
      SIMPLIFY = FALSE
    )

    pseudo_observations_data <- lapply(learners_predictions, function(mx) {
      out <- matrix(
        apply(as.matrix(
          mx,
          nrow = nrow(newdata),
          ncol = length(z_covariates)
        ), MARGIN = 2, log),
        nrow = nrow(data_pp),
        ncol = length(z_covariates)
      )
      colnames(out) <- z_covariates
      as.data.table(as.data.frame.matrix(out))
    })

    dt_pred <- mapply(
      function(superlearner, pseudo_observations_data) {
        object$metalearner$private_predictor(
          model = superlearner$meta_learner_fit,
          newdata = cbind(pseudo_observations_data, data_pp)
        )
      },
      object$superlearner,
      pseudo_observations_data,
      SIMPLIFY = FALSE
    )

  } else {

    learner_index <- if (model_sel$type == "learner") {
      model_sel$index
    } else {
      1L
    }

    dt_pred <- lapply(seq_len(object$data_info$n_crisks), function(k) {
      fit_k <- if (length(object$learners) == 1L) {
        object$superlearner[[k]]$learners_fit
      } else {
        object$superlearner[[k]]$learners_fit[[learner_index]]
      }

      object$learners[[learner_index]]$private_predictor(
        model = fit_k,
        newdata = data_pp
      )
    })
  }

  data_pp[, paste0("pwch_", 1:object$data_info$n_crisks) := dt_pred]

  sum_of_hazards <- paste(pwch_cols, collapse = " + ")
  pwch_dot_string <- paste0("data_pp[, pwch_dot :=", sum_of_hazards, "]")
  eval(parse(text = pwch_dot_string))

  mapply(function(pwch, name) {
    data_pp[, (paste0("cumulative_hazard_", name)) := cumsum(get(pwch) * deltatime), by = id]
  }, pwch_cols, gsub("pwch_", "", pwch_cols))

  haz <- as.matrix(data_pp[, .SD, .SDcols = patterns("^pwch_[0-9]+$")])
  S <- pch_survival(id = data_pp$id, dt = data_pp$deltatime, haz = haz)
  data_pp[, survival_function := S]

  data_pp[, absolute_risk := pch_absolute_risk(id, deltatime, haz, cause_idx = cause)]

  data_pp <- data_pp[, .SD[.N], by = id][, times := as.numeric(as.character(node)) + deltatime]

  columns_ss <- unique(
    c(
      colnames(newdata),
      object$data_info$event_time,
      pwch_cols,
      "survival_function",
      "absolute_risk",
      "internal_psl_ix"
    )
  )

  d <- data_pp[, ..columns_ss]

  if (cond_zero) {
    d <- rbind(zero_time, d)
  }

  if (any(cond_times_larger_than_max)) {
    eval(parse(
      text = paste0(
        "
        vec_dt2 <- data.table(
          ",
        object$data_info$event_time,
        " = times[cond_times_larger_than_max]
        )
        "
      )
    ))

    tmp[, dummy := 1]
    vec_dt2[, dummy := 1]

    d2 <- merge(tmp, vec_dt2, by = "dummy", allow.cartesian = TRUE)[, dummy := NULL]
    d2[, c(pwch_cols, "survival_function", "absolute_risk") := NA_real_]

    if (object$data_info$id %in% colnames(d)) {
      d2[[object$data_info$id]] <- (nrow(data_pp) + 1):(nrow(data_pp) + nrow(d2))
    }

    d <- rbind(d, d2, fill = TRUE)
  }

  d[, (object$data_info$id) := NULL]
  setnames(d, new = object$data_info$id, old = "internal_psl_ix")
  d <- d[order(get(object$data_info$id))]
  return(d)
}



#' Predict hazards, survival and absolute risk from a fitted base learner
#'
#' Computes **cause-specific piecewise-constant hazards** (`pwch_k`), the corresponding
#' **survival function**, and **absolute risk** for a given cause, at user-supplied
#' prediction horizons `times`, using a fitted `base_learner` object (single learner;
#' no stacking).
#'
#' Internally, `newdata` is expanded to a Cartesian product with `times`, converted to
#' long Poisson format on `object$data_info$nodes`, and the fitted learner for each
#' cause in `object$learner_fit` is used to predict the cause-specific hazards.
#' Survival and absolute risk are then computed from the predicted hazards.
#'
#' @param object `base_learner`. A fitted object returned by [fit_learner()].
#'   It contains the learner specification in `object$model` and cause-specific fitted
#'   models in `object$learner_fit`.
#' @param newdata `data.frame`/`data.table`. New covariate data (one row per subject).
#'   If `newdata` contains the original `event_time`, `status`, or `id` columns used
#'   for fitting, they are ignored for prediction.
#' @param times `numeric`. Prediction horizon(s). May include `0`.
#'   Times larger than `object$data_info$maximum_followup` are not supported:
#'   if **all** requested times exceed the maximum follow-up, a warning is issued and
#'   `NULL` is returned; if only **some** exceed, output rows for those times are
#'   returned with `NA` predictions.
#' @param cause `numeric(1)`. Cause index (1, 2, ...) used for the `absolute_risk`
#'   calculation.
#' @param ... Additional arguments (currently ignored).
#'
#' @details
#' **Special case `times = 0`:** when `0` is included in `times`, the returned rows
#' have `survival_function = 1`, `absolute_risk = 0`, and all `pwch_k = 0` at time 0.
#'
#' **Identifiers in the output:** if `newdata` contains the `id` column, it is carried
#' into the output. If `newdata` does not contain an id column, an internal id is
#' created for computation, but it is not guaranteed to appear in the returned table
#' unless it was present in `newdata`.
#'
#' @return A `data.table` with one row per `(row in newdata, time in times)` and columns:
#' \describe{
#'   \item{(original columns)}{All columns from `newdata` (excluding ignored event columns).}
#'   \item{time column}{A column with name `object$data_info$event_time` holding the requested horizon.}
#'   \item{pwch_1, pwch_2, ...}{Predicted cause-specific piecewise hazards at the horizon.}
#'   \item{survival_function}{Predicted survival probability at the horizon.}
#'   \item{absolute_risk}{Predicted cumulative incidence (absolute risk) for `cause` at the horizon.}
#' }
#'
#' @examples
#' d <- simulateStenoT1(120, competing_risks = TRUE)
#' lrn <- Learner_glmnet(covariates = c("age", "value_LDL"), lambda = 0, cross_validation = FALSE)
#' bl <- fit_learner(d, learner = lrn, id="id", status="status_cvd", event_time="time_cvd",
#'                   number_of_nodes=8)
#' p <- predict(bl, newdata = d[1:5], times = c(0, 2, 5), cause = 1)
#' head(p)
#'
#' @export
predict.base_learner <- function(object,
                                         newdata,
                                         times,
                                         cause = 1,
                                         ...) {

  setDT(newdata)

  tmp <- copy(newdata)
  tmp[,internal_psl_ix:=1:.N]
  # here we disregard the event_time column if present in the newdata
  tmp <- tmp[, setdiff(names(tmp), c(object$data_info$event_time,
                                     object$data_info$status,
                                     object$data_info$id)), with = FALSE]


  ## checks on the data
  cond_zero <- 0 %in% times

  all_zero <- all(times==0)

  cond_times_larger_than_max <- times > object$data_info$maximum_followup

  ## frame hazard problem
  pwch_cols <- paste0("pwch_", 1:object$data_info$n_crisks)

  if (all(cond_times_larger_than_max)) {
    warning(
      paste0(
        "All the entries in the input times are larger than the maximum follow-up: ",
        as.character(object$data_info$maximum_followup)
      )
    )
    d <- NULL

    return(d)

  } else{
    eval(parse(
      text = paste0(
        "
    vec_dt <- data.table(

    ",
        object$data_info$event_time,
        " = times[times <= object$data_info$maximum_followup]
  )
    "
      )
    ))

    # # no problem writing over id
    # if (is.null(tmp[[object$data_info$id]])) {
    #   tmp[[object$data_info$id]] <- 1:nrow(tmp)
    # }
    #
    # if (is.null(tmp[[object$data_info$status]])) {
    #   tmp[[object$data_info$status]] <- 0
    # }

    tmp[, dummy := 1]
    vec_dt[, dummy := 1]

    # Merge on dummy to create Cartesian product
    data_pp <- merge(tmp, vec_dt, by = "dummy", allow.cartesian = TRUE)[, dummy := NULL]
  }
  # }




# no problem writing over id
if (is.null(data_pp[[object$data_info$id]])) {
  data_pp[[object$data_info$id]] <- 1:nrow(data_pp)
}

if (is.null(data_pp[[object$data_info$status]])) {
  data_pp[[object$data_info$status]] <- 0
}


  if(cond_zero){


    tmptcol <- object$data_info$event_time  # character vector
    zero_time <- data_pp[get(tmptcol) == 0]
    for (cl in pwch_cols) set(zero_time, j = cl, value = 0)
    set(zero_time, j = "survival_function", value = 1)
    set(zero_time, j = "absolute_risk", value = 0)
    data_pp<-data_pp[get(tmptcol) != 0]
  }

  if(all_zero){

    return(zero_time)
  }



  data_pp <- data_pre_processing(
    data_pp,
    id = object$data_info$id,
    status = object$data_info$status,
    predictions=TRUE,
    event_time = object$data_info$event_time,
    nodes = object$data_info$nodes
  )



  #  ()
  if (!is.null(object$data_info$variable_transformation)) {

    apply_transformations(data_pp, object$data_info$variable_transformation)
  }


  # Set covariates for metalearner
  z_covariates <- paste0("Z", 1:length(object$learners))


  # Predict on the validation set your pseudo-observations ----
  #

  data_pp[, deltatime := tij][, tij := 1]


    dt_pred <- lapply(object$learner_fit, function(x) {
      object$model$private_predictor(
        model = x, newdata = data_pp)
    })


  # save casue-specific pwch

  data_pp[, paste0("pwch_", 1:object$data_info$n_crisks) := dt_pred]



  # save sum of pwch

  sum_of_hazards <- paste(pwch_cols, collapse = " + ")

  pwch_dot_string <- paste0("data_pp[, pwch_dot :=", sum_of_hazards, "]")

  eval(parse(text = pwch_dot_string))


  # compute cumulative hazard

  mapply(function(pwch, name) {
    data_pp[, (paste0("cumulative_hazard_", name)) := cumsum(get(pwch) * deltatime), by = id]
  }, pwch_cols, gsub("pwch_", "", pwch_cols))


  # compute survival function

  ## c++
  haz <- as.matrix(data_pp[, .SD, .SDcols = patterns("^pwch_[0-9]+$")])
  S <- pch_survival(id = data_pp$id, dt = data_pp$deltatime, haz = haz)
  data_pp[, survival_function := S]


  data_pp[, absolute_risk := pch_absolute_risk(id, deltatime, haz, cause_idx = cause)]

  # abs_risk approx
  # data_pp[, survival_function_shift := shift(survival_function, fill = 1), by =
  #           id]
  # absolute_risk_string <- paste0(
  #   "data_pp[, absolute_risk_2 := cumsum(survival_function_shift * pwch_",
  #   cause,
  #   "*deltatime), by = id]"
  # )
  # eval(parse(text = absolute_risk_string))


  ## non c++
  # hazard_terms <- paste0("cumulative_hazard_", 1:object$data_info$n_crisks)
  # sum_expr <- paste(pwch_cols, collapse = " + ")
  # survival_function_string <- paste0("data_pp[, survival_function := exp(-cumsum((", sum_expr, ")*deltatime)),by=id]")
  # eval(parse(text = survival_function_string))

  # shift survival function
  # data_pp[, survival_function_shift := shift(survival_function, fill = 1), by =
  #           id]
  # absolute_risk_string <- paste0(
  #   "data_pp[, absolute_risk := cumsum(survival_function_shift * pwch_",
  #   cause,
  #   "/pwch_dot * (1-exp(-pwch_dot*deltatime))), by = id]"
  # )
  # eval(parse(text = absolute_risk_string))

  ####
  data_pp <- data_pp[, .SD[.N], by = id][, times := as.numeric(as.character(node)) +
                                           deltatime]



  columns_ss <- unique(
    c(
      colnames(newdata),
      object$data_info$event_time,
      pwch_cols,
      "survival_function",
      "absolute_risk",
      "internal_psl_ix"
    )
  )

  d <- data_pp[, ..columns_ss]


  if (cond_zero) {

    d<- rbind(zero_time,
              d)

  }



  if (any(cond_times_larger_than_max)) {
    eval(parse(
      text = paste0(
        "
    vec_dt2 <- data.table(

    ",
        object$data_info$event_time,
        " = times[cond_times_larger_than_max]
  )
    "
      )
    ))

    tmp[, dummy := 1]
    vec_dt2[, dummy := 1]

    d2 <- merge(tmp, vec_dt2, by = "dummy", allow.cartesian = TRUE)[, dummy := NULL]
    d2[, c(pwch_cols, "survival_function", "absolute_risk") := NA_real_]


    if (object$data_info$id %in% colnames(d)) {
      d2[[object$data_info$id]] <- (nrow(data_pp) + 1):(nrow(data_pp) + nrow(d2))
    }



    d <- rbind(d, d2, fill = TRUE)


  }


  d[, (object$data_info$id) := NULL]
  setnames(d, new = object$data_info$id, old = "internal_psl_ix")
  d <- d[order(get(object$data_info$id))]
  return(d)




}

#' Print method for `base_learner`
#'
#' Prints a compact description of the fitted base learner, including the learner
#' type, the time-grid used, and (optionally) the fitted model object for a given
#' cause.
#'
#' @param x `base_learner` object returned by [fit_learner()].
#' @param cause `numeric(1)` or `NULL`. Which cause to print the fitted model for.
#'   If `NULL`, prints one line per cause (classes only) instead of printing the full
#'   fitted objects.
#' @param ... Passed to the underlying fitted object `print()` method when `cause`
#'   is a single integer.
#'
#' @return Invisibly returns `x`.
#'
#' @examples
#' d <- simulateStenoT1(30, competing_risks = TRUE)
#' lrn <- Learner_glmnet(
#'   covariates = c("sex", "value_LDL"),
#'   lambda = 0.01,
#'   cross_validation = FALSE
#' )
#' bl <- fit_learner(
#'   d,
#'   learner = lrn,
#'   id = "id",
#'   status = "status_cvd",
#'   event_time = "time_cvd",
#'   number_of_nodes = 3
#' )
#' print(bl, cause = NULL)
#'
#'
#' @export
print.base_learner <- function(x, cause=1, ...) {

  if (is.null(x$learner_fit)) {
    cat("No fitted model available (learner_fit is NULL).\n")
    return(invisible(x))
  }

  if (is.null(cause)) {
    return(invisible(lapply(x$learner_fit, print, ...)))
  }

  return(print(x$learner_fit[[cause]], ...))
}


#' Print method for `poisson_superlearner`
#'
#' Prints a compact description of the fitted Poisson Super Learner, including the
#' number of base learners, the meta-learner, the time-grid used, and competing-risk
#' structure. Optionally prints the fitted meta-learner for a given cause.
#'
#' @param x `poisson_superlearner` object returned by [Superlearner()].
#' @param cause `numeric(1)` or `NULL`. Which cause’s meta-learner fit to print.
#'   If `NULL`, prints one line per cause (classes only) instead of printing the full
#'   fitted objects.
#' @param model Scalar model selector. Default is `"sl"` for the stacked super learner.
#'   Other allowed values are:
#'   \describe{
#'     \item{`0` or `"sl"`}{Use the super learner prediction.}
#'     \item{learner label}{Use one stored base learner by its label in
#'       `object$data_info$learners_labels`.}
#'     \item{`"learner_j"`}{Use the `j`-th stored learner.}
#'     \item{integer `j >= 1`}{Use the `j`-th stored learner.}
#'   }
#' @param ... Passed to the underlying fitted meta-learner `print()` method when
#'   `cause` is a single integer.
#'
#' @return Invisibly returns `x`.
#'
#' @examples
#' d <- simulateStenoT1(30, competing_risks = TRUE)
#'
#' learners <- list(
#'   lasso = Learner_glmnet(
#'     covariates = "sex",
#'     alpha = 1,
#'     lambda = 0.01,
#'     cross_validation = FALSE
#'   ),
#'   ridge = Learner_glmnet(
#'     covariates = c("sex", "value_LDL"),
#'     alpha = 0,
#'     lambda = 0.01,
#'     cross_validation = FALSE
#'   )
#' )
#'
#' fit <- Superlearner(
#'   data = d,
#'   id = "id",
#'   status = "status_cvd",
#'   event_time = "time_cvd",
#'   learners = learners,
#'   number_of_nodes = 3,
#'   nfold = 2
#' )
#'
#' print(fit, cause = NULL)
#'
#' @export
print.poisson_superlearner <- function(x, cause = 1, model = "sl", ...) {

  if (is.null(x$superlearner)) {
    cat("No fitted model available (superlearner is NULL).\n")
    return(invisible(x))
  }

  if (is.null(cause)) {
    invisible(lapply(seq_len(x$data_info$n_crisks), function(k) {
      fit_info <- psl_get_stored_fit(x, cause = k, model = model)
      print(fit_info$fit, ...)
    }))
    return(invisible(x))
  }

  fit_info <- psl_get_stored_fit(x, cause = cause, model = model)
  print(fit_info$fit, ...)
  invisible(x)
}

# Data preprocessing ----

create_response_variable_c_risks <- function(nodes, time_to_event, delta, event_type){


  p_holder <- ifelse(delta == event_type, 1, 0)

  l <- sum(nodes < time_to_event)

  out <- c(rep(0, max(0, l - 1)),
           p_holder)

  return(out)
}

create_offset_variable <- function(nodes, delta, time_to_event){

  # if(time_to_event==0){
  #
  #   return(cbind(0,0))
  #
  # }else{
  tmp <- c(nodes[nodes < time_to_event],
           first(nodes[nodes >= time_to_event]))



  if (all(nodes < time_to_event)) {
    tmp <- c(tmp, time_to_event)
  } else{
    tmp[length(tmp)] <- time_to_event
  }


  tij <- diff(c(tmp))

  grid_nodes <- c(nodes[nodes < time_to_event])

  return(cbind(grid_nodes,tij))


  # }
}

data_pre_processing <- function(data,
                                id,
                                status,
                                event_time,
                                nodes=NULL,
                                predictions=FALSE,
                                uncensored_01=FALSE
){


  setDT(data)

  # Handle competing risks ----
  ## for each of the competing risks (CR) we need to create a table
  n_crisks <- pmax(length(unique(data[[status]])) - 1+uncensored_01,1)
  ## the CR tables are stuck on top of each other to allow for possible interactions
  dt_fit <- do.call(rbind, replicate(n_crisks, data, simplify = FALSE))
  ## we create an artificial k index. Table specific.
  dt_fit <- dt_fit[, k := rep(1:n_crisks, each = dim(data)[1])]

    # Data Transformation ----
    tmp <- c(id, "k")
    dt_fit <- dt_fit[ , {
        tte <- .SD[[1L]]   # time-to-event vector
        del <- .SD[[2L]]   # status/delta vector
        off <- create_offset_variable(nodes, time_to_event = tte)
        .(node    = off[, 1L],
          tij     = off[, 2L],
          deltaij = create_response_variable_c_risks(nodes,time_to_event = tte,delta = del,event_type = k)
          )
    }, by = c(id, "k"),
    .SDcols = c(event_time, status)]

    ## Retrieve covariates

  dt_fit <- merge(dt_fit, data, by = id, all.x = TRUE)

  setnames(dt_fit, c(id),c("id"))

  if(predictions){
    maxn <-last(nodes)
    lvls <- as.character(sort(unique(nodes)))


  }else{

    maxn <- max(dt_fit$node)
    lvls <- as.character(sort(unique(dt_fit$node)))
  }



  dt_fit[,c("node",
            "k"):=list(factor(node, levels=lvls),
                       as.factor(k))]

  dt_fit[,node:=relevel(node,ref=as.character(maxn))]

  # dt_fit[,node:=relevel(node,ref=as.character(last(nodes)))]

  return(dt_fit)

}



## Matrix transformation ----

apply_transformations <- function(dt, variable_transformation) {
  stopifnot(data.table::is.data.table(dt))

  if (is.null(variable_transformation)) return(invisible(dt))

  # Normalize to a list of items (strings or formulas)
  items <- variable_transformation
  if (inherits(items, "formula")) items <- list(items)
  else if (is.character(items))    items <- as.list(items)
  else if (is.list(items))         items <- items
  else stop("Unsupported 'variable_transformation' type.")

  # Collect all (lhs name, rhs expression) pairs
  lhs_all  <- character()
  rhs_all  <- vector("list", 0L)

  for (it in items) {
    if (inherits(it, "formula")) {
      # formula: lhs ~ rhs
      lhs_raw <- paste(deparse(it[[2L]]), collapse = "")
      rhs_raw <- paste(deparse(it[[3L]]), collapse = "")
    } else if (is.character(it) && length(it) == 1L) {
      parts <- strsplit(it, "~", fixed = TRUE)[[1L]]
      if (length(parts) != 2L) stop("Each transformation must contain a single '~'.")
      lhs_raw <- parts[1L]
      rhs_raw <- parts[2L]
    } else {
      stop("List elements must be formulas or single strings of the form 'lhs ~ rhs'.")
    }

    lhs_vec <- trimws(strsplit(lhs_raw, "\\+")[[1L]])
    rhs_vec <- trimws(strsplit(rhs_raw, "\\+")[[1L]])

    if (length(lhs_vec) != length(rhs_vec)) {
      stop(
        sprintf("LHS and RHS have different lengths in '%s ~ %s' (%d vs %d).",
                lhs_raw, rhs_raw, length(lhs_vec), length(rhs_vec))
      )
    }

    # Parse each RHS into an expression
    rhs_exprs <- lapply(rhs_vec, function(s) parse(text = s)[[1L]])

    lhs_all <- c(lhs_all, lhs_vec)
    rhs_all <- c(rhs_all, rhs_exprs)
  }

  # Evaluate and assign each transformation.
  # Using one-by-one assignment keeps evaluation within data.table's environment
  # so column names are visible and functions come from the calling env.
  for (i in seq_along(lhs_all)) {
    dt[, (lhs_all[i]) := eval(rhs_all[[i]])]
  }

  invisible(dt)
}

sl_cut <- function(x, breaks, include.lowest = TRUE, right = TRUE) {

  labels <- paste0(head(breaks, -1), "-", tail(breaks, -1) - if (right) 1 else 0)
  # Return labeled factor
  cut(x, breaks = breaks, labels = labels, include.lowest = include.lowest, right = right)
}


# Other utils ----
create_formula <- function(covariates=NA_character_,
                           competing_risks=FALSE,
                           intercept=FALSE,
                           add_nodes=TRUE){


 xs<- NULL

  if (!any(is.na( covariates))) {
  xs <- paste(covariates, collapse = "+")
  }

  # if (!is.na( treatment)) {
  #   xs <- paste(xs, "+", treatment)
  # }

  # if (competing_risks) {
  #   xs <- paste(xs, "+ k")
  # }

  if (add_nodes) {
    xs <- paste(xs, "+ node")
  }

  if(!intercept){
    xs <- paste(xs, "-1")
  }

  out <- paste("deltaij ~", xs, "+offset(log(tij))", sep =
                 "")

  return(out)




}


create_formula_glmnet <- function(covariates=NA_character_,
                           add_nodes=TRUE){


  xs<- NULL

  if (!any(is.na( covariates))) {
    xs <- paste(covariates, collapse = "+")
  }

  # if (!is.na( treatment)) {
  #   xs <- paste(xs, "+", treatment)
  # }

  if (add_nodes) {
    xs <- paste(xs, "+ node")
  }

  out <- paste("deltaij ~", xs, "+offset(log(tij))", sep =
                 "")

  return(out)




}

create_formula_gam <- function(covariates=NA_character_,
                           competing_risks=FALSE,
                           intercept=FALSE,
                           add_nodes=TRUE){



  if (!any(is.na( covariates))) {
    xs <- paste(covariates, collapse = "+")
  }

  # if (!is.na( treatment)) {
  #   xs <- paste(xs, "+", treatment)
  # }

  # if (competing_risks) {
  #   xs <- paste(xs, "+ k")
  # }

  if (add_nodes) {
    xs <- paste(xs, "+ node")
  }

  # if(!intercept){
  #   xs <- paste(xs, "-1")
  # }

  out <- paste("deltaij ~", xs, sep =
                 "")

  return(out)




}

create_formula_hal <- function(covariates=NA_character_,
                           competing_risks=FALSE,
                           intercept=FALSE,
                           add_nodes=TRUE){



  if (!any(is.na( covariates))) {
    xs <- paste(covariates, collapse = "*")
  }

  # if (!is.na( treatment)) {
  #   xs <- paste(xs, "*", treatment)
  # }

  # if (competing_risks) {
  #   xs <- paste(xs, "+ k")
  # }

  if (add_nodes) {
    xs <- paste(c(xs, "node"), collapse = "*")
  }

  if(!intercept){
    xs <- paste(xs, "-1")
  }

  out <- paste("deltaij ~", xs, "+offset(log(tij))", sep =
                 "")

  return(out)




}


## Make fit cheaper

make_failed_fit <- function(reason = NA_character_) {
  structure(
    list(reason = reason),
    class = c("psl_failed_fit", "psl_fit_stub")
  )
}

is_failed_fit <- function(model) {
  inherits(model, "psl_failed_fit")
}

attach_psl_fit_meta <- function(model, meta = NULL) {
  if (is.null(model) || is_failed_fit(model)) {
    return(model)
  }
  attr(model, "psl_meta") <- meta
  model
}

get_psl_fit_meta <- function(model) {
  attr(model, "psl_meta", exact = TRUE)
}

resolve_prediction_model <- function(object, model = "sl") {

  n_learners <- length(object$learners)
  labels <- object$data_info$learners_labels

  if (is.null(model) || length(model) != 1L) {
    stop("'model' must be a scalar.")
  }

  if (is.numeric(model)) {
    if (is.na(model) || model != as.integer(model)) {
      stop("Numeric 'model' must be one of 0, 1, 2, ...")
    }

    model <- as.integer(model)

    if (model == 0L) {
      return(list(type = "sl", index = 0L, label = "sl"))
    }

    if (model >= 1L && model <= n_learners) {
      return(list(
        type = "learner",
        index = model,
        label = labels[model]
      ))
    }

    stop(
      sprintf(
        "Numeric 'model' must be 0 for the superlearner or an integer in 1:%d.",
        n_learners
      )
    )
  }

  if (!is.character(model) || is.na(model)) {
    stop("'model' must be a character scalar or a numeric scalar.")
  }

  model_chr <- trimws(model)
  model_chr_lc <- tolower(model_chr)

  if (model_chr_lc %in% c("sl", "superlearner", "super_learner")) {
    return(list(type = "sl", index = 0L, label = "sl"))
  }

  if (model_chr %in% labels) {
    j <- match(model_chr, labels)
    return(list(
      type = "learner",
      index = j,
      label = labels[j]
    ))
  }

  if (grepl("^learner_[0-9]+$", model_chr_lc)) {
    j <- as.integer(sub("^learner_", "", model_chr_lc))

    if (j >= 1L && j <= n_learners) {
      return(list(
        type = "learner",
        index = j,
        label = labels[j]
      ))
    }

    stop(
      sprintf(
        "'%s' is out of range. Available learners are learner_1, ..., learner_%d.",
        model_chr,
        n_learners
      )
    )
  }

  stop(
    sprintf(
      paste(
        "Unknown 'model' selector '%s'.",
        "Use 'sl', 0, a learner label, 'learner_j', or an integer in 1:%d."
      ),
      model_chr,
      n_learners
    )
  )
}


psl_get_labels <- function(object) {
  labs <- NULL

  if (!is.null(object$data_info) && !is.null(object$data_info$learners_labels)) {
    labs <- object$data_info$learners_labels
  }

  if (is.null(labs) || length(labs) == 0L) {
    if (!is.null(object$learners) && length(object$learners) > 0L) {
      labs <- names(object$learners)
    }
  }

  if (is.null(labs) || length(labs) == 0L) {
    labs <- paste0("learner_", seq_along(object$learners))
  }

  as.character(labs)
}

psl_get_stored_fit <- function(object, cause, model = "sl") {

  if (is.null(object$superlearner) || length(object$superlearner) == 0L) {
    stop("No fitted superlearner object available.")
  }

  if (length(cause) != 1L || is.na(cause) || cause != as.integer(cause)) {
    stop("'cause' must be a single positive integer.")
  }

  cause <- as.integer(cause)

  if (cause < 1L || cause > object$data_info$n_crisks) {
    stop(
      sprintf(
        "'cause' must be between 1 and %d.",
        object$data_info$n_crisks
      )
    )
  }

  sel <- resolve_prediction_model(object, model)
  labels <- psl_get_labels(object)
  sl_k <- object$superlearner[[cause]]

  if (sel$type == "sl") {
    if (!is.null(object$metalearner) && !is.null(sl_k$meta_learner_fit)) {
      return(list(
        fit = sl_k$meta_learner_fit,
        type = "sl",
        index = 0L,
        label = "sl"
      ))
    }

    if (length(object$learners) == 1L) {
      return(list(
        fit = sl_k$learners_fit,
        type = "learner",
        index = 1L,
        label = labels[1L]
      ))
    }

    stop("No fitted meta-learner available for model = 'sl'.")
  }

  j <- sel$index

  fit_j <- if (length(object$learners) == 1L) {
    sl_k$learners_fit
  } else {
    sl_k$learners_fit[[j]]
  }

  list(
    fit = fit_j,
    type = "learner",
    index = j,
    label = labels[j]
  )
}

psl_extract_meta_coefs <- function(fit, ...) {
  if (is.null(fit)) return(NULL)

  if (inherits(fit, "cv.glmnet")) {
    cc <- stats::coef(fit, s = "lambda.min", ...)
    cc <- as.matrix(cc)
    return(cc[, 1])
  }

  if (inherits(fit, "glmnet")) {
    lam <- fit$lambda
    lam_use <- if (length(lam) >= 1L) lam[[1L]] else NULL
    cc <- if (!is.null(lam_use)) {
      stats::coef(fit, s = lam_use, ...)
    } else {
      stats::coef(fit, ...)
    }
    cc <- as.matrix(cc)
    return(cc[, 1])
  }

  stats::coef(fit, ...)
}

psl_rename_z_in_text <- function(x, zmap) {
  if (is.null(x)) return(x)
  x <- as.character(x)
  for (z in names(zmap)) {
    x <- gsub(paste0("\\b", z, "\\b"), zmap[[z]], x, perl = TRUE)
  }
  x
}


# Cross-validation helpers ----


psl_init_oof_buffer <- function(dt_k, z_covariates) {
  structure(
    list(
      Z = matrix(
        NA_real_,
        nrow = nrow(dt_k),
        ncol = length(z_covariates),
        dimnames = list(NULL, z_covariates)
      )
    ),
    class = "psl_oof_buffer"
  )
}

psl_write_oof_buffer <- function(oof_buffer, oof_chunk) {
  if (is.null(oof_chunk) || length(oof_chunk$row_ix) == 0L) {
    return(oof_buffer)
  }

  oof_buffer$Z[oof_chunk$row_ix, ] <- oof_chunk$Z
  oof_buffer
}

psl_compact_oof_to_dt <- function(dt, dt_z, z_covariates) {
  if (is.data.table(dt_z)) {
    return(
      merge(
        dt_z[, !c("tij", "deltaij"), with = FALSE],
        dt,
        by = c("id", "folder", "node")
      )
    )
  }

  if (!(is.list(dt_z) && !is.null(dt_z$Z))) {
    stop("'dt_z' must be either a data.table or a compact OOF buffer with element 'Z'.")
  }

  tmp <- copy(dt)

  if ("row_ix" %in% names(tmp)) {
    tmp[, row_ix := NULL]
  }

  z_dt <- as.data.table(dt_z$Z)
  setnames(z_dt, z_covariates)

  tmp[, (z_covariates) := z_dt]

  # Preserve the old meta-learning interface, where a competing_risk column
  # was present after merging dt_z back into dt.
  tmp[, competing_risk := as.integer(as.character(k))]

  tmp
}
psl_oof_loghaz_dt <- function(oof_buffer, z_covariates) {
  out <- data.table::as.data.table(oof_buffer$Z)
  data.table::setnames(out, z_covariates)
  out
}
create_pseudo_observations <- function(training_data,
                                       validation_data,
                                       competing_risk,
                                       learners,
                                       z_covariates,
                                       ix) {
  if (nrow(validation_data) == 0L) {
    return(list(
      row_ix = integer(0),
      Z = matrix(
        NA_real_,
        nrow = 0L,
        ncol = length(z_covariates),
        dimnames = list(NULL, z_covariates)
      )
    ))
  }

  train_list <- lapply(learners, function(f) f$private_fit(training_data))

  val_list <- mapply(
    function(f, model, newdata) {
      f$private_predictor(model = model, newdata = newdata)
    },
    learners,
    train_list,
    MoreArgs = list(newdata = validation_data),
    SIMPLIFY = FALSE
  )

  val_mat <- do.call(cbind, lapply(val_list, as.numeric))

  if (is.null(dim(val_mat))) {
    val_mat <- matrix(val_mat, ncol = length(z_covariates))
  }

  val_mat <- log(val_mat)
  storage.mode(val_mat) <- "double"
  colnames(val_mat) <- z_covariates

  list(
    row_ix = validation_data[["row_ix"]],
    Z = val_mat
  )
}
# create_pseudo_observations <- function(training_data,
#                                        validation_data,
#                                        competing_risk,
#                                        learners,
#                                        z_covariates,
#                                        ix){
#   "
#   This function creates the pseudo-observations.
#
#   "
#
#   train_list <- lapply(learners, function(f) f$private_fit(training_data))
#
#
#   # Predict on the validation set your pseudo-observations ----
#   val_list <- mapply(
#     function(f, model, newdata)
#       f$private_predictor(model = model, newdata = newdata),
#     learners,
#     train_list,
#     MoreArgs = list(newdata = validation_data)
#   )
#
#
#
#   # val_list<- as.matrix(val_list)
#
#
#   val_list<- apply(as.matrix(val_list),
#                  MARGIN = 2,
#                  log)
#
#
#   # Name the columns
#
#   colnames(val_list) <- z_covariates
#
#
#   dt_z <-  data.table(val_list)[, c("id", "folder","node",paste0("deltaij"), "tij") := validation_data[,.(id,folder,node,deltaij,tij)]] #
#
#   dt_z[,competing_risk:= competing_risk]
#
#   return(dt_z)
#
#
#
# }
#

select_covariate_path <- function(dt, z_covariates, min_depth) {
  # add covariates column
  dt[, covariate := z_covariates]

  # order by deviance
  ordered_cov <- dt[order(deviance), covariate]

  # build list from min_depth up to full length
  lapply(min_depth:length(ordered_cov), function(k) ordered_cov[1:k])
}



merge_deltas <- function(lst) {
  if (length(lst) == 0L) return(data.table(id = integer(), node = integer()))

  # Keep only id, node, and delta_* columns from each table
  cleaned <- lapply(lst, function(DT) {
    DT <- as.data.table(copy(DT))
    dcols <- grep("^delta_", names(DT), value = TRUE)
    if (length(dcols) == 0L) stop("Each element must contain at least one delta_* column.")
    DT <- DT[, c("id", "node", dcols), with = FALSE]
    setkey(DT, id, node)
    DT
  })

  # Full outer merge across the list on id+node
  out <- Reduce(function(x, y) merge(x, y, by = c("id", "node"), all = TRUE), cleaned)

  # Order columns: id, node, then delta_* in natural order (by numeric suffix if present)
  deltas <- grep("^delta_", names(out), value = TRUE)
  ord <- order(suppressWarnings(as.integer(sub(".*?(\\d+)$", "\\1", deltas))), deltas)
  setcolorder(out, c("id", "node", deltas[ord]))

  out[]
}


## Meta learning ----

meta_learners_candidates <- function(meta_learner_algorithms,
                                     z_covariates){

  out <- list()

  ## this should be removed from the production version of the package!!!
  if("glm_ml_1"%in% meta_learner_algorithms){

    one_time_learner=list(glm_ml_1 =  Learner_glmnet(
      covariates = z_covariates,
      cross_validation = FALSE,
      intercept = FALSE,
      add_nodes = FALSE,
      penalise_nodes = TRUE,
      lambda=0
    ))

    out <- c(out,one_time_learner)

  }

  if("glmnet_ml_1"%in% meta_learner_algorithms){

    one_time_learner=list(glmnet_ml_1 =  Learner_glmnet(
      covariates = z_covariates,
      cross_validation = TRUE,
      intercept = FALSE,
      add_nodes = FALSE,
      penalise_nodes = TRUE
    ))

out <- c(out,one_time_learner)

  }

  if("glm_ml_2"%in% meta_learner_algorithms){

    one_time_learner=list(glm_ml_2 =  Learner_glmnet(
      covariates = c(z_covariates, paste0(z_covariates,":node")),
      cross_validation = FALSE,
      intercept = FALSE,
      add_nodes = TRUE,
      penalise_nodes = TRUE,
      lambda=0
    ))

    out <- c(out,one_time_learner)

  }

  if("glmnet_ml_2"%in% meta_learner_algorithms){

    one_time_learner=list(glmnet_ml_2 =  Learner_glmnet(
      covariates = c(z_covariates, paste0(z_covariates,":node")),
      cross_validation = TRUE,
      intercept = FALSE,
      add_nodes = TRUE,
      penalise_nodes = TRUE
    ))

    out <- c(out,one_time_learner)

  }

  if("glm_ml_3"%in% meta_learner_algorithms){

    one_time_learner=list(glm_ml_3 =  Learner_glmnet(
      covariates = z_covariates,
      cross_validation = FALSE,
      intercept = FALSE,
      add_nodes = TRUE,
      penalise_nodes = TRUE,
      lambda=0
    ))

    out <- c(out,one_time_learner)

  }

  if("glmnet_ml_3"%in% meta_learner_algorithms){

    one_time_learner=list(glmnet_ml_3 =  Learner_glmnet(
      covariates =  c(z_covariates, paste0(z_covariates,":node")),
      cross_validation = TRUE,
      intercept = FALSE,
      add_nodes = TRUE,
      penalise_nodes = FALSE
    ))

    out <- c(out,one_time_learner)

  }

  if("glm_ml_4"%in% meta_learner_algorithms){

    one_time_learner=list(glm_ml_4 =  Learner_glmnet(
      covariates =c(z_covariates, paste0(z_covariates,":node")),
      cross_validation = FALSE,
      intercept = FALSE,
      lambda=0,
      add_nodes = FALSE,
      penalise_nodes = TRUE
    ))

    out <- c(out,one_time_learner)

  }

  if("glmnet_ml_4"%in% meta_learner_algorithms){

    one_time_learner=list(glmnet_ml_4 =  Learner_glmnet(
      covariates = z_covariates,
      cross_validation = TRUE,
      intercept = FALSE,
      add_nodes = TRUE,
      penalise_nodes = TRUE
    ))

    out <- c(out,one_time_learner)

  }

  if("glmnet_ml_5"%in% meta_learner_algorithms){

    one_time_learner=list(glmnet_ml_5 =  Learner_glmnet(
      covariates = z_covariates,
      cross_validation = TRUE,
      intercept = FALSE,
      add_nodes = TRUE,
      penalise_nodes = FALSE
    ))

    out <- c(out,one_time_learner)

  }

  if("glmnet_ml_6"%in% meta_learner_algorithms){

    one_time_learner=list(glmnet_ml_6 =  Learner_glmnet(
      covariates =c(z_covariates, paste0(z_covariates,":node")),
      cross_validation = TRUE,
      intercept = FALSE,
      add_nodes = FALSE,
      penalise_nodes = TRUE
    ))

    out <- c(out,one_time_learner)

  }

  #########################################################


  if("glmnet" %in% meta_learner_algorithms){


    glmnet_meta_learners <- list(

      # Z1 + Z2
      glmnet_ml_1 =  Learner_glmnet(
        covariates = z_covariates,
        cross_validation = TRUE,
        intercept = FALSE,
        add_nodes = FALSE,
        penalise_nodes = TRUE
      )#,


      #next paper ----
      ## Z1*Z2*node
      # glmnet_ml_2 =  Learner_glmnet(
      #   covariates = c(z_covariates, paste0(z_covariates,":node")),
      #   cross_validation = TRUE,
      #   intercept = FALSE,
      #   add_nodes = TRUE,
      #   penalise_nodes = TRUE
      # ),
      ## Z1*Z2*node - not penalised
      # glmnet_ml_3 =  Learner_glmnet(
      #   covariates =  c(z_covariates, paste0(z_covariates,":node")),
      #   cross_validation = TRUE,
      #   intercept = FALSE,
      #   add_nodes = TRUE,
      #   penalise_nodes = FALSE
      # ),
      ## Z1+Z2+node
      # glmnet_ml_4 =  Learner_glmnet(
      #   covariates = z_covariates,
      #   cross_validation = TRUE,
      #   intercept = FALSE,
      #   add_nodes = TRUE,
      #   penalise_nodes = TRUE
      # ),
      ## Z1+Z2+node - not penalised
      # glmnet_ml_5 =  Learner_glmnet(
      #   covariates = z_covariates,
      #   cross_validation = TRUE,
      #   intercept = FALSE,
      #   add_nodes = TRUE,
      #   penalise_nodes = FALSE
      # ),
      ## Z1+Z2+Z1:node+Z2:node
      # glmnet_ml_6 =  Learner_glmnet(
      #   covariates =c(z_covariates, paste0(z_covariates,":node")),
      #   cross_validation = TRUE,
      #   intercept = FALSE,
      #   add_nodes = FALSE,
      #   penalise_nodes = TRUE
      # )
      #

      )

    out <- c(out,glmnet_meta_learners)


  }


  if("glm" %in% meta_learner_algorithms){


    glm_meta_learners <- list(

      # Z1 + Z2
      glm_ml_1 =  Learner_glmnet(
        covariates = z_covariates,
        cross_validation = FALSE,
        intercept = FALSE,
        add_nodes = FALSE,
        penalise_nodes = TRUE,
        lambda=0
      )#,
      ## Z1*Z2*node
      # glm_ml_2 =  Learner_glmnet(
      #   covariates = c(z_covariates, paste0(z_covariates,":node")),
      #   cross_validation = FALSE,
      #   intercept = FALSE,
      #   add_nodes = TRUE,
      #   penalise_nodes = TRUE,
      #   lambda=0
      # ),
      ## Z1+Z2+node
      # glm_ml_3 =  Learner_glmnet(
      #   covariates = z_covariates,
      #   cross_validation = FALSE,
      #   intercept = FALSE,
      #   add_nodes = TRUE,
      #   penalise_nodes = TRUE,
      #   lambda=0
      # ),
      ## Z1+Z2+Z1:node+Z2:node
      # glm_ml_4 =  Learner_glmnet(
      #   covariates =c(z_covariates, paste0(z_covariates,":node")),
      #   cross_validation = FALSE,
      #   intercept = FALSE,
      #   lambda=0,
      #   add_nodes = FALSE,
      #   penalise_nodes = TRUE
      # )

      )

    out <- c(out,glm_meta_learners)


  }

  names(out) <- paste(names(out),paste0(z_covariates, collapse = ""),sep = "_")

  return(out)

}


# fit_meta_learner <- function(dt,
#                              dt_z,
#                              meta_learner,
#                              z_covariates) {
#
#   tmp <- merge(
#     dt_z[, !c("tij", "deltaij"), with = FALSE],
#     dt,
#     by = c("id", "folder", "node")
#   )
#
#   meta_learner_fit <- meta_learner$private_fit(tmp)
#
#   return(meta_learner_fit)
# }

fit_meta_learner <- function(dt,
                             dt_z,
                             meta_learner,
                             z_covariates) {
  tmp <- psl_compact_oof_to_dt(
    dt = dt,
    dt_z = dt_z,
    z_covariates = z_covariates
  )

  meta_learner_fit <- meta_learner$private_fit(tmp)

  return(meta_learner_fit)
}

# super-learner hyper parameters cross-validation ----
meta_learner_cross_validation <- function(dt,
                                          dt_z,
                                          cr_ix,
                                          nfold,
                                          meta_learner) {
  out <- NULL

  z_covariates <- grep("^Z[0-9]+$", names(dt_z), value = TRUE)
  if (length(z_covariates) == 0L && is.list(dt_z) && !is.null(dt_z$Z)) {
    z_covariates <- colnames(dt_z$Z)
  }

  dt_meta <- psl_compact_oof_to_dt(
    dt = dt,
    dt_z = dt_z,
    z_covariates = z_covariates
  )

  for (v_fold_id in seq_len(nfold)) {
    meta_learner_fit <- meta_learner$private_fit(dt_meta[folder != v_fold_id, ])

    oos_data <- copy(dt_meta[folder == v_fold_id, ])
    oos_data[, tij := 1]

    fitted_hazard <- data.table(
      oos_data[["id"]],
      oos_data[["node"]],
      oos_data[["deltaij"]],
      v_fold_id,
      as.vector(meta_learner$private_predictor(meta_learner_fit, oos_data))
    )

    setnames(
      fitted_hazard,
      c("id", "node", paste0("delta_", cr_ix), "folder", paste0("pwch_", cr_ix))
    )

    out <- rbind(out, fitted_hazard)
  }

  return(out)
}

# meta_learner_cross_validation <- function(dt,
#                              dt_z,
#                              cr_ix,
#                              nfold,
#                              meta_learner){
#
#
#
#   # model output ---
#   out<-NULL
#
#   dt_z <- merge(dt_z,dt,by=c("id","folder","node"))
#
#
#   for(v_fold_id in 1:nfold){
#
#   meta_learner_fit <- meta_learner$private_fit(dt_z[folder!=v_fold_id,])
#
#   oos_data <- copy(dt_z[folder == v_fold_id, ])
#
#   # forced tij to be one
#   oos_data[,tij:=1]
#
#
#   fitted_hazard = data.table(
#     oos_data[['id']],
#     oos_data[['node']],
#     oos_data[['deltaij']],
#     v_fold_id,
#     as.vector(meta_learner$private_predictor(meta_learner_fit, oos_data))
#   )
#
#   setnames(fitted_hazard,c("id","node",paste0("delta_",cr_ix),"folder",paste0("pwch_", cr_ix)))
#
#   out <- rbind(out,fitted_hazard)
#
#   }
#
#   return(out)
#
#
# }
#


learners_hat <- function(crisk_cause,superlearner,newdata,learners){

  learners_predictions <- mapply(
    function(f, model, newdata)
      f$private_predictor(model = model, newdata = newdata),
    learners,
    superlearner$learners_fit,
    MoreArgs = list(newdata = newdata)
  )

  return(learners_predictions)



}

# Hal helpers ----

mk_main <- function(x, v, K) {
  if (is.numeric(x)) {
    out <- mk_main_numeric_cpp(x, as.integer(K))
    cps <- out$cutpoints
    idxs <- out$idxs

    cps <- as.numeric(cps)
    if (!length(cps)) {
      return(list(idxs = list(), names = character(),
                  prim_meta = list(), var_meta = list(cutpoints = numeric())))
    }

    # nms <- sprintf("I(%s<=%.10g)", v, cps)
    nms <- sprintf("I(%s>=%.10g)", v, cps)

    prim_meta <- lapply(cps, function(cu) list(var = v, kind = "numeric", cutpoint = cu))
    list(
      idxs = idxs,
      names = nms,
      prim_meta = prim_meta,
      var_meta = list(cutpoints = cps)
    )

  } else if (is.factor(x)) {
    lv <- levels(x)
    if (!length(lv)) {
      return(list(idxs = list(), names = character(),
                  prim_meta = list(), var_meta = list(levels = character(), ref = NA_character_)))
    }

    # Drop a reference level (default: first level)
    ref <- lv[1]
    keep <- setdiff(lv, ref)

    # If only one level, nothing to add
    if (!length(keep)) {
      return(list(
        idxs = list(),
        names = character(),
        prim_meta = list(),
        var_meta = list(levels = lv, ref = ref)
      ))
    }

    # Compute idxs for all levels, then drop the reference entry
    out_all <- mk_main_factor_cpp(as.integer(x), length(lv))
    idxs_all <- out_all$idxs

    # Map level -> position (mk_main_factor_cpp is assumed to return in level order)
    pos_keep <- match(keep, lv)

    idxs <- idxs_all[pos_keep]
    nms <- sprintf("I(%s==%s)", v, make.names(keep, unique = TRUE))
    prim_meta <- lapply(keep, function(L) list(var = v, kind = "factor", level = L, ref = ref))

    list(
      idxs = idxs,
      names = nms,
      prim_meta = prim_meta,
      var_meta = list(levels = lv, ref = ref)
    )

  } else {
    stop(sprintf("Unsupported type for %s", v))
  }
}

##
add_cols <- function(state, idxs_list, nm_vec, col_meta_list) {
  out <- add_cols_cpp(idxs_list, p_start = state$p)
  ncol_added <- out$ncol
  if (!ncol_added) return(state)

  keep <- out$keep
  nm_vec <- nm_vec[keep]
  col_meta_list <- col_meta_list[keep]

  state$chunk_used <- state$chunk_used + 1L
  state$I_chunks[[state$chunk_used]] <- out$i
  state$J_chunks[[state$chunk_used]] <- out$j
  state$name_chunks[[state$chunk_used]] <- nm_vec
  state$colmeta_chunks[[state$chunk_used]] <- col_meta_list

  state$p <- state$p + ncol_added
  state
}

#' Absolute risk (cumulative incidence) for a cause under piecewise-constant hazards
#'
#' Computes, per row, the cumulative incidence function at the end of each interval,
#' grouped by `id`. The number of causes is inferred from the number of columns in `haz`.
#'
#' @param id Integer vector. Sorted by `id` then time.
#' @param dt Numeric vector of interval lengths.
#' @param haz Numeric matrix (n x C) of cause-specific hazards per interval.
#'           Columns correspond to causes 1..C.
#' @param cause_idx Integer. Index of the cause of interest (1-based by default).
#' @param one_based Logical. If `TRUE`, `cause_idx` is 1-based. If `FALSE`, 0-based.
#' @param na_is_zero Logical. If `TRUE`, treat NA/Inf hazards as zero.
#'
#' @return Numeric vector of cumulative incidence values at the end of each interval.
#'
#' @examples
#' id <- c(1L, 1L, 2L, 2L)
#' dt <- c(1, 1, 1, 1)
#' haz <- rbind(
#'   c(0.10, 0.05),
#'   c(0.20, 0.10),
#'   c(0.05, 0.02),
#'   c(0.10, 0.03)
#' )
#' pch_absolute_risk(id = id, dt = dt, haz = haz, cause_idx = 1)
#'
#'
#' @export
#' @useDynLib poissonsuperlearner, .registration = TRUE
#' @importFrom Rcpp evalCpp
pch_absolute_risk <- function(id, dt, haz, cause_idx, one_based = TRUE, na_is_zero = FALSE) {
  .Call(`_poissonsuperlearner_pch_absolute_risk`, id, dt, haz, as.integer(cause_idx),
        one_based, na_is_zero)
}


#' Absolute risk (Euler approximation) for a cause under piecewise-constant hazards
#'
#' Computes the cumulative incidence function using the first-order Euler
#' (discrete) approximation:
#'   \deqn{F_j(t) \approx \sum S(t_{k-1}) \lambda_{j,k} \Delta t_k}{
#'     F_j(t) = approx sum S(t_{k-1}) lambda_{j,k} Delta t_k
#'   }
#' Grouped by `id`, this returns the cumulative incidence at the end of each interval.
#'
#' @param id Integer vector. Sorted by `id` then time.
#' @param dt Numeric vector of interval lengths.
#' @param haz Numeric matrix (n x C) of cause-specific hazards per interval.
#' @param cause_idx Integer. Index of the cause of interest (1-based by default).
#' @param one_based Logical. If `TRUE`, `cause_idx` is 1-based. If `FALSE`, 0-based.
#' @param na_is_zero Logical. If `TRUE`, treat NA/Inf hazards as zero.
#'
#' @return Numeric vector of cumulative incidence values (Euler approximation)
#'         at the end of each interval.
#'
#'
#' @examples
#' id <- c(1L, 1L, 2L, 2L)
#' dt <- c(1, 1, 1, 1)
#' haz <- rbind(
#'   c(0.10, 0.05),
#'   c(0.20, 0.10),
#'   c(0.05, 0.02),
#'   c(0.10, 0.03)
#' )
#' pch_absolute_risk_euler(id = id, dt = dt, haz = haz, cause_idx = 1)
#'
#' @export
#' @useDynLib poissonsuperlearner, .registration = TRUE
#' @importFrom Rcpp evalCpp
pch_absolute_risk_euler <- function(id, dt, haz, cause_idx, one_based = TRUE, na_is_zero = FALSE) {
  .Call(`_poissonsuperlearner_pch_absolute_risk_euler`, id, dt, haz, as.integer(cause_idx),
        one_based, na_is_zero)
}

#' Absolute-risk matrix predictions for a fitted Poisson Super Learner
#'
#' S3 method compatible with `riskRegression::predictRisk` returning one column
#' per requested time.
#'
#' @param object `poisson_superlearner`. Fitted object.
#' @param newdata `data.frame`. New covariate data.
#' @param times `numeric`. Prediction times.
#' @param cause `numeric(1)`. Cause index.
#' @param model Scalar model selector. Default is `"sl"`. Allowed values are the
#' same as in [predict.poisson_superlearner()].
#' @param ... Unused.
#'
#' @return `numeric` matrix with `nrow(newdata)` rows and `length(times)` columns.
#'
#' @examples
#' d <- simulateStenoT1(30, competing_risks = TRUE)
#'
#' learners <- list(
#'   lasso = Learner_glmnet(
#'     covariates = "sex",
#'     alpha = 1,
#'     lambda = 0.01,
#'     cross_validation = FALSE
#'   ),
#'   ridge = Learner_glmnet(
#'     covariates = c("sex", "value_LDL"),
#'     alpha = 0,
#'     lambda = 0.01,
#'     cross_validation = FALSE
#'   )
#' )
#'
#' fit <- Superlearner(
#'   data = d,
#'   id = "id",
#'   status = "status_cvd",
#'   event_time = "time_cvd",
#'   learners = learners,
#'   number_of_nodes = 3,
#'   nfold = 2
#' )
#'
#' if (requireNamespace("riskRegression", quietly = TRUE)) {
#'   riskRegression::predictRisk(fit, newdata = d[1:3], times = c(1, 3), cause = 1)
#' }
#'
#' @export
#' @export predictRisk.poisson_superlearner
predictRisk.poisson_superlearner <- function(object,
                                             newdata,
                                             times,
                                             cause = 1,
                                             model = "sl",
                                             ...) {

  rr_output <- matrix(
    NA_real_,
    nrow = nrow(newdata),
    ncol = length(times)
  )

  pred <- predict(
    object,
    newdata = newdata,
    times = times,
    cause = cause,
    model = model
  )

  if (is.null(pred)) {
    return(NULL)
  }

  time_col <- object$data_info$event_time
  id_col <- object$data_info$id

  for (ix in seq_along(times)) {
    tmp <- pred[get(time_col) == times[ix]]

    if (nrow(tmp) == 0L) {
      next
    }

    data.table::setorderv(tmp, id_col)
    rr_output[, ix] <- tmp[["absolute_risk"]]
  }

  return(rr_output)
}
#' Absolute-risk matrix predictions for a fitted base learner
#'
#' @param object `base_learner`. Fitted object from [fit_learner()].
#' @param newdata `data.frame`. New covariate data.
#' @param times `numeric`. Prediction times.
#' @param cause `numeric(1)`. Cause index.
#' @param ... Unused.
#'
#' @return `numeric` matrix with `nrow(newdata)` rows and `length(times)` columns.
#'
#' @examples
#' d <- simulateStenoT1(30, competing_risks = TRUE)
#' lrn <- Learner_glmnet(
#'   covariates = c("sex", "value_LDL"),
#'   lambda = 0.01,
#'   cross_validation = FALSE
#' )
#' bl <- fit_learner(
#'   d,
#'   learner = lrn,
#'   id = "id",
#'   status = "status_cvd",
#'   event_time = "time_cvd",
#'   number_of_nodes = 3
#' )
#'
#' if (requireNamespace("riskRegression", quietly = TRUE)) {
#'   riskRegression::predictRisk(bl, newdata = d[1:3], times = c(1, 3), cause = 1)
#' }
#'
#'
#' @export
predictRisk.base_learner <- function(object,
                                     newdata,
                                     times,
                                     cause = 1,
                                     ...) {

  rr_output <- matrix(
    NA_real_,
    nrow = nrow(newdata),
    ncol = length(times)
  )

  pred <- predict(
    object,
    newdata = newdata,
    times = times,
    cause = cause
  )

  if (is.null(pred)) {
    return(NULL)
  }

  time_col <- object$data_info$event_time
  id_col <- object$data_info$id

  for (ix in seq_along(times)) {
    tmp <- pred[get(time_col) == times[ix]]

    if (nrow(tmp) == 0L) {
      next
    }

    data.table::setorderv(tmp, id_col)
    rr_output[, ix] <- tmp[["absolute_risk"]]
  }

  return(rr_output)
}

#' Piecewise-constant hazards survival function
#'
#' Computes survival at the end of each interval for competing risks with
#' piecewise constant hazards.
#'
#' @param id Integer vector of subject IDs, sorted by id then time.
#' @param dt Numeric vector of interval lengths.
#' @param haz Numeric matrix (n x C) of cause-specific hazards.
#' @param na_is_zero Logical. If TRUE, treat NA hazards as zero.
#'
#' @return Numeric vector of survival probabilities at the end of each interval.
#'
#' @examples
#' id <- c(1L, 1L, 2L, 2L)
#' dt <- c(1, 1, 1, 1)
#' haz <- rbind(
#'   c(0.10, 0.05),
#'   c(0.20, 0.10),
#'   c(0.05, 0.02),
#'   c(0.10, 0.03)
#' )
#' pch_survival(id = id, dt = dt, haz = haz)
#'
#' @export
#' @useDynLib poissonsuperlearner, .registration = TRUE
#' @importFrom Rcpp evalCpp
pch_survival <- function(id, dt, haz, na_is_zero = FALSE) {
  .Call(`_poissonsuperlearner_pch_survival`, id, dt, haz, na_is_zero)
}

#' Penalized Poisson learner via `glmnet`
#'
#' `Learner_glmnet` is a Reference Class implementing the learner interface
#' used by [Superlearner()] and [fit_learner()].
#'
#' **User-facing API:** users are expected to **initialize** the learner (i.e.,
#' call `Learner_glmnet(...)`) and pass the resulting object to
#' [Superlearner()] or [fit_learner()]. The remaining methods documented below
#' (e.g., `private_fit()`, `private_predictor()`) are part of the internal learner
#' interface and are **not meant to be called directly by users**.
#'
#' **Wrapper role:** this class is a user-friendly wrapper around the existing
#' `glmnet` implementation. The package-specific contribution is to provide a
#' piecewise-constant hazard workflow: create the long-format Poisson data with
#' offsets for time at risk, add the interval ("node") structure for the baseline
#' hazard when requested, and forward standard `glmnet` arguments supplied at
#' initialization to the backend fitter.
#'
#' @section Model:
#' Let \eqn{0=t_0 < t_1 < \cdots < t_m}{0=t0 < t1 < ... < tm} denote time knots and
#' define interval indicators \eqn{I_k(t)=1\{t\in(t_k,t_{k+1}]\}}{Ik(t)=I(t in (tk, t{k+1}])}.
#' The piecewise-constant hazard model is
#' \deqn{
#'   \lambda(t \mid x) = \sum_{k=0}^{m} I_k(t)\,\lambda_k(x),
#'   \qquad \lambda_k(x) = \exp(\beta^\top x + \gamma_k).
#' }{
#'   lambda(t|x) = sum_{k=0}^m Ik(t) * lambda_k(x),
#'   lambda_k(x) = exp(beta^T x + gamma_k).
#' }
#' Penalization is applied to the regression coefficients through the `glmnet`
#' elastic-net penalty. If you want node (baseline) terms to be unpenalized, use
#' `penalty.factor` via `...` (and set it consistently with how your design matrix
#' encodes nodes).
#'
#' @section Fields:
#' \describe{
#'   \item{`covariates` (`character`)}{Names of covariate columns used in the model.}
#'   \item{`cross_validation` (`logical`)}{If `TRUE`, chooses `lambda` by `glmnet::cv.glmnet`.}
#'   \item{`intercept` (`logical`)}{Whether to include an intercept in the backend fit.}
#'   \item{`add_nodes` (`logical`)}{If `TRUE`, include interval ("node") effects encoding the baseline hazard.}
#'   \item{`lambda` (`numeric`)}{If `cross_validation=FALSE`, the `lambda` used in the final fit.}
#'   \item{`formula` (`character`)}{Formula string used to create the design matrix in long format.}
#'   \item{`learner` (`function`)}{Backend fitter (`glmnet::glmnet` or `glmnet::cv.glmnet`).}
#'   \item{`fit_arguments` (`list`)}{Additional arguments forwarded to the backend fitter.}
#' }
#'
#' @section Methods (internal learner interface):
#' \describe{
#'   \item{`initialize(...)`}{Construct and configure the learner. This is the only method users should call.}
#'   \item{`private_fit(data, ...)`}{Internal. Fits a Poisson model with offset `log(tij)` on long-format data.}
#'   \item{`private_predictor(model, newdata, ...)`}{Internal. Predicts hazards on the response scale for long-format `newdata`.}
#' }
#'
#' @examples
#' lrn <- Learner_glmnet(covariates = c("age", "sex"), alpha = 1, cross_validation = TRUE)
#'
#' @export Learner_glmnet
#' @exportClass Learner_glmnet
Learner_glmnet <- setRefClass(
  "Learner_glmnet",
  fields = list(
    covariates = "character",
    cross_validation = "logical",
    intercept = "logical",
    formula = "character",
    learner = "function",
    add_nodes = "logical",
    lambda = "numeric",
    fit_arguments = "list"
  ),
  methods = list(
    initialize = function(covariates = NA_character_,
                          cross_validation = FALSE,
                          intercept = TRUE,
                          add_nodes = TRUE,
                          lambda = NA_real_,
                          ...) {
      .self$covariates <- covariates

      .self$cross_validation <- cross_validation

      .self$intercept <- intercept

      .self$add_nodes <- add_nodes

      tmp <- lambda
      if (is.null(lambda))
        tmp <- NA_real_
      if (length(lambda) == 0L)
        tmp <- NA_real_
      tmp <- as.numeric(tmp)

      # create formula for competing risks. It is correct in the fit method if survival.
      .self$formula <- create_formula_glmnet(
        covariates = .self$covariates,
        add_nodes = .self$add_nodes
      )



      .self$fit_arguments <- list(...)

      .self$fit_arguments[['family']] <- "poisson"

      .self$fit_arguments[['intercept']] <- .self$intercept

      if (.self$cross_validation) {
        .self$learner = glmnet::cv.glmnet

      } else{
        .self$learner = glmnet::glmnet
        .self$fit_arguments[['lambda']] <- tmp
      }


    },

    private_fit = function(data, ...) {
      .extract_symbols <- function(term) {
        tryCatch(
          all.vars(str2lang(term)),
          error = function(e) {
            tryCatch(
              all.vars(stats::terms(stats::as.formula(paste("~", term)))),
              error = function(e2) character(0)
            )
          }
        )
      }

      fit <- tryCatch({

        group_cols <- .self$covariates[complete.cases(.self$covariates)]
        group_cols <- group_cols[is.character(group_cols) & !is.na(group_cols) & nzchar(group_cols)]
        group_cols <- unlist(lapply(group_cols, .extract_symbols), use.names = FALSE)
        group_cols <- unique(group_cols)

        data <- data[, .(tij = sum(tij), deltaij = sum(deltaij)), by = c(group_cols, "node", "k")]
        data <- data[complete.cases(data), ]

        x <- Matrix::sparse.model.matrix(
          formula(.self$formula),
          data,
          contrasts.arg = NULL
        )[, -1, drop = FALSE]

        if (nrow(x) == 0L || ncol(x) == 0L) {
          return(make_failed_fit("Empty design matrix in Learner_glmnet::private_fit"))
        }

        pf <- 1 - grepl("node", colnames(x))

        if (.self$cross_validation) {
          cv_args <- .self$fit_arguments

          if (is.null(cv_args[["type.measure"]])) {
            cv_args[["type.measure"]] <- "deviance"
          }

          cv_fit <- tryCatch(
            suppressWarnings(
              do.call(glmnet::cv.glmnet, c(
                list(
                  x = x,
                  y = as.vector(data[["deltaij"]]),
                  offset = log(as.vector(data[["tij"]])),
                  penalty.factor = pf
                ),
                cv_args
              ))
            ),
            error = function(e1) {
              tryCatch(
                suppressWarnings(
                  do.call(glmnet::cv.glmnet, c(
                    list(
                      x = x,
                      y = as.vector(data[["deltaij"]]),
                      offset = log(as.vector(data[["tij"]])),
                      penalty.factor = rep(1, ncol(x))
                    ),
                    cv_args
                  ))
                ),
                error = function(e2) NULL
              )
            }
          )

          if (is.null(cv_fit) || length(cv_fit$lambda.min) == 0L || !is.finite(cv_fit$lambda.min)) {
            return(make_failed_fit("cv.glmnet failed in Learner_glmnet::private_fit"))
          }

          .self$lambda <- as.numeric(cv_fit$lambda.min)

          cv_fit

        } else {
          suppressWarnings(
            do.call(glmnet::glmnet, c(
              .self$fit_arguments,
              list(
                x = x,
                y = as.numeric(data[["deltaij"]]),
                offset = log(data[["tij"]])
              )
            ))
          )
        }

      }, error = function(e) {
        make_failed_fit(conditionMessage(e))
      })

      fit
    },

    private_predictor = function(model, newdata, ...) {
      if (is_failed_fit(model)) {
        return(rep(NA_real_, nrow(newdata)))
      }

      x_new <- tryCatch(
        Matrix::sparse.model.matrix(
          formula(.self$formula),
          newdata,
          contrasts.arg = NULL
        )[, -1, drop = FALSE],
        error = function(e) NULL
      )

      if (is.null(x_new) || nrow(x_new) == 0L || ncol(x_new) == 0L) {
        return(rep(NA_real_, nrow(newdata)))
      }

      pred_args <- list(
        object = model,
        newx = x_new,
        newoffset = log(1),
        type = "response"
      )

      dots <- list(...)
      if (inherits(model, "cv.glmnet") && is.null(dots[["s"]])) {
        pred_args[["s"]] <- "lambda.min"
      }

      out <- tryCatch(
        do.call(predict, c(pred_args, dots)),
        error = function(e) rep(NA_real_, nrow(newdata))
      )

      as.numeric(out)
    }

  )
)

#' HAL learner for piecewise Poisson hazards
#'
#' `Learner_hal` is a Reference Class implementing the learner interface
#' used by [Superlearner()] and [fit_learner()].
#'
#' **User-facing API:** users should **only initialize** the learner and pass it
#' to [Superlearner()] / [fit_learner()]. The methods `private_fit()` and
#' `private_predictor()` (and any basis-construction helpers) are part of the
#' internal learner interface and are **not meant to be called directly by users**.
#'
#' **Wrapper role:** this class provides a piecewise-constant hazard wrapper around
#' a HAL-style indicator-basis construction, estimated by L1-penalized Poisson
#' regression using a `glmnet` backend. The package-specific contribution is to
#' (i) construct the long-format Poisson representation with offsets for time at
#' risk, (ii) generate indicator bases compatible with piecewise hazards, and
#' (iii) forward backend fitting arguments supplied via `...`.
#'
#' @section Model:
#' Let \eqn{0=t_0 < t_1 < \cdots < t_m}{0=t0 < t1 < ... < tm} denote time knots and
#' define interval indicators \eqn{I_k(t)=1\{t\in(t_k,t_{k+1}]\}}{Ik(t)=I(t in (tk, t{k+1}])}.
#' The HAL piecewise-constant hazard model is
#' \deqn{
#'   \lambda(t \mid x) = \sum_{k=0}^{m} I_k(t)\,\exp\{f(t,x)\},
#' }{
#'   lambda(t|x) = sum_{k=0}^m Ik(t) * exp(f(t,x)).
#' }
#' where \eqn{f(t,x)}{f(t,x)} is approximated by a finite linear combination of
#' indicator basis functions.
#'
#' @section Two-covariate illustration:
#' Let \eqn{x=(x_1,x_2)}{x=(x1,x2)} be two covariates and let
#' \eqn{t_0 < t_1 < \cdots < t_R}{t0 < t1 < ... < tR} be time grid points used to
#' create step functions in time. Choose covariate cutpoints
#' \eqn{c_{1,1},\ldots,c_{1,K_1}}{c1_1,...,c1_K1} for \eqn{x_1}{x1} and
#' \eqn{c_{2,1},\ldots,c_{2,K_2}}{c2_1,...,c2_K2} for \eqn{x_2}{x2}.
#'
#' Define indicator bases:
#' \deqn{B_r(t) = 1\{t_r \le t\}}{Br(t) = I(tr <= t),}
#' \deqn{B_{1,p}(x) = 1\{c_{1,p} \le x_1\}}{B1p(x) = I(c1p <= x1),}
#' \deqn{B_{2,q}(x) = 1\{c_{2,q} \le x_2\}}{B2q(x) = I(c2q <= x2).}
#'
#' A main-effects HAL approximation on the log-hazard scale can be written as:
#' \deqn{
#'   f_\beta(t,x) = \beta_0
#'   + \sum_{r=1}^R \beta_r B_r(t)
#'   + \sum_{r=1}^R\sum_{p=1}^{K_1} \beta_{r,1,p} B_r(t) B_{1,p}(x)
#'   + \sum_{r=1}^R\sum_{q=1}^{K_2} \beta_{r,2,q} B_r(t) B_{2,q}(x).
#' }{
#'   f_beta(t,x) = beta0
#'   + sum_{r=1}^R beta_r Br(t)
#'   + sum_{r=1}^R sum_{p=1}^K1 beta_{r,1,p} Br(t) B1p(x)
#'   + sum_{r=1}^R sum_{q=1}^K2 beta_{r,2,q} Br(t) B2q(x).
#' }
#' If `max_degree >= 2`, the learner additionally includes interaction bases such as
#' \deqn{
#'   \sum_{r=1}^R\sum_{p=1}^{K_1}\sum_{q=1}^{K_2}
#'   \beta_{r,12,pq} B_r(t) B_{1,p}(x) B_{2,q}(x).
#' }{
#'   sum_{r=1}^R sum_{p=1}^K1 sum_{q=1}^K2 beta_{r,12,pq} Br(t) B1p(x) B2q(x).
#' }
#'
#' @section How reference class parameters map to the model:
#' \describe{
#'   \item{`covariates`}{Covariate columns used to build covariate indicator bases.}
#'   \item{`num_knots`}{Controls the number of cutpoints per covariate used for indicator bases.}
#'   \item{`max_degree`}{Maximum interaction order included in the basis expansion.}
#'   \item{`add_nodes`}{If `TRUE`, includes interval ("node") structure for the baseline hazard.}
#'   \item{`intercept`}{Whether the backend penalized regression includes an intercept term.}
#'   \item{`cross_validation`}{If `TRUE`, selects the penalty level using `glmnet::cv.glmnet`.}
#'   \item{`fit_arguments`}{Additional arguments forwarded to the `glmnet` backend (e.g. `nfolds`).}
#' }
#'
#' @section Fields:
#' \describe{
#'   \item{`covariates` (`character`)}{Names of covariate columns used in the basis.}
#'   \item{`cross_validation` (`logical`)}{Whether to use `cv.glmnet` to select the penalty.}
#'   \item{`intercept` (`logical`)}{Backend intercept flag.}
#'   \item{`add_nodes` (`logical`)}{Whether node (time-interval) effects are included.}
#'   \item{`max_degree` (`integer`)}{Maximum interaction order.}
#'   \item{`num_knots` (`numeric`)}{Knots used for basis construction.}
#'   \item{`lambda_opt` (`numeric`)}{Selected penalty level when using cross-validation.}
#'   \item{`fit_arguments` (`list`)}{Extra backend arguments forwarded to `glmnet`.}
#' }
#'
#' @section Methods (internal learner interface):
#' \describe{
#'   \item{`initialize(...)`}{Construct and configure the learner. This is the only method users should call.}
#'   \item{`private_fit(data, ...)`}{Internal. Builds bases and fits the penalized Poisson model with offset `log(tij)`.}
#'   \item{`private_predictor(model, newdata, ...)`}{Internal. Evaluates the fitted approximation and returns hazards on the response scale.}
#' }
#'
#' @examples
#' lrn <- Learner_hal(covariates = c("age", "sex"), max_degree = 2L, num_knots = c(10L, 5L))
#'
#' @export Learner_hal
#' @exportClass Learner_hal
Learner_hal <- setRefClass(
  "Learner_hal",
  fields = list(
    covariates = "character",
    cross_validation = "logical",
    intercept = "logical",
    formula = "character",
    learner = "function",
    add_nodes = "logical",
    max_degree = "integer",
    lambda_opt = "numeric",
    maxit_prefit = "numeric",
    fit_arguments = "list",
    num_knots = "numeric"
  ),
  methods = list(
    initialize = function(covariates = NA_character_,
                          intercept = TRUE,
                          cross_validation = TRUE,
                          add_nodes = TRUE,
                          max_degree = 2,
                          maxit_prefit = NA_real_,
                          num_knots = c(10L, 5L),
                          ...) {
      .self$covariates <- covariates

      .self$cross_validation <- cross_validation

      .self$intercept <- intercept

      .self$add_nodes <- add_nodes

      .self$max_degree <- max_degree

      .self$num_knots <- num_knots

      # normalize user input
      # tmp <- lambda_grid
      # if (is.null(lambda_grid))
      #   tmp <- NA_real_
      # if (length(lambda_grid) == 0L)
      #   tmp <- NA_real_
      #
      # .self$lambda_grid <- tmp

      .self$maxit_prefit <- maxit_prefit

      # create formula for competing risks. It is correct in the fit method if survival.
      .self$formula <- create_formula_hal(
        covariates = .self$covariates,
        intercept = FALSE,
        #in the glmnet case, intercept is handled separately.
        add_nodes = .self$add_nodes
      )

      if (.self$cross_validation) {
        .self$learner <- glmnet::cv.glmnet
      } else {
        .self$learner <- glmnet::glmnet
      }

      .self$fit_arguments <- list(...)

      .self$fit_arguments[['family']] <- "poisson"

      .self$fit_arguments[['intercept']] <- .self$intercept




    },
    hal_basis = function(vars,
                         DT,
                         max_interaction = 2L,
                         knots_per_order = rep(10L, 2L)) {
      stopifnot(requireNamespace("data.table", quietly = TRUE))
      stopifnot(requireNamespace("Matrix", quietly = TRUE))

      if (!data.table::is.data.table(DT))
        DT <- data.table::as.data.table(DT)
      if (length(knots_per_order) != max_interaction)
        stop("knots_per_order must have length max_interaction")
      if (!all(vars %in% names(DT)))
        stop("vars not all found in DT")

      n <- nrow(DT)

      DT_local <- DT[, ..vars]
      for (v in vars) {
        x <- DT_local[[v]]
        if (is.character(x) ||
            is.logical(x))
          DT_local[[v]] <- factor(x)
      }

      primitives <- setNames(vector("list", length(vars)), vars)
      meta_per_var <- setNames(vector("list", length(vars)), vars)

      for (v in vars) {
        res <- mk_main(DT_local[[v]], v, K = knots_per_order[1L])
        primitives[[v]] <- res
        meta_per_var[[v]] <- res$var_meta
      }

      # state for chunked construction (purely threaded)
      state <- list(
        I_chunks = vector("list", 0L),
        J_chunks = vector("list", 0L),
        name_chunks = vector("list", 0L),
        colmeta_chunks = vector("list", 0L),
        chunk_used = 0L,
        p = 0L
      )

      # mains
      for (v in vars) {
        prim <- primitives[[v]]
        if (!length(prim$idxs))
          next
        mains_meta <- lapply(prim$prim_meta, function(m)
          c(list(type = "main"), m))
        state <- add_cols(state, prim$idxs, prim$names, mains_meta)
      }

      # interactions
      if (max_interaction >= 2L) {
        for (d in 2:max_interaction) {
          if (length(vars) < d)
            break

          cmbs <- utils::combn(vars, d, simplify = FALSE)
          Kcap <- knots_per_order[d]

          for (cmb in cmbs) {
            prim_d <- lapply(cmb, function(v) {
              idxs <- primitives[[v]]$idxs
              pm   <- primitives[[v]]$prim_meta
              if (!length(idxs))
                return(list(
                  idxs = list(),
                  names = character(),
                  meta = list()
                ))

              is_num <- length(pm) > 0L &&
                identical(pm[[1]]$kind, "numeric")
              ncap <- if (is_num)
                min(length(idxs), Kcap)
              else
                length(idxs)

              list(idxs  = idxs[seq_len(ncap)],
                   names = primitives[[v]]$names[seq_len(ncap)],
                   meta  = pm[seq_len(ncap)])
            })

            if (any(vapply(prim_d, function(z)
              length(z$idxs), integer(1L)) == 0L))
              next

            lens <- vapply(prim_d, function(z)
              length(z$idxs), integer(1L))
            total <- prod(lens)
            if (!total)
              next

            for (t in 0:(total - 1L)) {
              sel <- integer(d)
              rem <- t
              for (j in seq_len(d)) {
                sel[j] <- (rem %% lens[j]) + 1L
                rem <- rem %/% lens[j]
              }

              chosen_idx <- vector("list", d)
              nm_parts <- character(d)
              parts_meta <- vector("list", d)

              for (j in seq_len(d)) {
                chosen_idx[[j]] <- prim_d[[j]]$idxs[[sel[j]]]
                nm_parts[j] <- prim_d[[j]]$names[[sel[j]]]
                parts_meta[[j]] <- prim_d[[j]]$meta[[sel[j]]]
              }

              idx_prod <- interN_cpp(chosen_idx)
              if (!length(idx_prod))
                next

              state <- add_cols(
                state,
                idxs_list = list(idx_prod),
                nm_vec = paste0(nm_parts, collapse = ":"),
                col_meta_list = list(
                  list(
                    type = "interaction",
                    order = d,
                    parts = parts_meta
                  )
                )
              )
            }
          }
        }
      }

      if (state$chunk_used == 0L)
        stop("No basis columns created.")

      I_chunks <- state$I_chunks
      J_chunks <- state$J_chunks
      name_chunks <- state$name_chunks
      colmeta_chunks <- state$colmeta_chunks

      i_vec <- unlist(I_chunks, use.names = FALSE)
      j_vec <- unlist(J_chunks, use.names = FALSE)

      names_all <- unlist(name_chunks, use.names = FALSE)
      per_col_meta <- unlist(colmeta_chunks, recursive = FALSE, use.names = FALSE)

      X <- Matrix::sparseMatrix(
        i = if (length(i_vec))
          i_vec
        else
          integer(),
        j = if (length(j_vec))
          j_vec
        else
          integer(),
        x = if (length(i_vec))
          1L
        else
          integer(),
        dims = c(n, if (length(j_vec))
          max(j_vec)
          else
            0L),
        dimnames = list(NULL, names_all)
      )

      list(
        X = X,
        colnames = names_all,
        meta = list(per_var = meta_per_var, per_col = per_col_meta)
      )
    },

    hal_prepare_new = function(DT_new, hal_obj) {
      stopifnot(requireNamespace("data.table", quietly = TRUE))
      stopifnot(requireNamespace("Matrix", quietly = TRUE))

      if (!data.table::is.data.table(DT_new))
        DT_new <- data.table::as.data.table(DT_new)

      vars <- names(hal_obj$meta$per_var)
      if (!all(vars %in% names(DT_new)))
        stop("DT_new is missing required variables")

      DT_local <- DT_new[, ..vars]

      if (!exists("interN_cpp", mode = "function", inherits = TRUE)) {
        stop("interN_cpp() not found. Ensure the C++ code is compiled and loaded.")
      }

      n <- nrow(DT_local)
      colnames_tr <- hal_obj$colnames
      colmeta <- hal_obj$meta$per_col
      if (length(colmeta) != length(colnames_tr))
        stop("hal_obj meta mismatch")

      # Stable key for numeric cutpoints (must be used consistently)
      key_num <- function(z)
        formatC(z, digits = 17, format = "fg")

      # --- coerce DT_local types using training meta_per_var ---
      for (v in vars) {
        per <- hal_obj$meta$per_var[[v]]

        if (!is.null(per$levels)) {
          # factor variables: enforce training levels
          x <- DT_local[[v]]

          if (is.character(x) || is.logical(x))
            x <- as.character(x)
          if (!is.factor(x))
            x <- as.character(x)

          lv_tr <- per$levels
          DT_local[[v]] <- factor(x, levels = lv_tr)

          if (anyNA(DT_local[[v]])) {
            bad <- unique(as.character(x[is.na(DT_local[[v]])]))
            stop(sprintf(
              "DT_new has unseen level(s) in %s: %s",
              v,
              paste(bad, collapse = ", ")
            ))
          }

        } else if (!is.null(per$cutpoints)) {
          DT_local[[v]] <- as.numeric(DT_local[[v]])
        } else {
          x <- DT_local[[v]]
          if (is.character(x) ||
              is.logical(x))
            DT_local[[v]] <- factor(x)
        }
      }
      # -------------------------------------------------------------

      # Precompute which primitives are needed (from per_col)
      need_num <- lapply(vars, function(v)
        numeric())
      names(need_num) <- vars
      need_fac <- lapply(vars, function(v)
        character())
      names(need_fac) <- vars

      for (m in colmeta) {
        if (m$type == "main") {
          if (m$kind == "numeric")
            need_num[[m$var]] <- unique(c(need_num[[m$var]], m$cutpoint))
          else
            need_fac[[m$var]] <- unique(c(need_fac[[m$var]], m$level))
        } else if (m$type == "interaction") {
          for (p in m$parts) {
            if (p$kind == "numeric")
              need_num[[p$var]] <- unique(c(need_num[[p$var]], p$cutpoint))
            else
              need_fac[[p$var]] <- unique(c(need_fac[[p$var]], p$level))
          }
        } else {
          stop("Unknown column meta type")
        }
      }

      # Build index maps: var -> key -> sorted row indices
      num_map <- vector("list", length(vars))
      names(num_map) <- vars
      fac_map <- vector("list", length(vars))
      names(fac_map) <- vars

      for (v in vars) {
        x <- DT_local[[v]]

        # numeric: STANDARD HAL uses I(x >= cut)
        if (is.numeric(x) && length(need_num[[v]])) {
          cuts <- sort(unique(need_num[[v]]))
          nn <- which(!is.na(x) & is.finite(x))

          if (length(nn)) {
            ord <- nn[order(x[nn])]
            xs <- x[ord]
            N <- length(ord)

            # k_lt[j] = number of xs < cuts[j]
            # (left.open=TRUE makes equality go to previous interval => strict <)
            k_lt <- findInterval(cuts, xs, left.open = TRUE)

            idxs <- lapply(k_lt, function(k) {
              start <- k + 1L
              if (start <= N)
                sort.int(ord[start:N], method = "quick")
              else
                integer()
            })

          } else {
            idxs <- lapply(cuts, function(.)
              integer())
          }

          names(idxs) <- key_num(cuts)
          num_map[[v]] <- idxs
        }

        # factor
        if (is.factor(x) && length(need_fac[[v]])) {
          chr <- as.character(x)
          lv_needed <- need_fac[[v]]
          idxs <- lapply(lv_needed, function(L)
            which(!is.na(chr) & chr == L))
          names(idxs) <- lv_needed
          fac_map[[v]] <- idxs
        }
      }

      # Assemble sparse triplets in training column order
      I_chunks <- vector("list", length(colmeta))
      J_chunks <- vector("list", length(colmeta))

      for (j in seq_along(colmeta)) {
        m <- colmeta[[j]]

        if (m$type == "main") {
          if (m$kind == "numeric") {
            idx <- num_map[[m$var]][[key_num(m$cutpoint)]]
          } else {
            idx <- fac_map[[m$var]][[m$level]]
          }

        } else {
          # interaction
          parts_idx <- vector("list", length(m$parts))
          ok <- TRUE

          for (k in seq_along(m$parts)) {
            p <- m$parts[[k]]
            if (p$kind == "numeric") {
              parts_idx[[k]] <- num_map[[p$var]][[key_num(p$cutpoint)]]
            } else {
              parts_idx[[k]] <- fac_map[[p$var]][[p$level]]
            }
            if (is.null(parts_idx[[k]]) ||
                !length(parts_idx[[k]])) {
              ok <- FALSE
              break
            }
          }

          if (!ok) {
            idx <- integer()
          } else if (length(parts_idx) == 1L) {
            idx <- parts_idx[[1L]]
          } else {
            idx <- interN_cpp(parts_idx)
          }
        }

        if (length(idx)) {
          I_chunks[[j]] <- idx
          J_chunks[[j]] <- rep.int(j, length(idx))
        } else {
          I_chunks[[j]] <- integer()
          J_chunks[[j]] <- integer()
        }
      }

      i_vec <- unlist(I_chunks, use.names = FALSE)
      j_vec <- unlist(J_chunks, use.names = FALSE)

      Matrix::sparseMatrix(
        i = if (length(i_vec))
          i_vec
        else
          integer(),
        j = if (length(j_vec))
          j_vec
        else
          integer(),
        x = if (length(i_vec))
          1L
        else
          integer(),
        dims = c(n, length(colnames_tr)),
        dimnames = list(NULL, colnames_tr)
      )
    },



    private_fit = function(data, ...) {

      fit <- tryCatch({

        data_copy <- data.table::copy(data)

        group_cols <- .self$covariates[complete.cases(.self$covariates)]

        data_copy <- data_copy[, .(tij = sum(tij), deltaij = sum(deltaij)), by = c(group_cols, "node", "k")]
        data_copy <- data_copy[complete.cases(data_copy), ]

        x_pp <- hal_basis(
          vars = c(group_cols, "node"),
          DT = data_copy,
          max_interaction = .self$max_degree,
          knots_per_order = .self$num_knots
        )

        if (nrow(x_pp$X) == 0L || ncol(x_pp$X) == 0L) {
          return(make_failed_fit("Empty HAL basis in Learner_hal::private_fit"))
        }

        pf <- 1 - grepl("^I\\(\\s*node\\s*==[^)]*\\)$", x_pp$colnames)

        if (.self$cross_validation) {
          prefit_args <- .self$fit_arguments

          if (!is.na(.self$maxit_prefit)) {
            prefit_args[["maxit"]] <- .self$maxit_prefit
          }

          cv_fit <- tryCatch(
            suppressWarnings(
              do.call(glmnet::cv.glmnet, c(
                list(
                  x = x_pp$X,
                  y = as.numeric(data_copy[["deltaij"]]),
                  offset = log(data_copy[["tij"]]),
                  penalty.factor = pf
                ),
                prefit_args
              ))
            ),
            error = function(e1) {
              tryCatch(
                suppressWarnings(
                  do.call(glmnet::cv.glmnet, c(
                    list(
                      x = x_pp$X,
                      y = as.numeric(data_copy[["deltaij"]]),
                      offset = log(data_copy[["tij"]]),
                      penalty.factor = rep(1, ncol(x_pp$X))
                    ),
                    prefit_args
                  ))
                ),
                error = function(e2) NULL
              )
            }
          )

          if (is.null(cv_fit) || length(cv_fit$lambda.min) == 0L || !is.finite(cv_fit$lambda.min)) {
            return(make_failed_fit("cv.glmnet failed in Learner_hal::private_fit"))
          }

          .self$lambda_opt <- as.numeric(cv_fit$lambda.min)

          attach_psl_fit_meta(
            cv_fit,
            meta = x_pp[c("colnames", "meta")]
          )

        } else {
          glmnet_args <- .self$fit_arguments
          glmnet_args[["penalty.factor"]] <- pf

          fit_obj <- suppressWarnings(
            do.call(glmnet::glmnet, c(
              glmnet_args,
              list(
                x = x_pp$X,
                y = as.numeric(data_copy[["deltaij"]]),
                offset = log(data_copy[["tij"]])
              )
            ))
          )

          attach_psl_fit_meta(
            fit_obj,
            meta = x_pp[c("colnames", "meta")]
          )
        }

      }, error = function(e) {
        make_failed_fit(conditionMessage(e))
      })

      fit
    },

    private_predictor = function(model, newdata, ...) {
      if (is_failed_fit(model)) {
        return(rep(NA_real_, nrow(newdata)))
      }

      fit_meta <- get_psl_fit_meta(model)

      if (is.null(fit_meta)) {
        return(rep(NA_real_, nrow(newdata)))
      }

      X_new <- tryCatch(
        .self$hal_prepare_new(newdata, fit_meta),
        error = function(e) NULL
      )

      if (is.null(X_new)) {
        return(rep(NA_real_, nrow(newdata)))
      }

      pred_args <- list(
        object = model,
        newx = X_new,
        newoffset = log(1),
        type = "response"
      )

      dots <- list(...)
      if (inherits(model, "cv.glmnet") && is.null(dots[["s"]])) {
        pred_args[["s"]] <- "lambda.min"
      }

      out <- tryCatch(
        do.call(predict, c(pred_args, dots)),
        error = function(e) rep(NA_real_, nrow(newdata))
      )

      as.numeric(out)
    }
  )
)


#' GAM learner via `mgcv::bam`
#'
#' `Learner_gam` is a Reference Class implementing the learner interface
#' used by [Superlearner()] and [fit_learner()].
#'
#' **User-facing API:** users should **only initialize** the learner and pass it
#' to [Superlearner()] / [fit_learner()]. The methods `private_fit()` and
#' `private_predictor()` are part of the internal learner interface and are
#' **not meant to be called directly by users**.
#'
#' **Wrapper role:** this class wraps `mgcv::bam` in a piecewise-constant hazard
#' workflow. The package-specific contribution is to provide a convenient
#' interface for the long-format Poisson likelihood with offsets for time at risk,
#' and optional node terms encoding the baseline hazard, while forwarding standard
#' `mgcv::bam` arguments supplied via `...`.
#'
#' @section Model:
#' Let \eqn{0=t_0 < t_1 < \cdots < t_m}{0=t0 < t1 < ... < tm} denote time knots and
#' define interval indicators \eqn{I_k(t)=1\{t\in(t_k,t_{k+1}]\}}{Ik(t)=I(t in (tk, t{k+1}])}.
#' The piecewise-constant hazard model with an additive predictor is
#' \deqn{
#'   \lambda(t \mid x) = \sum_{k=0}^{m} I_k(t)\,\exp\{\eta(x) + \gamma_k\}.
#' }{
#'   lambda(t|x) = sum_{k=0}^m Ik(t) * exp(eta(x) + gamma_k).
#' }
#' The additive predictor \eqn{\eta(x)}{eta(x)} is constructed from `covariates`
#' (smooth terms such as `s(age)` and/or linear terms) and estimated by `mgcv`.
#'
#' @param covariates `character`. Right-hand-side terms, including `mgcv` smooths
#'   (e.g. `"s(age)"`) and/or linear terms (e.g. `"value_LDL"`).
#' @param cross_validation `logical`. Included for compatibility with the learner
#'   interface; smoothing selection is controlled by `mgcv` and arguments in `...`.
#'
#' @section Fields:
#' \describe{
#'   \item{`covariates` (`character`)}{Terms used to build the additive predictor (may include `s()` terms).}
#'   \item{`cross_validation` (`logical`)}{Workflow flag; see Details.}
#'   \item{`intercept` (`logical`)}{Whether to include an intercept.}
#'   \item{`add_nodes` (`logical`)}{If `TRUE`, include interval ("node") effects encoding the baseline hazard.}
#'   \item{`formula` (`character`)}{Formula string passed to `mgcv::bam`.}
#'   \item{`learner` (`function`)}{Backend fitter (`mgcv::bam`).}
#'   \item{`fit_arguments` (`list`)}{Additional arguments forwarded to `mgcv::bam`.}
#' }
#'
#' @section Methods (internal learner interface):
#' \describe{
#'   \item{`initialize(...)`}{Construct and configure the learner. This is the only method users should call.}
#'   \item{`private_fit(data, ...)`}{Internal. Fits a Poisson GAM with offset `log(tij)` on long-format data.}
#'   \item{`private_predictor(model, newdata, ...)`}{Internal. Predicts hazards on the response scale.}
#' }
#'
#' @examples
#' lrn <- Learner_gam(covariates = c("s(age)", "value_LDL"))
#'
#' @export Learner_gam
#' @exportClass Learner_gam
Learner_gam <- setRefClass(
  "Learner_gam",
  fields = list(
    covariates = "character",
    cross_validation = "logical",
    intercept = "logical",
    formula = "character",
    learner = "function",
    add_nodes = "logical",
    fit_arguments = "list"
  ),
  methods = list(
    initialize = function(covariates = NULL,
                          cross_validation = FALSE,
                          intercept = TRUE,
                          add_nodes = TRUE,
                          ...) {
      .self$covariates <- covariates
      .self$cross_validation <- cross_validation
      .self$intercept <- intercept
      .self$add_nodes <- add_nodes

      .self$formula <- create_formula_gam(
        covariates = .self$covariates,
        intercept = .self$intercept,
        add_nodes = .self$add_nodes
      )

      .self$learner <- mgcv::bam

      .self$fit_arguments <- list(...)
      .self$fit_arguments[['family']] <- poisson()
    },
    private_fit = function(data, ...) {
      .extract_symbols <- function(term) {
        tryCatch(
          all.vars(str2lang(term)),
          error = function(e) {
            tryCatch(
              all.vars(stats::terms(stats::as.formula(paste("~", term)))),
              error = function(e2) character(0)
            )
          }
        )
      }

      fit <- tryCatch({

        group_cols <- .self$covariates[complete.cases(.self$covariates)]
        group_cols <- group_cols[is.character(group_cols) & !is.na(group_cols) & nzchar(group_cols)]
        group_cols <- unlist(lapply(group_cols, .extract_symbols), use.names = FALSE)
        group_cols <- unique(group_cols)

        data <- data[, .(tij = sum(tij), deltaij = sum(deltaij)), by = c(group_cols, "node", "k")]
        data <- data[complete.cases(data), ]

        .self$fit_arguments$formula <- as.formula(.self$formula)

        do.call(.self$learner, c(
          .self$fit_arguments,
          list(
            data = data,
            offset = log(data[["tij"]])
          )
        ))

      }, error = function(e) {
        make_failed_fit(conditionMessage(e))
      })

      fit
    },

    private_predictor = function(model, newdata, ...) {
      if (is_failed_fit(model)) {
        return(rep(NA_real_, nrow(newdata)))
      }

      pred <- tryCatch(
        predict(
          model,
          newdata = newdata[node %in% model$xlevels$node, ],
          type = "response",
          offset = log(1),
          ...
        ),
        error = function(e) NULL
      )

      if (is.null(pred)) {
        return(rep(NA_real_, nrow(newdata)))
      }

      if (all(levels(newdata$node) %in% model$xlevels$node)) {
        return(as.numeric(pred))
      } else {
        newdata[node %in% model$xlevels$node, predictions_model := pred]
        return(as.numeric(newdata$predictions_model))
      }
    }
  )
)

#' Simulate synthetic data inspired by the Steno Type-1 risk engine
#'
#' Generates baseline covariates and event times for CVD and censoring, with an
#' optional competing-risks setting, for examples, benchmarks and tests.
#'
#' The simulator uses a structural equation model (via `lava::lvm`) to generate
#' realistic correlations between covariates. Event times are then generated from
#' cause-specific Weibull proportional hazards models, where the linear predictor
#' depends on the simulated covariates (and scenario).
#'
#' @title Simulate time-to-event data for hypothetical type-1 diabetes patients
#' @param n `numeric(1)`. Number of subjects to simulate.
#' @param coefficient_age `numeric(1)`. Log-hazard coefficient for age in the CVD
#'   model (`time.event.1`).
#' @param coefficient_LDL `numeric(1)`. Log-hazard coefficient for LDL in the CVD
#'   model (`time.event.1`).
#' @param value_diabetis `numeric(1)`. Log-hazard coefficient for diabetes duration
#'   in the CVD model (`time.event.1`).
#' @param keep `character` or `NULL`. Optional subset of columns to retain. If
#'   supplied, only those columns are returned.
#' @param scenario `character(1)`. One of `"alpha"` or `"beta"`. Scenario `"beta"`
#'   modifies the CVD hazard by adding nonlinear hinge-squared terms in age and LDL.
#' @param competing_risks `logical(1)`. If `TRUE` and `scenario = "alpha"`, simulates
#'   two event causes (CVD and death without CVD). Otherwise simulates CVD vs censoring.
#'
#' @details
#' The following baseline covariates are generated (column name, type, interpretation):
#'
#' \describe{
#' \item{sex}{`factor`. Binary sex indicator (generated Bernoulli, then stored as factor).}
#' \item{age}{`numeric`. Age at baseline (years).}
#' \item{diabetes_duration}{`numeric`. Duration of diabetes at baseline (years).}
#' \item{value_SBP}{`numeric`. Systolic blood pressure (SBP).}
#' \item{value_LDL}{`numeric`. LDL cholesterol.}
#' \item{value_HBA1C}{`numeric`. HbA1c.}
#' \item{value_Albuminuria}{`factor` with levels `Normal`, `Micro`, `Macro`. Albuminuria category.}
#' \item{eGFR}{`numeric`. Estimated glomerular filtration rate, constructed from latent
#'   age-dependent log2 eGFR components (higher values indicate better kidney function).}
#' \item{value_Smoking}{`factor`. Smoking indicator (generated from a logistic model, then stored as factor).}
#' \item{value_Motion}{`factor`. Physical activity indicator (generated from a logistic model, then stored as factor).}
#' }
#'
#' Event time variables are generated from latent Weibull PH models:
#' `time.event.1` (CVD), `time.event.0` (censoring), and in scenario `"alpha"` also
#' `time.event.2` (death without prior CVD). These latent variables are used to
#' construct the observed outcome variables returned by the function (see below).
#'
#' @return A `data.table` with at least the following columns:
#'
#' \describe{
#' \item{id}{`integer`. Subject identifier (1, ..., `n`).}
#' \item{time_cvd}{`numeric`. Observed follow-up time (minimum of event and censoring times;
#'   also includes competing risk time if `competing_risks = TRUE` in scenario `"alpha"`).}
#' \item{status_cvd}{`integer`. Observed event status:
#'   `0` = censored, `1` = CVD, and if `competing_risks = TRUE` in scenario `"alpha"`,
#'   `2` = death without prior CVD.}
#' \item{time}{`numeric`. Alias of `time_cvd` (kept for convenience).}
#' \item{event}{`integer`. Alias of `status_cvd` (kept for convenience).}
#' \item{uncensored_time_cvd}{`numeric`. Event time ignoring censoring
#'   (minimum of event causes only).}
#' \item{uncensored_status_cvd}{`integer`. Event cause ignoring censoring.
#'   In scenario `"alpha"` this is `1` (CVD) or `2` (death without CVD);
#'   in scenario `"beta"` this is always `1`.}
#' \item{uncensored_time}{`numeric`. Alias of `uncensored_time_cvd`.}
#' \item{uncensored_event}{`integer`. Alias of `uncensored_status_cvd`.}
#' }
#'
#' In addition, the returned table contains all baseline covariates listed in
#' \strong{Details}. Internal latent variables used only for simulation are removed
#' before returning (e.g., log2 eGFR components and, in scenario `"beta"`, the
#' hinge-squared features).
#'
#' @examples
#' simulateStenoT1(n = 20, scenario = "alpha", competing_risks = TRUE)
#' @author Thomas A. Gerds <tag@@biostat.ku.dk>
#' @export
simulateStenoT1 <- function(
    n,
    coefficient_age = 0.05,
    coefficient_LDL = 0.10,
    value_diabetis = 0.02,
    keep = NULL,
    scenario = c("alpha", "beta"),
    competing_risks = FALSE
) {
  scenario <- match.arg(scenario)

  # ------------------------------------------------------------------
  # Shared model parts (covariates + structural equations)
  # ------------------------------------------------------------------
  sim_model <- lava::lvm()

  # covariates
  lava::distribution(sim_model, ~sex) <- lava::binomial.lvm(p = 0.462559634930512)
  lava::distribution(sim_model, ~age) <- lava::normal.lvm(mean = 42.2853710165578, sd = 16.1358967048911)
  # affects the age range
  lava::covariance(sim_model, ~age) <- 40
  lava::distribution(sim_model, ~diabetes_duration) <- lava::normal.lvm(mean = -0.721917658912036, sd = 11.8217218392295)
  lava::distribution(sim_model, ~value_SBP) <- lava::normal.lvm(mean = 113.015053986869, sd = 15.1994730394504)
  lava::distribution(sim_model, ~value_LDL) <- lava::normal.lvm(mean = 2.34380835266364, sd = 0.78907040283751)
  lava::distribution(sim_model, ~value_HBA1C) <- lava::normal.lvm(mean = 66.1642378865059, sd = 15.1450866080122)
  lava::distribution(sim_model, ~log2_eGFR_young) <- lava::normal.lvm(mean = -10.7725954841897, sd = 1.85777815372659)
  lava::distribution(sim_model, ~log2_eGFR_old) <- lava::normal.lvm(mean = -7.16808301118136, sd = 0.295067430820889)
  lava::distribution(sim_model, ~value_Smoking) <- lava::binomial.lvm(p = 0.00406502756222992)
  lava::distribution(sim_model, ~value_Motion) <- lava::binomial.lvm(p = 0.647591503001192)

  # albuminuria
  sim_model <- lava::categorical(
    sim_model, ~value_Albuminuria,
    labels = c("Normal", "Micro", "Macro"),
    K = 3, p = c(0.839037544077992, 0.121344119477287)
  )
  lava::transform(sim_model, value_AlbuminuriaMicro ~ value_Albuminuria) <- function(x) {
    1 * (c(x) == "Micro")
  }
  lava::transform(sim_model, value_AlbuminuriaMacro ~ value_Albuminuria) <- function(x) {
    1 * (c(x) == "Macro")
  }

  # structural equations (shared)
  lava::regression(sim_model, age ~ sex) <- c(-0.105030442652732)
  lava::regression(sim_model, diabetes_duration ~ age + sex) <- c(0.461453096059993, 1.8834280372868)
  lava::regression(sim_model, value_SBP ~ diabetes_duration + age + sex) <- c(0.173711840258398, 0.408090766933998, -5.26413576607228)
  lava::regression(sim_model, value_LDL ~ value_SBP + diabetes_duration + age + sex) <- c(
    0.0017997232707461, -0.00215485947423231, -0.00194439878409113, -0.0247941295129788
  )
  lava::regression(sim_model, value_HBA1C ~ value_LDL + value_SBP + diabetes_duration + age + sex) <- c(
    2.41425803932559, -0.0366520181671025, 0.105793793763282, -0.113517351234731, 1.14432850532257
  )

  lava::regression(
    sim_model,
    log2_eGFR_young ~ value_AlbuminuriaMicro + value_AlbuminuriaMacro + value_HBA1C +
      value_LDL + value_SBP + diabetes_duration + age + sex
  ) <- c(
    -0.0378605757282762, 0.369860071363398, -0.000381556996934381, 0.0759866855173213,
    -0.000921667597434942, 0.0109443806788427, 0.173985375155216, -0.05148192662188
  )

  lava::regression(
    sim_model,
    log2_eGFR_old ~ log2_eGFR_young + value_AlbuminuriaMicro + value_AlbuminuriaMacro +
      value_HBA1C + value_LDL + value_SBP + diabetes_duration + age + sex
  ) <- c(
    -0.998329448050345, 0.13212871605835, 0.627054487201391, -0.0017321066682523,
    -0.0168673394900242, 0.000656653716889479, 0.00231859330125351, 0.0105241259989393,
    0.0666215469676934
  )

  lava::regression(
    sim_model,
    value_Smoking ~ log2_eGFR_old + log2_eGFR_young + value_AlbuminuriaMicro +
      value_AlbuminuriaMacro + value_HBA1C + value_LDL + value_SBP +
      diabetes_duration + age + sex
  ) <- c(
    -0.573224437209381, -0.533358383827049, 0.417778953216348, 0.811962094950751,
    0.0258095967578417, 0.128112260674147, -0.0102291702067243, -0.00575108475194562,
    0.00529875681026287, -0.130756787272673
  )

  lava::regression(
    sim_model,
    value_Motion ~ value_Smoking + log2_eGFR_old + log2_eGFR_young +
      value_AlbuminuriaMicro + value_AlbuminuriaMacro + value_HBA1C +
      value_LDL + value_SBP + diabetes_duration + age + sex
  ) <- c(
    -0.31142087785739, -0.134144429182559, -0.0862952685994639, -0.121332306983678,
    -0.334529806176446, -0.0100110418345244, -0.138145721076664, 0.00690508520692886,
    -0.0080292376964733, -0.00517733988644732, 0.070748099701443
  )

  # construction of eGFR
  lava::transform(sim_model, eGFR ~ log2_eGFR_old + log2_eGFR_young + age) <- function(x) {
    as.numeric(x[["age"]] < 40) * 100 * 2^(x[["log2_eGFR_young"]]) +
      as.numeric(x[["age"]] > 40) * 100 * 2^(x[["log2_eGFR_old"]])
  }

  # ------------------------------------------------------------------
  # Outcome model (scenario-specific)
  # ------------------------------------------------------------------
  # latent event times: .1 is CVD, .2 is death without CVD (alpha only), .0 is censoring
  lava::distribution(sim_model, ~time.event.1) <- lava::coxWeibull.lvm(
    scale = 0.00344127508109558,
    shape = 1.1383691898171
  )
  lava::distribution(sim_model, ~time.event.0) <- lava::coxWeibull.lvm(
    scale = 0.0105086718102263,
    shape = 2.2578792372401
  )

  if (scenario == "alpha") {
    lava::distribution(sim_model, ~time.event.2) <- lava::coxWeibull.lvm(
      scale = 8.8338885667654e-05,
      shape = 1.49710154034773
    )
  }

  # censoring model (shared)
  lava::regression(
    sim_model,
    time.event.0 ~ value_Motion + value_Smoking + log2_eGFR_old + log2_eGFR_young +
      value_AlbuminuriaMicro + value_AlbuminuriaMacro + value_HBA1C + value_LDL +
      value_SBP + diabetes_duration + age + sex
  ) <- c(
    -0.0389267779620385, -0.106325304898881, -0.137101392329058, -0.118521604228187,
    -0.0330093143101979, 0.0763711547050245, -0.00617576911315076, 0.0509424048213361,
    -0.00757123095797519, -0.0125850223581346, -0.0130146695831196, -0.029633759327063
  )

  if (scenario == "alpha") {
    # event 1 hazard (linear)
    lava::regression(
      sim_model,
      time.event.1 ~ value_Motion + value_Smoking + log2_eGFR_old + log2_eGFR_young +
        value_AlbuminuriaMicro + value_AlbuminuriaMacro + value_HBA1C + value_LDL +
        value_SBP + diabetes_duration + age + sex
    ) <- c(
      -0.176043482185139, 0.311354666798778, 0.0442184803104845, 0.480004039250349,
      0.0458766977417934, 0.0760940706393913, 0.0143694959616522, coefficient_LDL,
      0.00607734573133048, value_diabetis, coefficient_age, -0.224377188536924
    )

    # event 2 hazard (death)
    lava::regression(
      sim_model,
      time.event.2 ~ value_Motion + value_Smoking + log2_eGFR_old + log2_eGFR_young +
        value_AlbuminuriaMicro + value_AlbuminuriaMacro + value_HBA1C + value_LDL +
        value_SBP + diabetes_duration + age + sex
    ) <- c(
      -0.410553409305998, 0.833815403131808, 0.3289175620287, 0.313118375461698,
      0.225523885897843, 0.604004891420425, 0.0102472549995989, 0.09,
      0.00156423178426314, 0.06, 0.09, -0.276186036048047
    )
  }

  if (scenario == "beta") {
    # Nonlinear terms for event 1 hazard
    lava::transform(sim_model, age_hinge55_sq ~ age) <- function(x) {
      a <- x[["age"]]
      (pmax(a - 55, 0)^2) / 100
    }
    lava::transform(sim_model, LDL_hinge3_sq ~ value_LDL) <- function(x) {
      ldl <- x[["value_LDL"]]
      (pmax(ldl - 3.0, 0)^2)
    }
    nl_age_strength <- 1
    nl_ldl_strength <- 1.5

    lava::regression(
      sim_model,
      time.event.1 ~ value_Motion + value_Smoking + log2_eGFR_old + log2_eGFR_young +
        value_AlbuminuriaMicro + value_AlbuminuriaMacro + value_HBA1C + value_LDL +
        value_SBP + diabetes_duration + age + sex +
        age_hinge55_sq + LDL_hinge3_sq
    ) <- c(
      -0.176043482185139, 0.311354666798778, 0.0442184803104845, 0.480004039250349,
      0.0458766977417934, 0.0760940706393913, 0.0143694959616522, coefficient_LDL,
      0.00607734573133048, value_diabetis, coefficient_age, -0.224377188536924,
      nl_age_strength, nl_ldl_strength
    )
  }

  # ------------------------------------------------------------------
  # Simulate and post-process
  # ------------------------------------------------------------------
  d <- lava::sim(sim_model, n)
  data.table::setDT(d)

  # drop internal variables
  d[, log2_eGFR_old := NULL]
  d[, log2_eGFR_young := NULL]
  d[, value_AlbuminuriaMicro := NULL]
  d[, value_AlbuminuriaMacro := NULL]

  if (scenario == "beta") {
    d[, age_hinge55_sq := NULL]
    d[, LDL_hinge3_sq := NULL]
  }

  # observed time/event
  if (scenario == "alpha" && isTRUE(competing_risks)) {
    d[, time_cvd := pmin(time.event.0, time.event.1, time.event.2)]
    d[, status_cvd := data.table::fifelse(time.event.1 < time.event.0 & time.event.1 <= time.event.2, 1,
                                          data.table::fifelse(time.event.2 < time.event.0 & time.event.2 < time.event.1, 2, 0)
    )]
  } else {
    d[, time_cvd := pmin(time.event.0, time.event.1)]
    d[, status_cvd := as.numeric(time.event.1 < time.event.0)]
  }

  # uncensored event time/status
  if (scenario == "alpha") {
    d[, uncensored_time_cvd := pmin(time.event.1, time.event.2)]
    d[, uncensored_status_cvd := as.numeric(time.event.2 < time.event.1) + 1L]
  } else {
    d[, uncensored_time_cvd := time.event.1]
    d[, uncensored_status_cvd := 1L]
  }

  # keep current simulator output names (time/event)
  d[, time := time_cvd]
  d[, event := status_cvd]
  d[, uncensored_time := uncensored_time_cvd]
  d[, uncensored_event := uncensored_status_cvd]

  # beta script drops time.event.*
  if (scenario == "beta") {
    d[, time.event.0 := NULL]
    d[, time.event.1 := NULL]
  }

  # id + ordering
  data.table::setorder(d, time, -event)
  d[, id := 1:.N]
  data.table::setcolorder(d, "id")

  # factors
  d[, c("sex", "value_Smoking", "value_Motion") := lapply(.SD, as.factor),
    .SDcols = c("sex", "value_Smoking", "value_Motion")
  ]

  if (length(keep) > 0) {
    return(d[, keep, with = FALSE][])
  }
  d[]
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

poisson_deviance_by_folder_cols <- function(log_hazard_cols, tij, delta, folder, nfold, eps = 1e-15) {
    .Call(`_poissonsuperlearner_poisson_deviance_by_folder_cols`, log_hazard_cols, tij, delta, folder, nfold, eps)
}

inter2_cpp <- function(a, b) {
    .Call(`_poissonsuperlearner_inter2_cpp`, a, b)
}

interN_cpp <- function(lst) {
    .Call(`_poissonsuperlearner_interN_cpp`, lst)
}

mk_main_numeric_cpp <- function(x, K) {
    .Call(`_poissonsuperlearner_mk_main_numeric_cpp`, x, K)
}

mk_main_factor_cpp <- function(x_int, n_levels) {
    .Call(`_poissonsuperlearner_mk_main_factor_cpp`, x_int, n_levels)
}

add_cols_cpp <- function(idxs_list, p_start) {
    .Call(`_poissonsuperlearner_add_cols_cpp`, idxs_list, p_start)
}

pch_absolute_risk <- function(id, dt, haz, cause_idx, one_based = TRUE, na_is_zero = FALSE) {
    .Call(`_poissonsuperlearner_pch_absolute_risk`, id, dt, haz, cause_idx, one_based, na_is_zero)
}

pch_absolute_risk_euler <- function(id, dt, haz, cause_idx, one_based = TRUE, na_is_zero = FALSE) {
    .Call(`_poissonsuperlearner_pch_absolute_risk_euler`, id, dt, haz, cause_idx, one_based, na_is_zero)
}

pch_survival <- function(id, dt, haz, na_is_zero = FALSE) {
    .Call(`_poissonsuperlearner_pch_survival`, id, dt, haz, na_is_zero)
}

poisson_deviance_by_folder <- function(log_hazard, tij, delta, folder, F, eps = 1e-15) {
    .Call(`_poissonsuperlearner_poisson_deviance_by_folder`, log_hazard, tij, delta, folder, F, eps)
}

#' @keywords internal
"_PACKAGE"

#' @importFrom riskRegression predictRisk
#' @importFrom methods new
#' @importFrom stats coef deviance predict quantile relevel time
#' @importFrom utils head tail
#' @importFrom glmnet glmnet cv.glmnet
#' @importFrom mgcv gam
#' @importFrom Matrix sparse.model.matrix
#' @import data.table
#' @import sampling
NULL
## usethis namespace: start
#' @useDynLib poissonsuperlearner, .registration = TRUE
## usethis namespace: end
NULL

---
title: "Basic Usage"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Basic Usage}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(poissonsuperlearner)
library(riskRegression)
```

# Introduction

This vignette gives fast examples of the Poisson Super Learner workflow after
the refactoring. It focuses on:

* fitting `Superlearner()` with one learner library shared by all causes;
* fitting `Superlearner()` with one learner library per cause;
* using `summary()` for the fitted super learner;
* using `predictRisk()` for the same model selectors.

The examples use small simulated data, two folds, and simple `glmnet` learners
with fixed `lambda` values so the vignette remains quick to run during package
checks.

# Data

We simulate a small competing-risks data set. The observed follow-up time is
stored in `time` and the event indicator in `event`, where `0` denotes
censoring, `1` denotes cardiovascular disease, and `2` denotes death without
prior cardiovascular disease.

```{r}
d <- simulateStenoT1(
  n = 45,
  scenario = "alpha",
  competing_risks = TRUE,
  seed = 1
)

d <- d[, .(
  id,
  time,
  event,
  sex,
  age,
  diabetes_duration,
  value_LDL,
  value_Smoking
)]

head(d)
```

# One Learner Library For All Causes

A learner library is a list of initialized learner objects. If a single library
is supplied in a competing-risks analysis, the same library is used for all
causes.

```{r}
shared_library <- list(
  simple = Learner_glmnet(
    covariates = c("sex", "diabetes_duration"),
    cross_validation = FALSE,
    lambda = 0
  ),
  shrink = Learner_glmnet(
    covariates = c("sex", "age", "value_LDL"),
    cross_validation = FALSE,
    lambda = 0.05,
    alpha = 1
  )
)

fit_shared <- Superlearner(
  data = d,
  id = "id",
  status = "event",
  event_time = "time",
  learners = shared_library,
  number_of_nodes = 3,
  nfold = 2
)
```

`summary()` gives a compact overview of the fitted super learner, including the
number of causes, retained learner labels, cross-validated deviances, and
meta-learner coefficients when a meta-learner was fitted.

```{r}
summary(fit_shared)
```

`predictRisk()` returns one row per subject and one column per requested
prediction time. The `model` argument uses the same selectors: `"sl"` for the
stacked ensemble, `"discrete_sl"` for the best cross-validated base learner per
cause, or learner labels such as `"simple"` and `"shrink"` for models stored in
the ensemble.

```{r}
newdata <- d[1:2]
times <- c(1, 2)

risk_shared_sl <- predictRisk(
  fit_shared, newdata = newdata, times = times, cause = 1, model = "sl"
)

risk_shared_discrete <- predictRisk(
  fit_shared, newdata = newdata, times = times, cause = 1, model = "discrete_sl"
)

risk_shared_simple <- predictRisk(
  fit_shared, newdata = newdata, times = times, cause = 1, model = "simple"
)

risk_shared_shrink <- predictRisk(
  fit_shared, newdata = newdata, times = times, cause = 1, model = "shrink"
)

list(
  sl = risk_shared_sl,
  discrete_sl = risk_shared_discrete,
  simple = risk_shared_simple,
  shrink = risk_shared_shrink
)
```

# One Learner Library Per Cause

For competing risks, `learners` can also be a list with one learner library per
cause. This allows different covariates, tuning parameters, or labels for each
cause.

```{r}
libraries_by_cause <- list(
  cvd = list(
    cvd_simple = Learner_glmnet(
      covariates = c("sex", "diabetes_duration"),
      cross_validation = FALSE,
      lambda = 0
    ),
    cvd_shrink = Learner_glmnet(
      covariates = c("age", "value_LDL"),
      cross_validation = FALSE,
      lambda = 0.05,
      alpha = 1
    )
  ),
  death = list(
    death_simple = Learner_glmnet(
      covariates = c("sex", "age"),
      cross_validation = FALSE,
      lambda = 0
    ),
    death_shrink = Learner_glmnet(
      covariates = c("diabetes_duration", "value_Smoking"),
      cross_validation = FALSE,
      lambda = 0.05,
      alpha = 1
    )
  )
)

fit_by_cause <- Superlearner(
  data = d,
  id = "id",
  status = "event",
  event_time = "time",
  learners = libraries_by_cause,
  number_of_nodes = 3,
  nfold = 2
)
```

The fitted object can be summarized in the same way.

```{r}
summary(fit_by_cause)
```

The stacked and discrete super learner selectors still work as scalar model
selectors for prediction.

```{r}
risk_by_cause_sl <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1, model = "sl"
)

risk_by_cause_discrete <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1, model = "discrete_sl"
)

list(
  sl = risk_by_cause_sl,
  discrete_sl = risk_by_cause_discrete
)
```

When selecting learners from cause-specific libraries, provide one selector per
cause. The first entry selects the learner for cause 1 and the second entry
selects the learner for cause 2.

```{r}
cause_specific_model <- c("cvd_simple", "death_shrink")
cause_specific_model_alt <- c("cvd_shrink", "death_simple")

risk_by_cause_selected <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1,
  model = cause_specific_model
)

risk_by_cause_selected_alt <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1,
  model = cause_specific_model_alt
)

list(
  selected_learners = risk_by_cause_selected,
  selected_learners_alt = risk_by_cause_selected_alt
)
```

Integer selectors are also supported. For cause-specific libraries, an integer
vector can choose different retained learner positions for different causes.

```{r}
risk_by_cause_indexed <- predictRisk(
  fit_by_cause, newdata = newdata, times = times, cause = 1, model = c(1, 2)
)

risk_by_cause_indexed
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/learners.R
\docType{class}
\name{Learner_gam-class}
\alias{Learner_gam-class}
\alias{Learner_gam}
\title{GAM learner via \code{mgcv::bam}}
\arguments{
\item{covariates}{\code{character}. Right-hand-side terms, including \code{mgcv} smooths
(e.g. \code{"s(age)"}) and/or linear terms (e.g. \code{"value_LDL"}).}

\item{cross_validation}{\code{logical}. Included for compatibility with the learner
interface; smoothing selection is controlled by \code{mgcv} and arguments in \code{...}.}
}
\description{
\code{Learner_gam} is a Reference Class implementing the learner interface
used by \code{\link[=Superlearner]{Superlearner()}} and \code{\link[=fit_learner]{fit_learner()}}.
}
\details{
\strong{User-facing API:} users should \strong{only initialize} the learner and pass it
to \code{\link[=Superlearner]{Superlearner()}} / \code{\link[=fit_learner]{fit_learner()}}. The methods \code{private_fit()} and
\code{private_predictor()} are part of the internal learner interface and are
\strong{not meant to be called directly by users}.

\strong{Wrapper role:} this class wraps \code{mgcv::bam} in a piecewise-constant hazard
workflow. The package-specific contribution is to provide a convenient
interface for the long-format Poisson likelihood with offsets for time at risk,
and optional node terms encoding the baseline hazard, while forwarding standard
\code{mgcv::bam} arguments supplied via \code{...}.
}

\section{Model}{

Let \eqn{0=t_0 < t_1 < \cdots < t_m}{0=t0 < t1 < ... < tm} denote time knots and
define interval indicators \eqn{I_k(t)=1\{t\in(t_k,t_{k+1}]\}}{Ik(t)=I(t in (tk, t{k+1}])}.
The piecewise-constant hazard model with an additive predictor is
\deqn{
  \lambda(t \mid x) = \sum_{k=0}^{m} I_k(t)\,\exp\{\eta(x) + \gamma_k\}.
}{
  lambda(t|x) = sum_{k=0}^m Ik(t) * exp(eta(x) + gamma_k).
}
The additive predictor \eqn{\eta(x)}{eta(x)} is constructed from \code{covariates}
(smooth terms such as \code{s(age)} and/or linear terms) and estimated by \code{mgcv}.
}

\section{Fields}{

\describe{
\item{\code{covariates} (\code{character})}{Terms used to build the additive predictor (may include \code{s()} terms).}
\item{\code{cross_validation} (\code{logical})}{Workflow flag; see Details.}
\item{\code{intercept} (\code{logical})}{Whether to include an intercept.}
\item{\code{add_nodes} (\code{logical})}{If \code{TRUE}, include interval ("node") effects encoding the baseline hazard.}
\item{\code{formula} (\code{character})}{Formula string passed to \code{mgcv::bam}.}
\item{\code{learner} (\code{function})}{Backend fitter (\code{mgcv::bam}).}
\item{\code{fit_arguments} (\code{list})}{Additional arguments forwarded to \code{mgcv::bam}.}
}
}

\section{Methods (internal learner interface)}{

\describe{
\item{\code{initialize(...)}}{Construct and configure the learner. This is the only method users should call.}
\item{\code{private_fit(data, ...)}}{Internal. Fits a Poisson GAM with offset \code{log(tij)} on long-format data.}
\item{\code{private_predictor(model, newdata, ...)}}{Internal. Predicts hazards on the response scale.}
}
}

\examples{
lrn <- Learner_gam(covariates = c("s(age)", "value_LDL"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coef.R
\name{coef.base_learner}
\alias{coef.base_learner}
\title{Extract coefficients from a fitted base learner}
\usage{
\method{coef}{base_learner}(object, cause = NULL, ...)
}
\arguments{
\item{object}{\code{base_learner}. A fitted object returned by \code{\link[=fit_learner]{fit_learner()}}.}

\item{cause}{\code{numeric(1)} or \code{NULL}. Which cause to extract coefficients for.
If \code{NULL}, coefficients are returned for all causes.
Causes are indexed \verb{1, 2, ..., object$data_info$n_crisks} (with \code{0} reserved for censoring).}

\item{...}{Passed to the underlying \code{coef()} method of the fitted learner object
(learner-dependent; e.g., \code{s} for \code{glmnet}).}
}
\value{
If \code{cause} is a single integer, returns the coefficient object produced by
\code{coef()} for that cause-specific fitted model.

If \code{cause = NULL}, returns a \code{list} of length \code{object$data_info$n_crisks},
where element \verb{[[k]]} contains coefficients for cause \code{k}.

If no fitted model is present (\code{object$learner_fit} is \code{NULL}), signals a message
and returns \code{invisible(object)}.
}
\description{
Convenience method to extract (cause-specific) model coefficients from a fitted
\code{base_learner} returned by \code{\link[=fit_learner]{fit_learner()}}.
}
\details{
For competing risks, \code{fit_learner()} fits one model per cause, stored in
\code{object$learner_fit[[k]]} for \verb{k = 1, 2, ..., K}. This method simply dispatches
to the underlying model’s \code{coef()} method for each fitted object.

\strong{Learner-dependent output.} The returned coefficient object depends on the
base learner used (e.g. a numeric vector, a sparse matrix, a list, etc.).
This method does not post-process or rename coefficients; it returns the output
of \code{coef(object$learner_fit[[k]], ...)} unchanged.
}
\examples{
d <- simulateStenoT1(50, competing_risks = TRUE)
lrn <- Learner_glmnet(covariates = c("age", "value_LDL"),
                      lambda = 0, cross_validation = FALSE)
bl <- fit_learner(d, learner = lrn, id = "id",
                  status = "status_cvd", event_time = "time_cvd",
                  number_of_nodes = 4)

# coefficients for cause 1
coef(bl, cause = 1)

# coefficients for all causes (list)
coef(bl)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/learners.R
\docType{class}
\name{Learner_hal-class}
\alias{Learner_hal-class}
\alias{Learner_hal}
\title{HAL learner for piecewise Poisson hazards}
\description{
\code{Learner_hal} is a Reference Class implementing the learner interface
used by \code{\link[=Superlearner]{Superlearner()}} and \code{\link[=fit_learner]{fit_learner()}}.
}
\details{
\strong{User-facing API:} users should \strong{only initialize} the learner and pass it
to \code{\link[=Superlearner]{Superlearner()}} / \code{\link[=fit_learner]{fit_learner()}}. The methods \code{private_fit()} and
\code{private_predictor()} (and any basis-construction helpers) are part of the
internal learner interface and are \strong{not meant to be called directly by users}.

\strong{Wrapper role:} this class provides a piecewise-constant hazard wrapper around
a HAL-style indicator-basis construction, estimated by L1-penalized Poisson
regression using a \code{glmnet} backend. The package-specific contribution is to
(i) construct the long-format Poisson representation with offsets for time at
risk, (ii) generate indicator bases compatible with piecewise hazards, and
(iii) forward backend fitting arguments supplied via \code{...}.
}

\section{Model}{

Let \eqn{0=t_0 < t_1 < \cdots < t_m}{0=t0 < t1 < ... < tm} denote time knots and
define interval indicators \eqn{I_k(t)=1\{t\in(t_k,t_{k+1}]\}}{Ik(t)=I(t in (tk, t{k+1}])}.
The HAL piecewise-constant hazard model is
\deqn{
  \lambda(t \mid x) = \sum_{k=0}^{m} I_k(t)\,\exp\{f(t,x)\},
}{
  lambda(t|x) = sum_{k=0}^m Ik(t) * exp(f(t,x)).
}
where \eqn{f(t,x)}{f(t,x)} is approximated by a finite linear combination of
indicator basis functions.
}

\section{Two-covariate illustration}{

Let \eqn{x=(x_1,x_2)}{x=(x1,x2)} be two covariates and let
\eqn{t_0 < t_1 < \cdots < t_R}{t0 < t1 < ... < tR} be time grid points used to
create step functions in time. Choose covariate cutpoints
\eqn{c_{1,1},\ldots,c_{1,K_1}}{c1_1,...,c1_K1} for \eqn{x_1}{x1} and
\eqn{c_{2,1},\ldots,c_{2,K_2}}{c2_1,...,c2_K2} for \eqn{x_2}{x2}.

Define indicator bases:
\deqn{B_r(t) = 1\{t_r \le t\}}{Br(t) = I(tr <= t),}
\deqn{B_{1,p}(x) = 1\{c_{1,p} \le x_1\}}{B1p(x) = I(c1p <= x1),}
\deqn{B_{2,q}(x) = 1\{c_{2,q} \le x_2\}}{B2q(x) = I(c2q <= x2).}

A main-effects HAL approximation on the log-hazard scale can be written as:
\deqn{
  f_\beta(t,x) = \beta_0
  + \sum_{r=1}^R \beta_r B_r(t)
  + \sum_{r=1}^R\sum_{p=1}^{K_1} \beta_{r,1,p} B_r(t) B_{1,p}(x)
  + \sum_{r=1}^R\sum_{q=1}^{K_2} \beta_{r,2,q} B_r(t) B_{2,q}(x).
}{
  f_beta(t,x) = beta0
  + sum_{r=1}^R beta_r Br(t)
  + sum_{r=1}^R sum_{p=1}^K1 beta_{r,1,p} Br(t) B1p(x)
  + sum_{r=1}^R sum_{q=1}^K2 beta_{r,2,q} Br(t) B2q(x).
}
If \code{max_degree >= 2}, the learner additionally includes interaction bases such as
\deqn{
  \sum_{r=1}^R\sum_{p=1}^{K_1}\sum_{q=1}^{K_2}
  \beta_{r,12,pq} B_r(t) B_{1,p}(x) B_{2,q}(x).
}{
  sum_{r=1}^R sum_{p=1}^K1 sum_{q=1}^K2 beta_{r,12,pq} Br(t) B1p(x) B2q(x).
}
}

\section{How reference class parameters map to the model}{

\describe{
\item{\code{covariates}}{Covariate columns used to build covariate indicator bases.}
\item{\code{num_knots}}{Controls the number of cutpoints per covariate used for indicator bases.}
\item{\code{max_degree}}{Maximum interaction order included in the basis expansion.}
\item{\code{add_nodes}}{If \code{TRUE}, includes interval ("node") structure for the baseline hazard.}
\item{\code{intercept}}{Whether the backend penalized regression includes an intercept term.}
\item{\code{cross_validation}}{If \code{TRUE}, selects the penalty level using \code{glmnet::cv.glmnet}.}
\item{\code{fit_arguments}}{Additional arguments forwarded to the \code{glmnet} backend (e.g. \code{nfolds}).}
}
}

\section{Fields}{

\describe{
\item{\code{covariates} (\code{character})}{Names of covariate columns used in the basis.}
\item{\code{cross_validation} (\code{logical})}{Whether to use \code{cv.glmnet} to select the penalty.}
\item{\code{intercept} (\code{logical})}{Backend intercept flag.}
\item{\code{add_nodes} (\code{logical})}{Whether node (time-interval) effects are included.}
\item{\code{max_degree} (\code{integer})}{Maximum interaction order.}
\item{\code{num_knots} (\code{numeric})}{Knots used for basis construction.}
\item{\code{lambda_opt} (\code{numeric})}{Selected penalty level when using cross-validation.}
\item{\code{fit_arguments} (\code{list})}{Extra backend arguments forwarded to \code{glmnet}.}
}
}

\section{Methods (internal learner interface)}{

\describe{
\item{\code{initialize(...)}}{Construct and configure the learner. This is the only method users should call.}
\item{\code{private_fit(data, ...)}}{Internal. Builds bases and fits the penalized Poisson model with offset \code{log(tij)}.}
\item{\code{private_predictor(model, newdata, ...)}}{Internal. Evaluates the fitted approximation and returns hazards on the response scale.}
}
}

\examples{
lrn <- Learner_hal(covariates = c("age", "sex"), max_degree = 2L, num_knots = c(10L, 5L))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poissonsuperlearner-package.R
\docType{package}
\name{poissonsuperlearner-package}
\alias{poissonsuperlearner}
\alias{poissonsuperlearner-package}
\title{poissonsuperlearner: Poisson Super Learner}
\description{
Provides tools for fitting piece-wise constant hazard models for survival and competing risks data, including ensemble hazard estimation via the Super Learner framework. The package supports estimation of survival functions and absolute risk predictions from fitted cause-specific hazard models. For the Super Learner framework see van der Laan, Polley and Hubbard (2007) \doi{10.2202/1544-6115.1309}.
}
\author{
\strong{Maintainer}: Gabriele Pittarello \email{gabriele.pittarello@sund.ku.dk}

Authors:
\itemize{
  \item Helene Rytgaard \email{hely@sund.ku.dk}
  \item Thomas Gerds \email{tag@biostat.ku.dk}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.base_learner}
\alias{print.base_learner}
\title{Print method for \code{base_learner}}
\usage{
\method{print}{base_learner}(x, cause = 1, ...)
}
\arguments{
\item{x}{\code{base_learner} object returned by \code{\link[=fit_learner]{fit_learner()}}.}

\item{cause}{\code{numeric(1)} or \code{NULL}. Which cause to print the fitted model for.
If \code{NULL}, prints one line per cause (classes only) instead of printing the full
fitted objects.}

\item{...}{Passed to the underlying fitted object \code{print()} method when \code{cause}
is a single integer.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Prints a compact description of the fitted base learner, including the learner
type, the time-grid used, and (optionally) the fitted model object for a given
cause.
}
\examples{
d <- simulateStenoT1(30, competing_risks = TRUE)
lrn <- Learner_glmnet(
  covariates = c("sex", "value_LDL"),
  lambda = 0.01,
  cross_validation = FALSE
)
bl <- fit_learner(
  d,
  learner = lrn,
  id = "id",
  status = "status_cvd",
  event_time = "time_cvd",
  number_of_nodes = 3
)
print(bl, cause = NULL)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pch_survival.R
\name{pch_survival}
\alias{pch_survival}
\title{Piecewise-constant hazards survival function}
\usage{
pch_survival(id, dt, haz, na_is_zero = FALSE)
}
\arguments{
\item{id}{Integer vector of subject IDs, sorted by id then time.}

\item{dt}{Numeric vector of interval lengths.}

\item{haz}{Numeric matrix (n x C) of cause-specific hazards.}

\item{na_is_zero}{Logical. If TRUE, treat NA hazards as zero.}
}
\value{
Numeric vector of survival probabilities at the end of each interval.
}
\description{
Computes survival at the end of each interval for competing risks with
piecewise constant hazards.
}
\examples{
id <- c(1L, 1L, 2L, 2L)
dt <- c(1, 1, 1, 1)
haz <- rbind(
  c(0.10, 0.05),
  c(0.20, 0.10),
  c(0.05, 0.02),
  c(0.10, 0.03)
)
pch_survival(id = id, dt = dt, haz = haz)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.base_learner}
\alias{predict.base_learner}
\title{Predict hazards, survival and absolute risk from a fitted base learner}
\usage{
\method{predict}{base_learner}(object, newdata, times, cause = 1, ...)
}
\arguments{
\item{object}{\code{base_learner}. A fitted object returned by \code{\link[=fit_learner]{fit_learner()}}.
It contains the learner specification in \code{object$model} and cause-specific fitted
models in \code{object$learner_fit}.}

\item{newdata}{\code{data.frame}/\code{data.table}. New covariate data (one row per subject).
If \code{newdata} contains the original \code{event_time}, \code{status}, or \code{id} columns used
for fitting, they are ignored for prediction.}

\item{times}{\code{numeric}. Prediction horizon(s). May include \code{0}.
Times larger than \code{object$data_info$maximum_followup} are not supported:
if \strong{all} requested times exceed the maximum follow-up, a warning is issued and
\code{NULL} is returned; if only \strong{some} exceed, output rows for those times are
returned with \code{NA} predictions.}

\item{cause}{\code{numeric(1)}. Cause index (1, 2, ...) used for the \code{absolute_risk}
calculation.}

\item{...}{Additional arguments (currently ignored).}
}
\value{
A \code{data.table} with one row per \verb{(row in newdata, time in times)} and columns:
\describe{
\item{(original columns)}{All columns from \code{newdata} (excluding ignored event columns).}
\item{time column}{A column with name \code{object$data_info$event_time} holding the requested horizon.}
\item{pwch_1, pwch_2, ...}{Predicted cause-specific piecewise hazards at the horizon.}
\item{survival_function}{Predicted survival probability at the horizon.}
\item{absolute_risk}{Predicted cumulative incidence (absolute risk) for \code{cause} at the horizon.}
}
}
\description{
Computes \strong{cause-specific piecewise-constant hazards} (\code{pwch_k}), the corresponding
\strong{survival function}, and \strong{absolute risk} for a given cause, at user-supplied
prediction horizons \code{times}, using a fitted \code{base_learner} object (single learner;
no stacking).
}
\details{
Internally, \code{newdata} is expanded to a Cartesian product with \code{times}, converted to
long Poisson format on \code{object$data_info$nodes}, and the fitted learner for each
cause in \code{object$learner_fit} is used to predict the cause-specific hazards.
Survival and absolute risk are then computed from the predicted hazards.

\strong{Special case \code{times = 0}:} when \code{0} is included in \code{times}, the returned rows
have \code{survival_function = 1}, \code{absolute_risk = 0}, and all \code{pwch_k = 0} at time 0.

\strong{Identifiers in the output:} if \code{newdata} contains the \code{id} column, it is carried
into the output. If \code{newdata} does not contain an id column, an internal id is
created for computation, but it is not guaranteed to appear in the returned table
unless it was present in \code{newdata}.
}
\examples{
d <- simulateStenoT1(120, competing_risks = TRUE)
lrn <- Learner_glmnet(covariates = c("age", "value_LDL"), lambda = 0, cross_validation = FALSE)
bl <- fit_learner(d, learner = lrn, id="id", status="status_cvd", event_time="time_cvd",
                  number_of_nodes=8)
p <- predict(bl, newdata = d[1:5], times = c(0, 2, 5), cause = 1)
head(p)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coef.R
\name{coef.poisson_superlearner}
\alias{coef.poisson_superlearner}
\title{Extract stacking (meta-learner) coefficients from a fitted Poisson Super Learner}
\usage{
\method{coef}{poisson_superlearner}(object, cause = NULL, model = "sl", ...)
}
\arguments{
\item{object}{\code{poisson_superlearner}. A fitted ensemble returned by \code{\link[=Superlearner]{Superlearner()}}.}

\item{cause}{\code{numeric(1)} or \code{NULL}. Which cause to extract meta-learner
coefficients for. If \code{NULL}, coefficients are returned for all causes.
Causes are indexed \verb{1, 2, ..., object$data_info$n_crisks}.}

\item{model}{Scalar model selector. Default is \code{"sl"} for the stacked super learner.
Other allowed values are:
\describe{
\item{\code{0} or \code{"sl"}}{Use the super learner prediction.}
\item{learner label}{Use one stored base learner by its label in
\code{object$data_info$learners_labels}.}
\item{\code{"learner_j"}}{Use the \code{j}-th stored learner.}
\item{integer \code{j >= 1}}{Use the \code{j}-th stored learner.}
}}

\item{...}{Passed to the underlying \code{coef()} method of the fitted meta-learner
(learner-dependent; e.g., \code{s} for \code{glmnet}).}
}
\value{
If \code{cause} is a single integer, returns the coefficient object produced by
\code{coef()} for the cause-specific fitted meta-learner.

If \code{cause = NULL}, returns a \code{list} of length \code{object$data_info$n_crisks},
where element \verb{[[k]]} contains meta-learner coefficients for cause \code{k}.

If no fitted ensemble is present (\code{object$superlearner} is \code{NULL}), signals a message
and returns \code{invisible(object)}.
}
\description{
Extracts the \strong{meta-learner coefficients} (stacking weights) from a fitted
\code{poisson_superlearner} object returned by \code{\link[=Superlearner]{Superlearner()}}.
}
\details{
For each cause \code{k}, the ensemble stores a fitted meta-learner in
\code{object$superlearner[[k]]$meta_learner_fit}. This method dispatches to the
underlying \code{coef()} method for that fitted meta-learner.

\strong{What coefficients represent.} These coefficients correspond to the meta-learner
regression of the outcome on the cross-validated base-learner predictions
(\code{Z1}, \code{Z2}, ...). Under the default meta-learner, they are the stacking
weights (on the scale defined by the meta-learner).

\strong{Learner-dependent output.} The returned coefficient object depends on the
meta-learner implementation (by default a \code{glmnet} fit, often returning a sparse
matrix). This method does not rename \verb{Z*} terms or post-process coefficients; it
returns the output of \code{coef(object$superlearner[[k]]$meta_learner_fit, ...)}
unchanged.

\strong{Single-learner special case.} If the ensemble was fit with only one base learner,
no meta-learner is fit and \code{meta_learner_fit} is \code{NULL}. In that case, \code{coef()}
for the \code{poisson_superlearner} does not have meta-learner coefficients to return.
}
\examples{
d <- simulateStenoT1(50, competing_risks = TRUE)
learners <- list(
  glm = Learner_glmnet(covariates = c("age", "value_LDL"), lambda = 0, cross_validation = FALSE),
  gam = Learner_gam(covariates = c("age", "value_LDL"))
)
fit <- Superlearner(d, id="id", status="status_cvd", event_time="time_cvd",
                    learners=learners, number_of_nodes=4, nfold=2)

# meta-learner coefficients (cause 1)
coef(fit, cause = 1)

# meta-learner coefficients for all causes (list)
coef(fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.poisson_superlearner}
\alias{summary.poisson_superlearner}
\title{Summarize a fitted Poisson Super Learner object}
\usage{
\method{summary}{poisson_superlearner}(object, cause = NULL, model = "sl", ...)
}
\arguments{
\item{object}{\code{poisson_superlearner} returned by \code{\link[=Superlearner]{Superlearner()}}.}

\item{cause}{\code{numeric(1)} or \code{NULL}. Which cause’s meta-learner fit to print.
If \code{NULL}, prints one line per cause (classes only) instead of printing the full
fitted objects.}

\item{model}{Scalar model selector. Default is \code{"sl"} for the stacked super learner.
Other allowed values are:
\describe{
\item{\code{0} or \code{"sl"}}{Use the super learner prediction.}
\item{learner label}{Use one stored base learner by its label in
\code{object$data_info$learners_labels}.}
\item{\code{"learner_j"}}{Use the \code{j}-th stored learner.}
\item{integer \code{j >= 1}}{Use the \code{j}-th stored learner.}
}}

\item{...}{Passed to the underlying \code{coef()} method for the fitted meta-learner
(learner-dependent; e.g. \code{s} for \code{glmnet}).}
}
\value{
Invisibly returns a \code{list} with elements:
\describe{
\item{cross_validation_deviance}{\code{data.table} (or \code{NULL}).}
\item{meta_coefficients}{List of length \code{n_crisks} with cause-specific coefficient objects (or \code{NULL}).}
}
}
\description{
Prints:
\enumerate{
\item a compact description of the fitted ensemble,
\item cross-validated deviances for base learners (when available),
\item cause-specific meta-learner coefficients (stacking weights).
}
}
\examples{
d <- simulateStenoT1(30, competing_risks = TRUE)

learners <- list(
  lasso = Learner_glmnet(
    covariates = "sex",
    alpha = 1,
    lambda = 0.01,
    cross_validation = FALSE
  ),
  ridge = Learner_glmnet(
    covariates = c("sex", "value_LDL"),
    alpha = 0,
    lambda = 0.01,
    cross_validation = FALSE
  )
)

fit <- Superlearner(
  data = d,
  id = "id",
  status = "status_cvd",
  event_time = "time_cvd",
  learners = learners,
  number_of_nodes = 3,
  nfold = 2
)

s <- summary(fit, cause = 1)
names(s)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.base_learner}
\alias{summary.base_learner}
\title{Summarize a fitted base learner object}
\usage{
\method{summary}{base_learner}(object, cause = 1, ...)
}
\arguments{
\item{object}{\code{base_learner} returned by \code{\link[=fit_learner]{fit_learner()}}.}

\item{cause}{\code{numeric(1)} or \code{NULL}. Which cause to summarize. If \code{NULL},
returns one summary per cause.}

\item{...}{Passed to the underlying \code{summary()} method (learner-dependent).}
}
\value{
If \code{cause} is a single integer, returns the underlying model summary for
that cause. If \code{cause = NULL}, returns a list of summaries (one per cause).
}
\description{
Dispatches to the underlying fitted model’s \code{summary()} method for the selected
cause, or returns a list of summaries for all causes.
}
\examples{
d <- simulateStenoT1(30, competing_risks = TRUE)
lrn <- Learner_glmnet(
  covariates = c("sex", "value_LDL"),
  lambda = 0.01,
  cross_validation = FALSE
)
bl <- fit_learner(
  d,
  learner = lrn,
  id = "id",
  status = "status_cvd",
  event_time = "time_cvd",
  number_of_nodes = 3
)
out <- summary(bl, cause = 1)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulateStenoT1.R
\name{simulateStenoT1}
\alias{simulateStenoT1}
\title{Simulate time-to-event data for hypothetical type-1 diabetes patients}
\usage{
simulateStenoT1(
  n,
  coefficient_age = 0.05,
  coefficient_LDL = 0.1,
  value_diabetis = 0.02,
  keep = NULL,
  scenario = c("alpha", "beta"),
  competing_risks = FALSE
)
}
\arguments{
\item{n}{\code{numeric(1)}. Number of subjects to simulate.}

\item{coefficient_age}{\code{numeric(1)}. Log-hazard coefficient for age in the CVD
model (\code{time.event.1}).}

\item{coefficient_LDL}{\code{numeric(1)}. Log-hazard coefficient for LDL in the CVD
model (\code{time.event.1}).}

\item{value_diabetis}{\code{numeric(1)}. Log-hazard coefficient for diabetes duration
in the CVD model (\code{time.event.1}).}

\item{keep}{\code{character} or \code{NULL}. Optional subset of columns to retain. If
supplied, only those columns are returned.}

\item{scenario}{\code{character(1)}. One of \code{"alpha"} or \code{"beta"}. Scenario \code{"beta"}
modifies the CVD hazard by adding nonlinear hinge-squared terms in age and LDL.}

\item{competing_risks}{\code{logical(1)}. If \code{TRUE} and \code{scenario = "alpha"}, simulates
two event causes (CVD and death without CVD). Otherwise simulates CVD vs censoring.}
}
\value{
A \code{data.table} with at least the following columns:

\describe{
\item{id}{\code{integer}. Subject identifier (1, ..., \code{n}).}
\item{time_cvd}{\code{numeric}. Observed follow-up time (minimum of event and censoring times;
also includes competing risk time if \code{competing_risks = TRUE} in scenario \code{"alpha"}).}
\item{status_cvd}{\code{integer}. Observed event status:
\code{0} = censored, \code{1} = CVD, and if \code{competing_risks = TRUE} in scenario \code{"alpha"},
\code{2} = death without prior CVD.}
\item{time}{\code{numeric}. Alias of \code{time_cvd} (kept for convenience).}
\item{event}{\code{integer}. Alias of \code{status_cvd} (kept for convenience).}
\item{uncensored_time_cvd}{\code{numeric}. Event time ignoring censoring
(minimum of event causes only).}
\item{uncensored_status_cvd}{\code{integer}. Event cause ignoring censoring.
In scenario \code{"alpha"} this is \code{1} (CVD) or \code{2} (death without CVD);
in scenario \code{"beta"} this is always \code{1}.}
\item{uncensored_time}{\code{numeric}. Alias of \code{uncensored_time_cvd}.}
\item{uncensored_event}{\code{integer}. Alias of \code{uncensored_status_cvd}.}
}

In addition, the returned table contains all baseline covariates listed in
\strong{Details}. Internal latent variables used only for simulation are removed
before returning (e.g., log2 eGFR components and, in scenario \code{"beta"}, the
hinge-squared features).
}
\description{
Simulate synthetic data inspired by the Steno Type-1 risk engine
}
\details{
Generates baseline covariates and event times for CVD and censoring, with an
optional competing-risks setting, for examples, benchmarks and tests.

The simulator uses a structural equation model (via \code{lava::lvm}) to generate
realistic correlations between covariates. Event times are then generated from
cause-specific Weibull proportional hazards models, where the linear predictor
depends on the simulated covariates (and scenario).

The following baseline covariates are generated (column name, type, interpretation):

\describe{
\item{sex}{\code{factor}. Binary sex indicator (generated Bernoulli, then stored as factor).}
\item{age}{\code{numeric}. Age at baseline (years).}
\item{diabetes_duration}{\code{numeric}. Duration of diabetes at baseline (years).}
\item{value_SBP}{\code{numeric}. Systolic blood pressure (SBP).}
\item{value_LDL}{\code{numeric}. LDL cholesterol.}
\item{value_HBA1C}{\code{numeric}. HbA1c.}
\item{value_Albuminuria}{\code{factor} with levels \code{Normal}, \code{Micro}, \code{Macro}. Albuminuria category.}
\item{eGFR}{\code{numeric}. Estimated glomerular filtration rate, constructed from latent
age-dependent log2 eGFR components (higher values indicate better kidney function).}
\item{value_Smoking}{\code{factor}. Smoking indicator (generated from a logistic model, then stored as factor).}
\item{value_Motion}{\code{factor}. Physical activity indicator (generated from a logistic model, then stored as factor).}
}

Event time variables are generated from latent Weibull PH models:
\code{time.event.1} (CVD), \code{time.event.0} (censoring), and in scenario \code{"alpha"} also
\code{time.event.2} (death without prior CVD). These latent variables are used to
construct the observed outcome variables returned by the function (see below).
}
\examples{
simulateStenoT1(n = 20, scenario = "alpha", competing_risks = TRUE)
}
\author{
Thomas A. Gerds \href{mailto:tag@biostat.ku.dk}{tag@biostat.ku.dk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_learner.R
\name{fit_learner}
\alias{fit_learner}
\title{Fit a single base learner}
\usage{
fit_learner(
  data,
  learner,
  id = "id",
  stratified_k_fold = FALSE,
  status = "status",
  event_time = NULL,
  number_of_nodes = NULL,
  nodes = NULL,
  variable_transformation = NULL,
  ...
)
}
\arguments{
\item{data}{\code{data.frame}. Subject-level input data (one row per subject).}

\item{learner}{Reference-class learner object (e.g. from \code{\link[=Learner_glmnet]{Learner_glmnet()}},
\code{\link[=Learner_hal]{Learner_hal()}} or \code{\link[=Learner_gam]{Learner_gam()}}). Must implement a \verb{$private_fit(dt_long)}
method that fits the learner on long Poisson data for one cause.}

\item{id}{\code{character(1)}. Name of the subject identifier column. If not found
in \code{data}, an \code{id} column is created automatically.}

\item{stratified_k_fold}{\code{logical(1)}. Reserved argument for future fold strategy.
Currently ignored.}

\item{status}{\code{character(1)}. Name of the event-status column.
Must be coded with \code{0} = censoring and \verb{1,2,...,K} for event types (causes).
If there is no \code{0} in \code{status}, the data are treated as uncensored.}

\item{event_time}{\code{character(1)}. Name of the event/censoring time column.
Must be present in \code{data}.}

\item{number_of_nodes}{\code{numeric(1)} or \code{NULL}. If not \code{NULL}, constructs a
quantile-based node grid with \code{number_of_nodes + 1} cut points (including
endpoints), then adds \code{0} if missing.}

\item{nodes}{\code{numeric} or \code{NULL}. Explicit time-node grid (cut points). If
supplied, \code{number_of_nodes} is ignored. \code{0} is added if missing. Nodes
beyond \code{max(event_time)} are dropped.}

\item{variable_transformation}{\code{list}/\code{character}/\code{formula} or \code{NULL}.
Optional transformations applied to the internally created long Poisson data
before fitting (via \code{apply_transformations()}).}

\item{...}{Additional arguments currently ignored.}
}
\value{
An object of class \code{base_learner}, i.e. a named \code{list} with:

\describe{
\item{model}{The \strong{learner object} that was fit (the input \code{learner}), stored
for later prediction. This contains the learner specification (e.g.,
covariates, tuning parameters).}

\item{learner_fit}{A \code{list} of fitted model objects, \strong{one per cause}.
Its length equals \code{data_info$n_crisks}. The list is created by splitting the
internally pre-processed long data by cause indicator \code{k} and calling
\code{model$private_fit()} on each split.

\itemize{
\item Names typically correspond to the cause labels \code{"1"}, \code{"2"}, ..., \code{"K"}.
\item Each element is \strong{learner-dependent}: e.g. for \code{Learner_glmnet} it
may be a \code{"glmnet"} (often wrapped, e.g. \code{"fishnet"}) fit; for other
learners it will be whatever \verb{$private_fit()} returns.
\item Each fitted object is trained on long Poisson data representing the
piecewise-constant hazard for that cause across the node intervals.
}}

\item{data_info}{A \code{list} of bookkeeping information needed for prediction and
interpretation:
\describe{
\item{id}{Identifier column name used.}
\item{status}{Status column name used.}
\item{event_time}{Event/censoring time column name used.}
\item{nodes}{Numeric vector of node cut points used for the piecewise grid
(includes \code{0} and is sorted). These are the interval boundaries used in
the long Poisson representation.}
\item{maximum_followup}{\code{max(data[[event_time]])}.}
\item{n_crisks}{Number of event types (causes) detected.
If censoring is present (\code{0} in \code{status}), then \verb{n_crisks = #unique(status) - 1};
otherwise \verb{n_crisks = #unique(status)}.}
\item{variable_transformation}{The transformation specification passed in
\code{variable_transformation} (or \code{NULL}).}
}}
}
}
\description{
Pre-processes subject-level time-to-event data into a long Poisson format on a
piecewise-constant time grid, then fits \strong{one} initialized learner object.
For competing risks, a separate model is fit for each event type (cause)
using the standard cause-specific Poisson likelihood on the long data.
}
\examples{
d <- simulateStenoT1(50, competing_risks = TRUE)
lrn <- Learner_glmnet(covariates = c("age", "value_LDL"),
                      lambda = 0,
                      cross_validation = FALSE)
bl <- fit_learner(d,
                  learner = lrn,
                  id = "id",
                  status = "status_cvd",
                  event_time = "time_cvd",
                  number_of_nodes = 4)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictRisk.R
\name{predictRisk.base_learner}
\alias{predictRisk.base_learner}
\title{Absolute-risk matrix predictions for a fitted base learner}
\usage{
\method{predictRisk}{base_learner}(object, newdata, times, cause = 1, ...)
}
\arguments{
\item{object}{\code{base_learner}. Fitted object from \code{\link[=fit_learner]{fit_learner()}}.}

\item{newdata}{\code{data.frame}. New covariate data.}

\item{times}{\code{numeric}. Prediction times.}

\item{cause}{\code{numeric(1)}. Cause index.}

\item{...}{Unused.}
}
\value{
\code{numeric} matrix with \code{nrow(newdata)} rows and \code{length(times)} columns.
}
\description{
Absolute-risk matrix predictions for a fitted base learner
}
\examples{
d <- simulateStenoT1(30, competing_risks = TRUE)
lrn <- Learner_glmnet(
  covariates = c("sex", "value_LDL"),
  lambda = 0.01,
  cross_validation = FALSE
)
bl <- fit_learner(
  d,
  learner = lrn,
  id = "id",
  status = "status_cvd",
  event_time = "time_cvd",
  number_of_nodes = 3
)

if (requireNamespace("riskRegression", quietly = TRUE)) {
  riskRegression::predictRisk(bl, newdata = d[1:3], times = c(1, 3), cause = 1)
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/learners.R
\docType{class}
\name{Learner_glmnet-class}
\alias{Learner_glmnet-class}
\alias{Learner_glmnet}
\title{Penalized Poisson learner via \code{glmnet}}
\description{
\code{Learner_glmnet} is a Reference Class implementing the learner interface
used by \code{\link[=Superlearner]{Superlearner()}} and \code{\link[=fit_learner]{fit_learner()}}.
}
\details{
\strong{User-facing API:} users are expected to \strong{initialize} the learner (i.e.,
call \code{Learner_glmnet(...)}) and pass the resulting object to
\code{\link[=Superlearner]{Superlearner()}} or \code{\link[=fit_learner]{fit_learner()}}. The remaining methods documented below
(e.g., \code{private_fit()}, \code{private_predictor()}) are part of the internal learner
interface and are \strong{not meant to be called directly by users}.

\strong{Wrapper role:} this class is a user-friendly wrapper around the existing
\code{glmnet} implementation. The package-specific contribution is to provide a
piecewise-constant hazard workflow: create the long-format Poisson data with
offsets for time at risk, add the interval ("node") structure for the baseline
hazard when requested, and forward standard \code{glmnet} arguments supplied at
initialization to the backend fitter.
}

\section{Model}{

Let \eqn{0=t_0 < t_1 < \cdots < t_m}{0=t0 < t1 < ... < tm} denote time knots and
define interval indicators \eqn{I_k(t)=1\{t\in(t_k,t_{k+1}]\}}{Ik(t)=I(t in (tk, t{k+1}])}.
The piecewise-constant hazard model is
\deqn{
  \lambda(t \mid x) = \sum_{k=0}^{m} I_k(t)\,\lambda_k(x),
  \qquad \lambda_k(x) = \exp(\beta^\top x + \gamma_k).
}{
  lambda(t|x) = sum_{k=0}^m Ik(t) * lambda_k(x),
  lambda_k(x) = exp(beta^T x + gamma_k).
}
Penalization is applied to the regression coefficients through the \code{glmnet}
elastic-net penalty. If you want node (baseline) terms to be unpenalized, use
\code{penalty.factor} via \code{...} (and set it consistently with how your design matrix
encodes nodes).
}

\section{Fields}{

\describe{
\item{\code{covariates} (\code{character})}{Names of covariate columns used in the model.}
\item{\code{cross_validation} (\code{logical})}{If \code{TRUE}, chooses \code{lambda} by \code{glmnet::cv.glmnet}.}
\item{\code{intercept} (\code{logical})}{Whether to include an intercept in the backend fit.}
\item{\code{add_nodes} (\code{logical})}{If \code{TRUE}, include interval ("node") effects encoding the baseline hazard.}
\item{\code{lambda} (\code{numeric})}{If \code{cross_validation=FALSE}, the \code{lambda} used in the final fit.}
\item{\code{formula} (\code{character})}{Formula string used to create the design matrix in long format.}
\item{\code{learner} (\code{function})}{Backend fitter (\code{glmnet::glmnet} or \code{glmnet::cv.glmnet}).}
\item{\code{fit_arguments} (\code{list})}{Additional arguments forwarded to the backend fitter.}
}
}

\section{Methods (internal learner interface)}{

\describe{
\item{\code{initialize(...)}}{Construct and configure the learner. This is the only method users should call.}
\item{\code{private_fit(data, ...)}}{Internal. Fits a Poisson model with offset \code{log(tij)} on long-format data.}
\item{\code{private_predictor(model, newdata, ...)}}{Internal. Predicts hazards on the response scale for long-format \code{newdata}.}
}
}

\examples{
lrn <- Learner_glmnet(covariates = c("age", "sex"), alpha = 1, cross_validation = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pch_absolute_risk.R
\name{pch_absolute_risk_euler}
\alias{pch_absolute_risk_euler}
\title{Absolute risk (Euler approximation) for a cause under piecewise-constant hazards}
\usage{
pch_absolute_risk_euler(
  id,
  dt,
  haz,
  cause_idx,
  one_based = TRUE,
  na_is_zero = FALSE
)
}
\arguments{
\item{id}{Integer vector. Sorted by \code{id} then time.}

\item{dt}{Numeric vector of interval lengths.}

\item{haz}{Numeric matrix (n x C) of cause-specific hazards per interval.}

\item{cause_idx}{Integer. Index of the cause of interest (1-based by default).}

\item{one_based}{Logical. If \code{TRUE}, \code{cause_idx} is 1-based. If \code{FALSE}, 0-based.}

\item{na_is_zero}{Logical. If \code{TRUE}, treat NA/Inf hazards as zero.}
}
\value{
Numeric vector of cumulative incidence values (Euler approximation)
at the end of each interval.
}
\description{
Computes the cumulative incidence function using the first-order Euler
(discrete) approximation:
\deqn{F_j(t) \approx \sum S(t_{k-1}) \lambda_{j,k} \Delta t_k}{
    F_j(t) = approx sum S(t_{k-1}) lambda_{j,k} Delta t_k
  }
Grouped by \code{id}, this returns the cumulative incidence at the end of each interval.
}
\examples{
id <- c(1L, 1L, 2L, 2L)
dt <- c(1, 1, 1, 1)
haz <- rbind(
  c(0.10, 0.05),
  c(0.20, 0.10),
  c(0.05, 0.02),
  c(0.10, 0.03)
)
pch_absolute_risk_euler(id = id, dt = dt, haz = haz, cause_idx = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/superlearner.R
\name{Superlearner}
\alias{Superlearner}
\title{Fit a Poisson Super Learner ensemble}
\usage{
Superlearner(
  data,
  id = "id",
  status = "status",
  event_time = NULL,
  learners,
  number_of_nodes = NULL,
  nodes = NULL,
  variable_transformation = NULL,
  nfold = 3,
  ...
)
}
\arguments{
\item{data}{\code{data.frame}. Subject-level input data, one row per subject.}

\item{id}{\code{character(1)}. Name of the subject identifier column. If missing,
an \code{id} column is created automatically.}

\item{status}{\code{character(1)}. Name of the event-status column. It must be coded
with \code{0} for censoring and \verb{1, 2, ..., K} for event types. If there is no
\code{0} in \code{status}, the data are treated as uncensored.}

\item{event_time}{\code{character(1)}. Name of the event or censoring time column.}

\item{learners}{\code{list}. List of initialized learner reference-class objects,
for example \code{\link[=Learner_glmnet]{Learner_glmnet()}}, \code{\link[=Learner_hal]{Learner_hal()}}, or \code{\link[=Learner_gam]{Learner_gam()}}. If
unnamed, learners are named \code{"learner_1"}, \code{"learner_2"}, and so on. Each
learner must implement \verb{$private_fit(dt_long)} and
\verb{$private_predictor(model, newdata)}.}

\item{number_of_nodes}{\code{numeric(1)} or \code{NULL}. If not \code{NULL}, constructs a
quantile-based node grid with \code{number_of_nodes + 1} cut points. Ignored when
\code{nodes} is supplied.}

\item{nodes}{\code{numeric} or \code{NULL}. Explicit time-node grid. If supplied,
\code{number_of_nodes} is ignored. \code{0} is added if missing, and nodes larger than
\code{max(event_time)} are dropped.}

\item{variable_transformation}{Optional transformation specification passed to
\code{apply_transformations()} on the internally created long-format data.}

\item{nfold}{\code{numeric(1)}. Number of folds for cross-validation stacking.}

\item{...}{Additional arguments currently ignored.}
}
\value{
An object of class \code{poisson_superlearner}, stored as a named \code{list}
with the following components:

\code{learners}:
the retained base learner objects.

\code{metalearner}:
the meta-learner object used for stacking. If no stacking is performed because
only one learner remains, \code{metalearner} is \code{NULL}.

\code{superlearner}:
a \code{list} of length \code{data_info$n_crisks}, one entry per cause. For cause \code{k},
\code{superlearner[[k]]} is a \code{list} with two elements:
\itemize{
\item \code{learners_fit}: the fitted base learner object or objects for cause \code{k}.
If more than one learner is retained, this is a \code{list} with one fitted
object per retained learner. If only one learner remains, this is the
single fitted learner object itself.
\item \code{meta_learner_fit}: the fitted cause-specific meta-learner for cause
\code{k}. If no stacking is performed, this is \code{NULL}.
}

\code{cross_validation_deviance}:
a \code{data.table} with columns \code{learner} and \code{deviance}, giving the mean
cross-validated Poisson deviance for each retained base learner. This
component is present when cross-validated model comparison is available.

\code{data_info}:
a \code{list} of bookkeeping information used for prediction and interpretation,
containing:
\itemize{
\item \code{id}: identifier column name used.
\item \code{status}: status column name used.
\item \code{event_time}: event-time column name used.
\item \code{nodes}: numeric vector of node cut points used for the piecewise grid.
\item \code{nfold}: number of folds used for stacking.
\item \code{maximum_followup}: maximum observed follow-up time.
\item \code{n_crisks}: number of event types detected.
\item \code{learners_labels}: character vector of retained learner labels.
\item \code{variable_transformation}: the transformation specification passed in
\code{variable_transformation}, or \code{NULL}.
}
}
\description{
Fits an ensemble of cause-specific piecewise-constant hazard models using a
long-format Poisson representation and combines them through a meta-learner
(stacking).
}
\details{
Internally, the function:
\enumerate{
\item builds a time grid (\code{nodes}) and converts the subject-level data to a
long Poisson format;
\item fits each base learner once on the full long data for each cause;
\item removes learners that already fail on the full data;
\item uses \code{nfold} cross-validation to obtain out-of-sample base-learner
predictions (\code{Z1}, \code{Z2}, ...) for stacking;
\item removes learners whose cross-validated prediction column is entirely
missing for at least one cause;
\item fits a cause-specific meta-learner on the retained stacked predictions.
}

If all learners fail on the full data, the function stops with an error.
If only one learner remains after the full-data screening step or after the
cross-validation screening step, no meta-learner is fit. In that case,
\code{metalearner} is \code{NULL}, each \code{superlearner[[k]]$meta_learner_fit} is \code{NULL},
and prediction is based directly on the stored fitted base learner.Numeric
learner positions always refer to the learners actually retained in the
fitted object.
}
\examples{
data <- simulateStenoT1(50, competing_risks = TRUE)

learners <- list(
  glm = Learner_glmnet(
    covariates = c("sex", "value_LDL"),
    lambda = 0,
    cross_validation = FALSE
  ),
  ridge = Learner_glmnet(
    covariates = c("sex", "value_LDL"),
    alpha = 0,
    lambda = 0.01,
    cross_validation = FALSE
  )
)

fit <- Superlearner(
  data = data,
  id = "id",
  status = "status_cvd",
  event_time = "time_cvd",
  learners = learners,
  number_of_nodes = 3,
  nfold = 2
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.poisson_superlearner}
\alias{predict.poisson_superlearner}
\title{Predict hazards, survival and absolute risk from a fitted Poisson Super Learner}
\usage{
\method{predict}{poisson_superlearner}(object, newdata, times, cause = 1, model = "sl", ...)
}
\arguments{
\item{object}{\code{poisson_superlearner}. A fitted ensemble from \code{\link[=Superlearner]{Superlearner()}}.}

\item{newdata}{\code{data.frame}/\code{data.table}. New covariate data (one row per subject).
If \code{newdata} contains the original \code{event_time}, \code{status}, or \code{id} columns used
for fitting, they are ignored for prediction.}

\item{times}{\code{numeric}. Prediction horizon(s). May include \code{0}.
Times larger than \code{object$data_info$maximum_followup} are not supported:
if \strong{all} requested times exceed the maximum follow-up, a warning is issued and
\code{NULL} is returned; if only \strong{some} exceed, output rows for those times are
returned with \code{NA} predictions.}

\item{cause}{\code{numeric(1)}. Cause index (1, 2, ...) used for the \code{absolute_risk}
calculation.}

\item{model}{Scalar model selector. Default is \code{"sl"} for the stacked super learner.
Other allowed values are:
\describe{
\item{\code{0} or \code{"sl"}}{Use the super learner prediction.}
\item{learner label}{Use one stored base learner by its label in
\code{object$data_info$learners_labels}.}
\item{\code{"learner_j"}}{Use the \code{j}-th stored learner.}
\item{integer \code{j >= 1}}{Use the \code{j}-th stored learner.}
}
Numeric positions refer to the learners actually stored in the fitted object.}

\item{...}{Additional arguments (currently ignored).}
}
\value{
A \code{data.table} with one row per \verb{(row in newdata, time in times)} and columns:
\describe{
\item{(original columns)}{All columns from \code{newdata} (excluding ignored event columns).}
\item{time column}{A column with name \code{object$data_info$event_time} holding the requested horizon.}
\item{pwch_1, pwch_2, ...}{Predicted cause-specific piecewise hazards at the horizon.}
\item{survival_function}{Predicted survival probability at the horizon.}
\item{absolute_risk}{Predicted cumulative incidence (absolute risk) for \code{cause} at the horizon.}
}
}
\description{
Computes \strong{cause-specific piecewise-constant hazards} (\code{pwch_k}), the corresponding
\strong{survival function}, and \strong{absolute risk} for a given cause, at user-supplied
prediction horizons \code{times}, for each row in \code{newdata}.
}
\details{
Internally, \code{newdata} is expanded to a Cartesian product with the requested
\code{times}, converted to long Poisson format on \code{object$data_info$nodes}, and hazards
are predicted either from the stacked super learner (\code{model = "sl"}) or from one
selected fitted base learner. Survival and absolute risk are then computed from
the predicted hazards.

\strong{Special case \code{times = 0}:} when \code{0} is included in \code{times}, the returned rows
have \code{survival_function = 1}, \code{absolute_risk = 0}, and all \code{pwch_k = 0} at time 0.

\strong{Identifiers in the output:} if \code{newdata} contains the \code{id} column, it is carried
into the output. If \code{newdata} does not contain an id column, an internal id is
created for computation, but it is not guaranteed to appear in the returned table
unless it was present in \code{newdata}.
}
\examples{
d <- simulateStenoT1(30, competing_risks = TRUE)

learners <- list(
  lasso = Learner_glmnet(
    covariates = "sex",
    alpha = 1,
    lambda = 0.01,
    cross_validation = FALSE
  ),
  ridge = Learner_glmnet(
    covariates = c("sex", "value_LDL"),
    alpha = 0,
    lambda = 0.01,
    cross_validation = FALSE
  )
)

fit <- Superlearner(
  data = d,
  id = "id",
  status = "status_cvd",
  event_time = "time_cvd",
  learners = learners,
  number_of_nodes = 3,
  nfold = 2
)
p <- predict(fit, newdata = d[1:3], times = c(0, 2), cause = 1)
p[, .(id, time_cvd, absolute_risk)]


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.poisson_superlearner}
\alias{print.poisson_superlearner}
\title{Print method for \code{poisson_superlearner}}
\usage{
\method{print}{poisson_superlearner}(x, cause = 1, model = "sl", ...)
}
\arguments{
\item{x}{\code{poisson_superlearner} object returned by \code{\link[=Superlearner]{Superlearner()}}.}

\item{cause}{\code{numeric(1)} or \code{NULL}. Which cause’s meta-learner fit to print.
If \code{NULL}, prints one line per cause (classes only) instead of printing the full
fitted objects.}

\item{model}{Scalar model selector. Default is \code{"sl"} for the stacked super learner.
Other allowed values are:
\describe{
\item{\code{0} or \code{"sl"}}{Use the super learner prediction.}
\item{learner label}{Use one stored base learner by its label in
\code{object$data_info$learners_labels}.}
\item{\code{"learner_j"}}{Use the \code{j}-th stored learner.}
\item{integer \code{j >= 1}}{Use the \code{j}-th stored learner.}
}}

\item{...}{Passed to the underlying fitted meta-learner \code{print()} method when
\code{cause} is a single integer.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Prints a compact description of the fitted Poisson Super Learner, including the
number of base learners, the meta-learner, the time-grid used, and competing-risk
structure. Optionally prints the fitted meta-learner for a given cause.
}
\examples{
d <- simulateStenoT1(30, competing_risks = TRUE)

learners <- list(
  lasso = Learner_glmnet(
    covariates = "sex",
    alpha = 1,
    lambda = 0.01,
    cross_validation = FALSE
  ),
  ridge = Learner_glmnet(
    covariates = c("sex", "value_LDL"),
    alpha = 0,
    lambda = 0.01,
    cross_validation = FALSE
  )
)

fit <- Superlearner(
  data = d,
  id = "id",
  status = "status_cvd",
  event_time = "time_cvd",
  learners = learners,
  number_of_nodes = 3,
  nfold = 2
)

print(fit, cause = NULL)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictRisk.R
\name{predictRisk.poisson_superlearner}
\alias{predictRisk.poisson_superlearner}
\title{Absolute-risk matrix predictions for a fitted Poisson Super Learner}
\usage{
\method{predictRisk}{poisson_superlearner}(object, newdata, times, cause = 1, model = "sl", ...)
}
\arguments{
\item{object}{\code{poisson_superlearner}. Fitted object.}

\item{newdata}{\code{data.frame}. New covariate data.}

\item{times}{\code{numeric}. Prediction times.}

\item{cause}{\code{numeric(1)}. Cause index.}

\item{model}{Scalar model selector. Default is \code{"sl"}. Allowed values are the
same as in \code{\link[=predict.poisson_superlearner]{predict.poisson_superlearner()}}.}

\item{...}{Unused.}
}
\value{
\code{numeric} matrix with \code{nrow(newdata)} rows and \code{length(times)} columns.
}
\description{
S3 method compatible with \code{riskRegression::predictRisk} returning one column
per requested time.
}
\examples{
d <- simulateStenoT1(30, competing_risks = TRUE)

learners <- list(
  lasso = Learner_glmnet(
    covariates = "sex",
    alpha = 1,
    lambda = 0.01,
    cross_validation = FALSE
  ),
  ridge = Learner_glmnet(
    covariates = c("sex", "value_LDL"),
    alpha = 0,
    lambda = 0.01,
    cross_validation = FALSE
  )
)

fit <- Superlearner(
  data = d,
  id = "id",
  status = "status_cvd",
  event_time = "time_cvd",
  learners = learners,
  number_of_nodes = 3,
  nfold = 2
)

if (requireNamespace("riskRegression", quietly = TRUE)) {
  riskRegression::predictRisk(fit, newdata = d[1:3], times = c(1, 3), cause = 1)
}

}
